//
// DCUcontroller.m
// DCU
//
// Created by Christoph Birk on 2013-02-27
// Copyright 2013 Carnegie Observatories. All rights reserved.
//

/* ---------------------------------------------------------------- */

#define DEBUG              1

/* ---------------------------------------------------------------- */

#import "DCUcontroller.h"
#import "Notice.h"

/* ---------------------------------------------------------------- */

enum DCU_errors { E_DCU_error=1,E_DCU_config,E_DCU_busy,E_DCU_disabled,
                  E_DCU_ret,E_DCU_air,E_DCU_selector,E_DCU_failed,
                  E_DCU_last };

static const char* dcu_errors[] = { "DCUcontroller: error",
                                    "DCU: not configured",
                                    "DCU: busy",
                                    "DCU: disabled (requires PLC reset)",
                                    "DCU: retract-sensor not active",
                                    "DCU: air pressure too low",
                                    "DCU: selector failed (PLC interlock)",
                                    "DCU: sensor failed",
                                    "" };
                                    
/* ---------------------------------------------------------------- */

static char*     str2chr       (NSString*,char*,int);
static NSString* chr2str       (const char*);
static int       alertPanel    (const char*,const char*,int,
                                const char*,const char*,const char*);
static BOOL      alertQuestion (const char*,const char*,int,
                                const char*,const char*);
static float     level2volt    (int);

/* ---------------------------------------------------------------- */

@implementation DCUcontroller

@synthesize drop_ffs,drop_mcal;
@synthesize error_state;

/* ---------------------------------------------------------------- */

- (id)init
{
  NSString *file;
  NSNumber *number;
  char     buf[128];
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p)\n",PREFUN,self);
#endif

  if ((self = [super init]) == nil) return nil;
  
  configured = FALSE;
  
  NSBundle *bundle = [NSBundle mainBundle];
  NSString *r_path = [bundle resourcePath];
  file = [r_path stringByAppendingString:@"/Bonjour.icns"];
  icon_on  = [[NSImage alloc] initWithContentsOfFile:file];
  file = [r_path stringByAppendingString:@"/AlertCautionIcon.icns"];
  icon_err = [[NSImage alloc] initWithContentsOfFile:file];
  icon_off = nil;
  
  s_title = [[bundle objectForInfoDictionaryKey:@"CFBundleExecutable"] retain];
  str2chr(s_title,p_title,sizeof(p_title));
  
  for (int i=0; i<N_LAMPS; i++) lamps[i] = nil;
  
  logger = [[Logger alloc] initWithBase:nil];
#if (DEBUG > 0)
  [logger append:PREFUN];
#endif

  NSUserDefaults *defaults = [NSUserDefaults standardUserDefaults];

  number = [defaults objectForKey:@"watchdog_interval"];
  interval_dog = (number) ? [number intValue] : 60;
  sprintf(buf,"watchdog_interval= %d",interval_dog); [logger append:buf];
  next_dog = time(NULL) + interval_dog;  // v0.902

  number = [defaults objectForKey:@"min_pressure"];
  min_pressure = (number) ? [number floatValue] : 90.0f; 
  sprintf(buf,"min_pressure= %.1f",min_pressure); [logger append:buf];
  air_pressure = min_pressure+5.0f;    // Simulator preset
  
  return self;
}

/* ---------------------------------------------------------------- */

- (void)awakeFromNib
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p)\n",PREFUN,self);
#endif

  but_lamps = [[NSArray alloc] initWithObjects:but_lamp1,but_lamp2,
    but_lamp3,but_lamp4,but_lamp5,but_lamp6,but_lamp7,but_lamp8,nil];
#ifndef NDEBUG
  { NSSet *set = [NSSet setWithArray:but_lamps]; assert([set count] >= N_LAMPS); }
#endif
  
  assert([NSThread isMainThread]);     // Note: timer on main thread

  [Notice openNotice:NULL Title:NULL]; // force initialization on main thread

  [NSTimer scheduledTimerWithTimeInterval:2.0
           target:self selector:@selector(fireTimer:)
           userInfo:nil repeats:YES];
}

/* ---------------------------------------------------------------- */

- (void)terminate:(const char*)text
{
  if (text) { CFStringRef str;
    [logger message:text Level:2 File:YES];
    sleep(3);
    str = CFStringCreateWithCString(kCFAllocatorDefault,text,kCFStringEncodingASCII);
    CFUserNotificationDisplayNotice(0,
      kCFUserNotificationStopAlertLevel | kCFUserNotificationNoDefaultButtonFlag,
      NULL,NULL,NULL,CFSTR(P_TITLE),str,CFSTR("OK"));
    CFRelease(str);
  }
  [logger append:"terminate"];
  [[NSApplication sharedApplication] terminate:self];  
}

/* ---------------------------------------------------------------- */

- (BOOL)configure
{
  int  i;
  char buf[256];
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,self);
#endif

  BOOL was_configured = configured;
  configured = FALSE;
  [self updateState:NO];                  // disable everything
  if (was_configured) [self shutdown];    // close all connections

  NSUserDefaults *defaults = [NSUserDefaults standardUserDefaults];
#ifdef NDEBUG
  use_interlocks = YES;
#else
  use_interlocks = ![defaults boolForKey:@"no_interlocks"];
#endif
  sprintf(buf,"use_interlocks= %d",(int)use_interlocks); [logger append:buf];
  eng_mode = [defaults boolForKey:KEY_ENGMODE];
  sprintf(buf,"eng_mode= %d",(int)eng_mode); [logger append:buf];
  
  /* --- determine telescope and TCSIS host --- */
  
  if (telname) { [telname release]; telname = nil; }
  telname = [defaults stringForKey:KEY_TELNAME]; 
  if (!telname) { [self terminate:"no telescope defined"]; return FALSE; }
#ifdef SIM_ONLY
#ifdef NDEBUG
  strcpy(telhost,"localhost");
#else
  strcpy(telhost,"andromeda");
#endif
#else // !SIM_ONLY
  *telhost = '\0';
  if      ([telname compare:@"Baade"] == NSOrderedSame) strcpy(telhost,"mag1tcs");
  else if ([telname compare:@"Clay"]  == NSOrderedSame) strcpy(telhost,"mag2tcs");
#endif
  if (*telhost == '\0') { 
    [self terminate:"invalid telescope\rdelete Preferences and re-start"]; 
    return FALSE;
  }
  [telname retain];

  NSString *title = [NSString stringWithFormat:@"%@/%@ ... connecting to PLC ...",
                                      s_title,telname];
  [window setTitle:title];

  /* --- setup button/lamp labels --- */
  
  NSString *path  = [[NSBundle mainBundle] pathForResource:telname ofType:@"xml"];
  NSDictionary *dict = [NSDictionary dictionaryWithContentsOfFile:path];
  if (dict) { NSString *key,*label; NSButton *button;
    for (i=0; i<N_LAMPS; i++) {
      key = [NSString stringWithFormat:@"Lamp%d_Name",(1+i)];
      label = [dict objectForKey:key];
      assert(label);
      button = [but_lamps objectAtIndex:i];
      [button setTitle:label];
    }
    label = [dict objectForKey:@"varLamp_Name"];
    assert(label);
    [label_quartz setStringValue:label];
#if 1 // hide controls ?Povilas
    [label_quartz setHidden:([label compare:@"-"] == NSOrderedSame)];
    [popup_quartz setHidden:([label compare:@"-"] == NSOrderedSame)];
    [edit_quartz  setHidden:([label compare:@"-"] == NSOrderedSame)];
#endif
  }
  
  /* --- get some PLC addresses --- */

  cookie_addr = o2d([[dict objectForKey:@"Cookie_Address"] intValue]);
#if (DEBUG > 1)
  fprintf(stderr,"cookie_addr=%d\n",cookie_addr);
#endif
  assert(cookie_addr > 0);
  
  selector_addr = o2d([[dict objectForKey:@"Selector_Output"] intValue]);
#if (DEBUG > 1)
  fprintf(stderr,"selector_addr=%d\n",selector_addr);
#endif
  assert(selector_addr > 0);
  
  error_addr = o2d([[dict objectForKey:@"Error_Address"] intValue]);
#if (DEBUG > 1)
  fprintf(stderr,"error_addr=%d\n",error_addr);
#endif
  assert(error_addr > 0);
  
  /* --- release FFmodel devices --- */
  
  if (tcs) { [tcs release]; tcs=nil; } 
  if (plc) { [plc release]; plc=nil; }
  
  for (i=0; i<N_LAMPS; i++) {
    if (lamps[i]) {                    // remove old devices
      [lamps[i] removeObserver:self forKeyPath:KEY_MOVING];
      [lamps[i] release]; lamps[i] = nil;
    }
  }
  if (q_lamp) {                        // remove old device
    [q_lamp removeObserver:self forKeyPath:KEY_MOVING];
    [q_lamp release]; q_lamp = nil;
  }
  if (ff_enable) { [ff_enable release]; ff_enable = nil; }
  if (ff_screen) {                     // remove old device
    [ff_screen.brake.pswitch removeObserver:self forKeyPath:KEY_STATUS];
    [ff_screen.brake removeObserver:self forKeyPath:KEY_MOVING];
    [ff_screen removeObserver:self forKeyPath:KEY_MOVING];
    [ff_screen release]; ff_screen = nil;
  }
  if (mc_enable) { [mc_enable release]; mc_enable = nil; }
  if (mc_screen) {                     // remove old device
    [mc_screen.brake.pswitch removeObserver:self forKeyPath:KEY_STATUS];
    [mc_screen.brake removeObserver:self forKeyPath:KEY_MOVING];
    [mc_screen removeObserver:self forKeyPath:KEY_MOVING];
    [mc_screen release]; mc_screen = nil;
  }
  if (air_sensor) {
    [air_sensor release]; air_sensor = nil;
  }
  
  /* ---  the remainder of the configuration (I/O) is running in a thread */
  
  [self performSelectorInBackground:@selector(run_conf:) withObject:nil];
  
  return YES;
}

/* ---------------------------------------------------------------- */

- (void)run_conf:(id)param
{
  NSString *name,*string;
  int      status;
  char     host[128],buf[256];
  int      port,err=0;
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,param);
#endif

  NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
  
  NSUserDefaults *defaults = [NSUserDefaults standardUserDefaults];

  /* --- setup DCU PLC connection, check cookie --- */
  
  port = 502; *host = '\0';
  name = [defaults stringForKey:KEY_DCUNAME]; // Note: 192.91.178.187 at SBS
#ifdef SIM_ONLY
  port = 0xffff-port;
#ifdef NDEBUG
  if ([name compare:@"localhost"] != NSOrderedSame) name = @"localhost";
#else
  if ([name compare:@"localhost"] != NSOrderedSame) {
    if ([name compare:@"andromeda"] != NSOrderedSame) name = @"andromeda";
  }
#endif
#else  // !SIM_ONLY
  if ([name compare:@"andromeda"] == NSOrderedSame) port = 0xffff-port;
  if ([name compare:@"localhost"] == NSOrderedSame) port = 0xffff-port;
#endif
#if defined(SIM_ONLY) && defined(NDEBUG)
  status = 0;
#else
  status = [defaults integerForKey:KEY_DCUSTAT];
#endif
  str2chr(name,host,sizeof(host));
  plc = [[PLC_IO alloc] initWithHost:host Port:port Online:status];
  sprintf(buf,"PLC[%s:%d,%d]= %p",host,port,status,plc); [logger append:buf];
  if (plc.up_and_running) {            // set/check cookie
    if (!plc.online) { cookie = 0; sleep(1); }
    err = [plc get32bit:cookie_addr Value:&cookie];
    if (!err) { BOOL r=YES;
      sprintf(buf,"get cookie=0x%x",cookie); [logger append:buf];
      if (cookie != 0) {
        sprintf(buf,"The DCU-PLC might already be in use.\rContinue anyway?");
        r = alertQuestion(p_title,buf,1,"No","Yes");
      }
      if (r) {
        cookie = (getpid() << 20) ^ time(NULL); 
        sprintf(buf,"set cookie=0x%x",cookie); [logger append:buf];
        err = [plc set32bit:cookie_addr Value:cookie];
      } else {
        [self terminate:NULL]; 
        return;
      }
    }
    if (err) {
      sprintf(buf,"setting cookie failed (err=%d)\r%s",err,ErrorText(err));
      [self terminate:buf]; return;
    }
  } else err = E_plcio_error;
  if (err) {                           // terminate application
    [self terminate:"connecting to PLC failed"]; return;
  }

  string = [NSString stringWithFormat:@"%@/%@",s_title,telname];
  [window performSelectorOnMainThread:@selector(setTitle:) withObject:string 
          waitUntilDone:YES];

  /* --- create FFmodel devices (lamps, screens) --- */

  NSString *path  = [[NSBundle mainBundle] pathForResource:telname ofType:@"xml"];
  NSDictionary *dict = [NSDictionary dictionaryWithContentsOfFile:path];
  
  for (int i=0; i<N_LAMPS; i++) {
    assert(!lamps[i]);
    NSButton *button = [but_lamps objectAtIndex:i];
    if ([[button title] compare:@"-"] == NSOrderedSame) continue;
    NSString *prefix = [NSString stringWithFormat:@"Lamp%d",(1+i)];
    lamps[i] = [[FFlamp alloc] initWithPLC:plc XML:dict Prefix:prefix];
    [lamps[i] addObserver:self forKeyPath:KEY_MOVING
                               options:NSKeyValueObservingOptionNew
                               context:(void*)0];
  }

  assert(!q_lamp);
  if ([[label_quartz stringValue] compare:@"-"] != NSOrderedSame) {
    q_lamp = [[FFvarLamp alloc] initWithPLC:plc XML:dict Prefix:@"varLamp"];
    [q_lamp addObserver:self forKeyPath:KEY_MOVING
                             options:NSKeyValueObservingOptionNew
                             context:(void*)0];
  }
  
  assert(!ff_enable);                  // create FFS-Enable coil
  ff_enable = [[FFcoil alloc] initWithPLC:plc XML:dict Prefix:@"FFS_Enable"];
  assert(!ff_screen);                  // create FFS-Screen ?
  ff_screen = [[FFscreen alloc] initWithPLC:plc XML:dict Prefix:@"FFS"];
  [ff_screen addObserver:self forKeyPath:KEY_MOVING
             options:NSKeyValueObservingOptionNew context:(void*)0];
  [ff_screen.brake addObserver:self forKeyPath:KEY_MOVING
                   options:NSKeyValueObservingOptionNew context:(void*)0];
  [ff_screen.brake.pswitch addObserver:self forKeyPath:KEY_STATUS
                           options:NSKeyValueObservingOptionNew context:(void*)0];
  ff_disabled = [defaults boolForKey:KEY_FFS_DSBLD];
  if (ff_disabled) [logger append:"FFS disabled"];
  
  assert(!mc_enable);
  mc_enable = [[FFcoil alloc] initWithPLC:plc XML:dict Prefix:@"MCAL_Enable"];
  assert(!mc_screen);                  // Mcal at 'Clay' only
  if ([telname compare:@"Clay"] == NSOrderedSame) {
    mc_screen = [[FFscreen alloc] initWithPLC:plc XML:dict Prefix:@"MCAL"];  
    [mc_screen addObserver:self forKeyPath:KEY_MOVING
               options:NSKeyValueObservingOptionNew context:(void*)0];
    [mc_screen.brake addObserver:self forKeyPath:KEY_MOVING
                     options:NSKeyValueObservingOptionNew context:(void*)0];
    [mc_screen.brake.pswitch addObserver:self forKeyPath:KEY_STATUS
                             options:NSKeyValueObservingOptionNew context:(void*)0];
  } else {
    [check_mcins setState:NSOffState];
    [check_mcret setState:NSOffState];
    [popup_mcbrake selectItemAtIndex:0];
  }
  mc_disabled = [defaults boolForKey:KEY_MCAL_DSBLD];
  if (mc_disabled) [logger append:"Mcal disabled"];

  assert(!air_sensor);
  air_sensor = [[FFsensor alloc] initWithPLC:plc XML:dict Prefix:@"AIR"];
  
  /* --- setup TCSIS connection --- */
  
  port = 5801;
  status = [defaults integerForKey:KEY_TELSTAT];
  tcs = [[TEL_IO alloc] initWithHost:telhost Port:port Online:status];
  sprintf(buf,"TCS[%s:%d,%d]= %p",telhost,port,status,tcs); [logger append:buf];

  err = [self getFullStatus];
  if (err) {                           // terminate application
    [self terminate:"reading status from PLC failed"]; return;
  }
  
  configured = TRUE;
  [self updateState:YES];              // send full state to TCS

  [pool drain];
}

/* ---------------------------------------------------------------- */

- (void)fireTimer:(NSTimer*)theTimer   
{
#if (DEBUG > 2)
  fprintf(stderr,"%s(%p): counter=%lu\n",PREFUN,theTimer,counter);
#endif

 if (!configured) return;
 
#ifndef SIM_ONLY
  if (!plc.online) {                   // Simulator
#endif
    NSString *string;
    if (counter % 2) string = [NSString stringWithFormat:@"%@/%@",s_title,telname];
    else             string = [NSString stringWithFormat:@"%@ (Simulator)",s_title];
    [window setTitle:string];          // v0.902
#ifndef SIM_ONLY
  }
#endif

  if ([self running]) return;          // v0.808
      
  if (plc.up_and_running && interval_dog) { // check cookie
    if (time(NULL) >= next_dog) { 
      [self performSelectorInBackground:@selector(run_dog) withObject:nil];
      next_dog = time(NULL) + interval_dog;
    }
  }

  counter++;
}

/* ---------------------------------------------------------------- */

- (int)getFullStatus
{
  int err;
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,self);
#endif
  
  err = [plc open]; 
  if (!err) {
    for (int i=0; i<N_LAMPS; i++) {
      err = [lamps[i] get]; if (err) break; // Note: [nil] is ok
    }
    if (!err) err = [q_lamp get];           // Note: [nil] is ok
    if (!err) err = [ff_screen get];
    if (!err) err = [mc_screen get];
    if (!err) [self checkAir:FALSE];
    [plc close];
  } 
  
  if (err) { char buf[128];            
    sprintf(buf,"getting status from PLC failed (err=%d)",err);
    [Notice openNotice:buf Title:p_title Time:0 Level:2];
  }
  
  return err;
}

/* ---------------------------------------------------------------- */

- (void)shutdown 
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,self);
#endif

  if (plc.up_and_running) {
    [logger append:"set cookie=0"];
    [plc set32bit:cookie_addr Value:0];
    [logger append:"PLC shutdown"];
    [plc shutdown];
  }
  if (tcs.up_and_running) { 
    [logger append:"TCS shutdown"];
    [tcs shutdown];
  }
}

/* ---------------------------------------------------------------- */

- (BOOL)running
{
  if (prog_screen.counter) return TRUE;
  if (prog_mcal.counter) return TRUE;
  return FALSE;
}

- (BOOL)deployed
{
  if (ff_screen.status == SCREEN_INS) return TRUE;
  if (mc_screen.status == SCREEN_INS) return TRUE;
  return FALSE;
}

/* ---------------------------------------------------------------- */

- (void)action_lamp:(id)sender
{
  int  req;
  char buf[256];
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,sender);
#endif
  assert([NSThread isMainThread]);

  [sender setEnabled:NO];

  if (sender == popup_quartz) {        // variable quartz lamp
    assert(q_lamp);
    req = 1000*[sender indexOfSelectedItem];  // TODO
    sprintf(buf,"%s: var= %d",PREFUN,req); [logger append:buf];
    NSNumber *pos = [NSNumber numberWithInt:req];
    [q_lamp performSelectorInBackground:@selector(run:) withObject:pos];
  } else { int i;                      // fixed level lamps
    i = [but_lamps indexOfObject:sender];
    assert(i < N_LAMPS);
    assert(lamps[i]);
    req = (lamps[i].status != LAMP_OFF) ? LAMP_OFF : LAMP_ON;
    sprintf(buf,"%s: i= %d req= %d",PREFUN,i,req); [logger append:buf];
    NSNumber *pos = [NSNumber numberWithInt:req];
    [lamps[i] performSelectorInBackground:@selector(run:) withObject:pos];
  }
}

/* ---------------------------------------------------------------- */

- (void)action_ffs:(id)sender
{
  int  req=0,err=0;
  char buf[256];
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,sender);
#endif
  assert([NSThread isMainThread]);
  assert(ff_screen);

  BOOL script = ([sender isKindOfClass:[NSScriptCommand class]]) ? TRUE : FALSE;
  if (script) {                        // v0.900
    if (!eng_mode) {
      [Notice openNotice:"Scripting allowed in 'engineering mode' only" 
              Title:P_TITLE Time:30 Level:1];
      // TODO error_state = E_DCU_error; // stops script
      return;
    }
  }
  if (!configured) err = E_DCU_config;  // too early
  if ([self running]) err = E_DCU_busy; // catch scripting error v0.807

  if (!err) err = [plc open];
  if (!err) { int value=DCU_FFS;       // Simulator preset
    err = [self checkPLC:YES];
    if (mc_screen) {                   // v0.802
      if (!err) err = [mc_screen get]; // other DCU
      if (!err) {                      // is out of the way?
        sprintf(buf,"mc_screen.ret_limit= %d (%d)",
                mc_screen.ret_limit.status,mc_screen.ins_limit.status);
        [logger append:buf];
        if (use_interlocks) {
          if ((mc_screen.ret_limit.status != 1) || (mc_screen.ins_limit.status != 0)) {
            err = E_DCU_ret; 
          } 
        }
      }
    }
    if (!err) { if (![self checkAir:TRUE]) { err = E_DCU_air; } }
    if (!err) err = [mc_enable set:0]; // enable proper DCU v0.804
    if (!err) err = [ff_enable set:1];
    [NSThread sleepForTimeInterval:PLC_DELAY];  
    if (!err) err = [plc getOutput:selector_addr Value:&value];
    if (!err) { if (value != DCU_FFS) err = E_DCU_selector; }
    [plc close];
  }
  if (err) {
    error_state = err;
    [self updateFFS]; if (err == E_DCU_ret) [self updateMcal];
    sprintf(buf,"enabling FFS-DCU failed (err=%d)\r%s",err,ErrorText(err));
    [Notice openNotice:buf Title:P_TITLE Time:0 Level:2];
    return;
  }

  if (!script) [sender setEnabled:FALSE];

  if (sender == popup_ffbrake) {       // FFS brake
    req = [sender indexOfSelectedItem];
    sprintf(buf,"%s: brake= %d",PREFUN,req); [logger append:buf];
    NSNumber *pos = [NSNumber numberWithInt:req];
    [ff_screen.brake performSelectorInBackground:@selector(run:) withObject:pos];
  } else {                             // FFS screen
    switch ([drop_ffs indexOfSelectedItem]) {
      case 1: req = SCREEN_RET; break;
      case 2: req = SCREEN_INS; break;
      default: assert(0); break;
    }
    if (!eng_mode) {
      if (req == ff_screen.status) {   // v0.900
        [self updateFFS];
        sprintf(buf,"FF-Screen already %s",(req==SCREEN_RET) ? "retracted" : "deployed");
        [Notice openNotice:buf Title:P_TITLE Time:30 Level:1];
        return;
      }
    }
    sprintf(buf,"%s: screen= %d",PREFUN,req); [logger append:buf];
    NSNumber *pos = [NSNumber numberWithInt:req];
    [ff_screen performSelectorInBackground:@selector(run:) withObject:pos];
    if (req == SCREEN_RET) { int i;    // turn off all lamps
      for (i=0; i<N_LAMPS; i++) {
        if (!lamps[i]) continue;
        if (lamps[i].status != LAMP_OFF) {
          [self action_lamp:[but_lamps objectAtIndex:i]];
        }
      }
      if (q_lamp) {
        if (q_lamp.status != LAMP_OFF) {
          [popup_quartz selectItemAtIndex:0];
          [self action_lamp:popup_quartz];
        }
      }
    }
  }
}

/* ---------------------------------------------------------------- */

- (void)action_mcal:(id)sender         // takes about 35 seconds
{
  int  req=0,err=0;
  char buf[256];
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,sender);
#endif
  assert([NSThread isMainThread]);
  
  if (!mc_screen) {                    // might be called by script
    err = error_state = E_DCU_error;   // global 'error_state' v0.807
    sprintf(buf,"MCAL-DCU not available (err=%d)\r",err,ErrorText(err));
    [Notice openNotice:buf Title:P_TITLE Time:0 Level:2];
    return;
  }
  BOOL script = ([sender isKindOfClass:[NSScriptCommand class]]) ? TRUE : FALSE;
  if (script) {                        // v0.900
    if (!eng_mode) {
      [Notice openNotice:"Scripting allowed in 'engineering mode' only" 
              Title:P_TITLE Time:30 Level:1];
      // TODO error_state = E_DCU_error; // stops script
      return;
    }
  }
  
  if (!configured) err = E_DCU_config;   // too early
  if ([self running]) err = E_DCU_busy;  // catch scripting error
  
  if (!err) err = [plc open];
  if (!err) { int value=DCU_MCAL;      // Simulator preset
    err = [self checkPLC:YES];
    if (ff_screen) {                   // v0.802
      if (!err) err = [ff_screen get]; // other DCU
      if (!err) {                      // is out of the way?
        sprintf(buf,"ff_screen.ret_limit= %d (%d)",
                ff_screen.ret_limit.status,ff_screen.ins_limit.status);
        [logger append:buf];
        if (use_interlocks) {
          if ((ff_screen.ret_limit.status != 1) || (ff_screen.ins_limit.status != 0)) {
            err = E_DCU_ret; 
          }
        }
      }
    }
    if (!err) { if (![self checkAir:TRUE]) { err = E_DCU_air; } }
    if (!err) err = [ff_enable set:0];
    if (!err) err = [mc_enable set:1];
    [NSThread sleepForTimeInterval:PLC_DELAY];
    if (!err) err = [plc getOutput:selector_addr Value:&value];
    if (!err) { if (value != DCU_MCAL) err = E_DCU_selector; }
    [plc close];
  } 
  if (err) {
    error_state = err;
    [self updateMcal]; if (err == E_DCU_ret) [self updateFFS];
    sprintf(buf,"enabling MCAL-DCU failed (err=%d)\r%s",err,ErrorText(err));
    [Notice openNotice:buf Title:P_TITLE Time:0 Level:2];
    return;
  }

  if (!script) [sender setEnabled:FALSE];

  if (sender == popup_mcbrake) {       // move MCAL brake
    req = [sender indexOfSelectedItem];
    sprintf(buf,"%s: brake= %d",PREFUN,req); [logger append:buf];
    NSNumber *pos = [NSNumber numberWithInt:req];
    [mc_screen.brake performSelectorInBackground:@selector(run:) withObject:pos];
  } else {                             // move MCAL screen
    switch ([drop_mcal indexOfSelectedItem]) {
      case 1: req = SCREEN_RET; break;
      case 2: req = SCREEN_INS; break;
      default: assert(0); break;
    }
    if (!eng_mode) {
      if (req == mc_screen.status) {   // v0.900
        [self updateMcal];
        sprintf(buf,"Mcal already %s",(req==SCREEN_RET) ? "retracted" : "deployed");
        [Notice openNotice:buf Title:P_TITLE Time:30 Level:1];
        return;
      }
    }
    sprintf(buf,"%s: screen= %d",PREFUN,req); [logger append:buf];
    NSNumber *pos = [NSNumber numberWithInt:req];
    [mc_screen performSelectorInBackground:@selector(run:) withObject:pos];
  }
}

/* ---------------------------------------------------------------- */

- (void)updateState:(BOOL)toTCS
{
  for (int i=0; i<N_LAMPS; i++) [self updateLamp:i];
  
  if (toTCS) [self updateTCS];
  
  [self updateVar];
  [self updateFFS];
  [self updateMcal];
}

/* ---------------------------------------------------------------- */

- (void)update_lamp:(NSNumber*)lamp    // v0.810
{
  assert([NSThread isMainThread]);
  [self updateLamp:[lamp intValue]];
}

- (void)updateLamp:(int)lamp
{
  NSButton *button = [but_lamps objectAtIndex:lamp];
  BOOL enable=configured;
#if (DEBUG > 1)
  fprintf(stderr,"%s(%d)\n",PREFUN,lamp);
#endif

  if (![NSThread isMainThread]) {      // v0.810
    [self performSelectorOnMainThread:@selector(update_lamp:) 
          withObject:[NSNumber numberWithInt:lamp] waitUntilDone:NO];
    return;
  }
  assert([NSThread isMainThread]);
  
  if (!lamps[lamp]) { [button setEnabled:NO]; return; }

  if (lamps[lamp].moving) {
    [button setImage:((lamps[lamp].request) ? icon_on : icon_off)];
  } else {
    if (lamps[lamp].request == lamps[lamp].status) {
      [button setImage:((lamps[lamp].status) ? icon_on : icon_off)];
    } else {
      [button setImage:icon_err];
    }
    if (ff_screen.status == SCREEN_INS) {
      [button setEnabled:enable];
    } else {
      if (!eng_mode) {                 // v0.811
        [button setEnabled:((lamps[lamp].status != LAMP_OFF) ? enable : NO)];
      } else {
        [button setEnabled:enable];
      }
    }
  }
}

/* ---------------------------------------------------------------- */

- (void)updateVar
{
  BOOL enable=configured;
#if (DEBUG > 1)
  fprintf(stderr,"%s()\n",PREFUN);
#endif

  if (![NSThread isMainThread]) {      // v0.810
    [self performSelectorOnMainThread:_cmd withObject:nil waitUntilDone:NO];
    return;
  } 
  assert([NSThread isMainThread]);

  if (!q_lamp) { [popup_quartz setEnabled:NO]; return; }  

  if (!q_lamp.moving) {
    if (ff_screen.status == SCREEN_INS) {
      [popup_quartz setEnabled:enable];
    } else {
      if (!eng_mode) {
        [popup_quartz setEnabled:((q_lamp.status != 0) ? enable : NO)];
      } else {
        [popup_quartz setEnabled:enable];
      }
    }
  }
  NSString *str = [NSString stringWithFormat:@"%.3f",level2volt(q_lamp.level)];
  [edit_quartz setStringValue:str];
}

/* ---------------------------------------------------------------- */

- (void)updateTCS
{
  int  i;
  char cmd[256]="fflamps",state[32];   // v0.812
#if (DEBUG > 1)
  fprintf(stderr,"%s(): up_and_running=%d\n",PREFUN,tcs.up_and_running);
#endif
   
  for (i=0; i<N_LAMPS; i++) {          // fixed level lamps
    if (!lamps[i]) continue;
    if (lamps[i].status) {             // lamp ON
      NSButton *button = [but_lamps objectAtIndex:i];
      str2chr([button title],state,sizeof(state));
      strcat(cmd," "); strcat(cmd,state);
    } 
  }
  if (q_lamp) {
    if (q_lamp.level) {
      sprintf(state," vQ%.2f",level2volt(q_lamp.level));  // TODO
      strcat(cmd,state);
    }
  }
  if (strlen(cmd) < 8) strcat(cmd," Off"); // v0.812
#if (DEBUG > 0)
  fprintf(stderr,"%s\n",cmd);
#endif
  if (tcs.up_and_running) { int err;
    err = [tcs command:cmd];
    if (err) { char buf[128];
      sprintf(buf,"sending lamp state to TCSIS failed (err=%d)",err);
      [Notice openNotice:buf Title:p_title Time:30 Level:1];
    }
#if (DEBUG > 1)
    if (!err) { char buf[128];
      err = [tcs request:"fflamps" Response:buf Max:sizeof(buf)];
      if (err) printf("err=%d\n",err);
      else     printf("response='%s'\n",buf);
    }
#endif
  }
}

/* ---------------------------------------------------------------- */

- (void)updateScreen:(FFscreen*)screen Popup:(NSPopUpButton*)drop
{
  NSButton *but;

  if (!screen) return;                 // 'Baade' has no MCal unit
#if (DEBUG >  1)
  fprintf(stderr,"%s(%p,%p)\n",PREFUN,screen,drop);
#endif
  assert([NSThread isMainThread]);

  if (screen.moving) {                 // screen is moving
    [drop setTitle:@"moving"];
    [drop setEnabled:NO];
  } else { NSString *status;           // not moving
    switch (screen.status) { 
    case SCREEN_RET:
      status = @"retracted";
      break;
    case SCREEN_INS:
      status = @"deployed";
      break;
    case 0:
      status = @"floating";
      break;    
    default:
      status = @"Error";
      break;
    }
    if (screen.status == screen.request) { // consistent state
      [drop setTitle:status];
    } else {
      NSString *string = [NSString stringWithFormat:@"?%@",status];
      [drop setTitle:string];
    }
  }

  but = (screen == ff_screen) ? check_ffins : check_mcins;
  [but setState:((screen.ins_limit.status) ? NSOnState : NSOffState)];
  but = (screen == ff_screen) ? check_ffret : check_mcret;
  [but setState:((screen.ret_limit.status) ? NSOnState : NSOffState)];
}

/* ---------------------------------------------------------------- */

- (void)updateFFS
{
  BOOL enable=(configured && !ff_disabled);

#if (DEBUG > 1)
  fprintf(stderr,"%s()\n",PREFUN);
#endif
  if (![NSThread isMainThread]) {      // v0.810
    [self performSelectorOnMainThread:_cmd withObject:nil waitUntilDone:NO];
    return;
  } 
  assert([NSThread isMainThread]);

  [self updateScreen:ff_screen Popup:drop_ffs];
  
  if (!ff_screen.brake.moving) [popup_ffbrake selectItemAtIndex:ff_screen.brake.status];
  [check_ffbrake setState:((ff_screen.brake.pswitch.status) ? NSOnState : NSOffState)];

  if (!ff_screen.moving && !ff_screen.brake.moving) { BOOL allow=!ff_disabled;
    if (use_interlocks && mc_screen && allow) {
      allow = (mc_screen.status == SCREEN_RET) && (!mc_screen.moving);
      if (mc_screen.brake.status != BRAKE_ON) allow = NO; // v0.807
    }
    [drop_ffs setEnabled:(allow && enable)];
    allow = (eng_mode && !ff_disabled) ? YES : NO;     // v0.809
    [popup_ffbrake setEnabled:(enable && allow)];
  } else {                             // disable brake and screen
    [popup_ffbrake setEnabled:NO];
    [drop_ffs setEnabled:NO];
  }
}

/* ---------------------------------------------------------------- */

- (void)updateMcal
{
  BOOL enable=(configured && !mc_disabled);

#if (DEBUG > 1)
  fprintf(stderr,"%s()\n",PREFUN);
#endif

  if (![NSThread isMainThread]) {
    [self performSelectorOnMainThread:_cmd withObject:nil waitUntilDone:NO];
    return;
  } 
  assert([NSThread isMainThread]);

  if (mc_screen) {
    [self updateScreen:mc_screen Popup:drop_mcal];
  } else {
    enable = FALSE;
  }

  if (!mc_screen.brake.moving) [popup_mcbrake selectItemAtIndex:mc_screen.brake.status];
  [check_mcbrake setState:((mc_screen.brake.pswitch.status) ? NSOnState : NSOffState)];

  if (!mc_screen.moving && !mc_screen.brake.moving) { BOOL allow=!mc_disabled;
    if (use_interlocks && allow) {
      allow = (ff_screen.status == SCREEN_RET) && (!ff_screen.moving);
      if (ff_screen.brake.status != BRAKE_ON) allow = NO; // v0.807
    }
    [drop_mcal setEnabled:(allow && enable)];
    allow = (eng_mode && !mc_disabled) ? YES : NO;
    [popup_mcbrake setEnabled:(enable && allow)];
  } else {                             // disable brake and screen
    [popup_mcbrake setEnabled:NO];
    [drop_mcal setEnabled:NO];
  }
}

/* ---------------------------------------------------------------- */

- (void)run_dog                        // v0.811
{
  u_int ckie=cookie;                   // Simulator preset
  int   err;
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,self);
#endif

  NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];

  err = [plc open];
  if (!err) { 
    err = [plc get32bit:cookie_addr Value:&ckie];
    if (!err) {
      if (ckie != cookie) { char buf[128];
        sprintf(buf,"Warning: DCU-PLC cookie has changed (%x)",ckie);
        strcat(buf,"\rAnother GUI might be accessing the DCU-PLC");
        [Notice openNotice:buf Title:p_title Time:interval_dog];
        err = [plc set32bit:cookie_addr Value:cookie]; 
        if (!err) {
          [self getFullStatus];
          [self updateState:YES];
        }
      }
    }
    if (!err) [self checkAir:FALSE];   // as part of 'watchdog' v0.901
    [plc close];
  }
  
  if (err) { char buf[128];
    sprintf(buf,"getting/setting cookie at DCU-PLC failed (err=%d)",err);
    [Notice openNotice:buf Title:p_title Time:interval_dog];
  }

  [pool drain];
}

/* ---------------------------------------------------------------- */

- (void)updateAir                      // v0.810
{
  assert([NSThread isMainThread]);

  if (air_pressure < 0) { 
    [edit_air setStringValue:@"?"];
  } else {
    if (air_pressure < min_pressure+1) [edit_air setTextColor:[NSColor redColor]];
    else                               [edit_air setTextColor:[NSColor blackColor]];
    [edit_air setStringValue:[NSString stringWithFormat:@"%.1f",air_pressure]];
  }
}

- (BOOL)checkAir:(BOOL)interactive
{
  int  err,value;
  BOOL r=FALSE; 
  char buf[128];

  assert(air_sensor);
  value = (int)(10.0*air_pressure);    // v0.813 [0.1 PSI]
  if (air_pressure <= min_pressure) value += (lrand48()%5)-1; // Simulator
  else                              value += (lrand48()%7)-3;
  if (value < 0) value = 0;
  err = [air_sensor get:value];
  if (!err) {
    air_pressure = (float)air_sensor.status/10.0f;  // v0.813
    if (air_pressure < min_pressure) {  // TODO test @ LCO
      if (interactive) { char buf[128];
        sprintf(buf,"air-pressure too low: %.1f [psi]",air_pressure);
        [Notice openNotice:buf Title:P_TITLE Time:30 Level:1];
      }  
    } else r = TRUE;
    sprintf(buf,"air_pressure= %.1f",air_pressure); [logger append:buf];
  } else { 
    if (interactive) {
      sprintf(buf,"reading air-pressure failed (err=%d)",err);
      [logger message:buf Level:2 Time:30 File:YES];
    }
    air_pressure = -1.0f;
  }

  [self performSelectorOnMainThread:@selector(updateAir) 
        withObject:nil waitUntilDone:NO];

  return r;
}

/* ---------------------------------------------------------------- */

static char *reasons[] = {             // TODO test
  "ok",
  "Mcal not retracted while using FFS",
  "FFS not retracted while using Mcal",
  ""                                   // terminator
}; 

- (int)checkPLC:(BOOL)interactive      // v0.807
{
  int err,error=0;
  
#ifdef SIM_ONLY
  if (!plc.online) error = (error_state) ? lrand48() & 0x03 : 0;
#endif
  err = [plc getValue:error_addr Value:&error]; 
  if (!err) { if (error) err = E_DCU_disabled; }

  if (error) { char text[512],*p;
    sprintf(text,"PLC disabled (code=0x%02x)",error);
    for (int i=0; *reasons[i]; i++) {
      if (error & (0x01 << i)) {
        strcat(text,"\r"); strcat(text,reasons[i]); 
      }
    }
    if (interactive) {
      [Notice openNotice:text Title:P_TITLE Time:0 Level:2];
    }                                  // replace <CR>
    p = text; while (*p) { if (*p =='\r') *p = ' '; p++; }
    [logger append:text];
  }
  
  return err;
}

/* ---------------------------------------------------------------- */
/* KVO */
/* ---------------------------------------------------------------- */

- (void)observeValue:(NSArray*)param   // v0.810
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p)\n",PREFUN,param);
#endif
  assert([NSThread isMainThread]);

  NSString *key = [param objectAtIndex:0];
  id object = [param objectAtIndex:1];

  if (object == ff_screen) {           // FF screen
    if ([key compare:KEY_MOVING] == NSOrderedSame) {
      if (ff_screen.moving) {
        error_state = 0;               // set global error v0.807
        [prog_screen turnOn];
      } else {
        error_state = (ff_screen.request == ff_screen.status) ? 0 : E_DCU_failed;
        [prog_screen turnOff];
        if (error_state) [self checkPLC:YES]; // v0.807
      }
      [self updateState:NO];
    } else {
      assert(0);
    }
  } else 
  if (object == ff_screen.brake) {     // FFS brake
    if ([key compare:KEY_MOVING] == NSOrderedSame) {
      if (ff_screen.brake.moving) [prog_screen turnOn];
      else                        [prog_screen turnOff];
      [self updateFFS]; if (use_interlocks) [self updateMcal];
    }
  } else
  if (object == ff_screen.brake.pswitch) { // FFS brake pressure-switch
    if ([key compare:KEY_STATUS] == NSOrderedSame) {
      [self updateFFS];
    } else {
      assert(0);
    }    
  } else
  if (object == mc_screen) {           // Mcal screen
    if ([key compare:KEY_MOVING] == NSOrderedSame) {
      if (mc_screen.moving) {
        error_state = 0;               // set global error v0.807
        [prog_mcal turnOn];
      } else {
        error_state = (mc_screen.request == mc_screen.status) ? 0 : E_DCU_failed;
        [prog_mcal turnOff];
        if (error_state) [self checkPLC:YES]; // v0.807
      }
      [self updateState:NO];
    } else {
      assert(0);
    }
  } else
  if (object == mc_screen.brake) {     // Mcal DCU brake
    if ([key compare:KEY_MOVING] == NSOrderedSame) {
      if (mc_screen.brake.moving) [prog_mcal turnOn];
      else                        [prog_mcal turnOff]; 
      [self updateMcal]; if (use_interlocks) [self updateFFS];
    }
  } else 
  if (object == mc_screen.brake.pswitch) { // Mcal DCU brake sensor
    if ([key compare:KEY_STATUS] == NSOrderedSame) {
      [self updateMcal]; 
    } else {
      assert(0);
    }    
  } else
  if (object == q_lamp) {              // variable quartz lamp
    if ([key compare:KEY_MOVING] == NSOrderedSame) {
      if (q_lamp.moving) {
        [prog_screen turnOn];
      } else {
        [prog_screen turnOff];
        [self updateTCS];
      }
      [self updateVar];      
    } else {
      assert(0);
    }
  } else { int i;
    for (i=0; i<N_LAMPS; i++) {        // fixed level lamps
      if (object == lamps[i]) {
        if ([key compare:KEY_MOVING] == NSOrderedSame) {
          if (lamps[i].moving) {       // move started
            [prog_screen turnOn];
          } else {                     // move ended
            [prog_screen turnOff];
            [self updateTCS];
          }
          [self updateLamp:i];
        } else {
          assert(0);
        }
        break;
      }
    }
#ifndef NDEBUG
    if (i == N_LAMPS) { assert(0); }
#endif
  }
}

- (void)observeValueForKeyPath:(NSString*)key
                      ofObject:(id)object
                        change:(NSDictionary*)dict
                       context:(void*)param
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p,%p,%p,%p)\n",PREFUN,key,object,dict,param);
#endif
#if (DEBUG > 2)
  fprintf(stderr,"%s(): mainThread=%d\n",PREFUN,[NSThread isMainThread]);
#endif

  NSString *key_copy = [key copy];
  NSArray *ary = [[NSArray alloc] initWithObjects:key_copy,object,nil];
  [key_copy release];
    
  [self performSelectorOnMainThread:@selector(observeValue:)
        withObject:ary waitUntilDone:YES]; 
  [ary release];  
}

/* ---------------------------------------------------------------- */
/* NSWindow delegates */
/* ---------------------------------------------------------------- */

#if 0 // unused
// - (void)windowDidExpose:(NSNotification*)note
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p): window=%p\n",PREFUN,note,[note object]);
#endif
}
#endif

#if 0 // unused
// - (void)windowDidBecomeKey:(NSNotification*)note
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p): window=%p\n",PREFUN,note,[note object]);
#endif
}
#endif

#if 0 // unused
// - (void)windowDidBecomeMain:(NSNotification*)note
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p): window=%p\n",PREFUN,note,[note object]);
#endif
}
#endif

- (BOOL)windowShouldClose:(id)sender   // NEW v0.903
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,sender);
#endif

#ifdef NDEBUG
  if ([self deployed]) {
    if (!alertQuestion(P_TITLE,"DCU deployed - Quit anyway?",1,"No","Yes")) return NO;
  }
#endif

  return YES;
}

/* ---------------------------------------------------------------- */

@end

/* ---------------------------------------------------------------- */

static NSString* chr2str(const char *text)
{
  assert(text);
  NSString *string = [[NSString alloc] initWithCString:text 
                                       encoding:NSASCIIStringEncoding];
                                       
  return [string autorelease];
}

/* ---------------------------------------------------------------- */

static char* str2chr(NSString* string,char* buf,int maxlen)
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p,%p,%d)\n",PREFUN,string,buf,maxlen);
#endif
  
  assert(string);
  if (![string getCString:buf maxLength:maxlen encoding:NSASCIIStringEncoding]) {
    if (![string getCString:buf maxLength:maxlen encoding:NSUTF8StringEncoding]) {
      assert(0);
    }
  }
  return buf;
}

/* ---------------------------------------------------------------- */

static int alertPanel(const char* title,const char* message,int level,
               const char* but1,const char* but2,const char* but3)
{ // Warning: this blocks the main-thread (application modal)
  int  i;

  NSAlert *alert = [[NSAlert alloc] init];
  
  if (level) [alert setAlertStyle:NSCriticalAlertStyle];
  [alert setMessageText:chr2str(title)];
  [alert setInformativeText:chr2str(message)];
  if (but1) [alert addButtonWithTitle:chr2str(but1)];
  if (but2) [alert addButtonWithTitle:chr2str(but2)];
  if (but3) [alert addButtonWithTitle:chr2str(but3)];
  i = [alert runModal];
  [alert release];
                                      
  return i;
}

static BOOL alertQuestion(const char* title,const char* message,int level,
                   const char* okbut,const char* ccbut)
{
  int i;
  
  i = alertPanel(title,message,level,okbut,ccbut,NULL);
  return ((i != NSAlertFirstButtonReturn) ? YES : NO);
}

/* ---------------------------------------------------------------- */

static float level2volt(int level)
{
  return (float)((24.0*level)/5000.0);
}

/* ---------------------------------------------------------------- */

const char* ErrorText(int err)
{
  static char text[128];               // Warning: not MT-safe
  
  if ((err >= E_DCU_error) && (err < E_DCU_last)) {
    return dcu_errors[err-E_DCU_error];
  } else 
  if ((err >= E_tcpip_error) && (err < E_tcpip_last)) {
    return tcpip_errors[err-E_tcpip_error];
  } else 
  if ((err >= E_plcio_error) && (err < E_plcio_last)) {
    return plcio_errors[err-E_plcio_error];
  } else 
  if ((err >= E_FF_error) && (err < E_FF_last)) {
    return ffmodel_errors[err-E_FF_error];
  }  
  if ((err >= E_modbus_error) && (err < E_modbus_last)) {
    return modio_errors[err-E_modbus_error];
  }  
  sprintf(text,"%s(): unknown error %d",PREFUN,err);
  
  return text;
}

/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
