//
// Logger.m
//
// Created by Christoph Birk on 2013-04-10
// Copyright 2013 Carnegie Observatories. All rights reserved.
//

/* ---------------------------------------------------------------- */

#define DEBUG      0                   // debug level

#define N_MSG      128
#define M_FILE     512

/* ---------------------------------------------------------------- */

#import "Logger.h"
#import "Notice.h"

/* ---------------------------------------------------------------- */

static NSDateFormatter *timeFormatter=nil;

static NSString* chr2str(const char *text);
static char*     str2chr(NSString* string,char* buf,int maxlen);

/* ---------------------------------------------------------------- */

@implementation Logger

@synthesize logfile;

/* ---------------------------------------------------------------- */

+ (void)initialize
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,self);
#endif

  if (self == [Logger class]) {        // avoid call by sub-class
    timeFormatter = [[NSDateFormatter alloc] initWithDateFormat:@"%H:%M:%S"
                                             allowNaturalLanguage:NO];
  }
}

/* ---------------------------------------------------------------- */

- (NSString*)logfileName
{
  time_t    ut; 
  struct tm *now,res;
  char      tstr[32];
  NSArray   *ary;
  NSString  *path,*name;

  ut = time(NULL) - 9*3600;            // 9am UT
#if 1 // def NDEBUG
  now = localtime_r(&ut,&res);
  strftime(tstr,sizeof(tstr),"%Y_%W",now);
#else
  ut = 300*(ut/300);
  now = localtime_r(&ut,&res);
  strftime(tstr,sizeof(tstr),"%Y_%H%M",now);
#endif  
  ary = NSSearchPathForDirectoriesInDomains(NSLibraryDirectory,NSUserDomainMask,YES);
  if ([ary count] > 0) {
    path = [NSString stringWithFormat:@"%@/Logs/%@",[ary objectAtIndex:0],p_exec];
  } else {
    path = NSHomeDirectory();
  }
  name = [NSString stringWithFormat:@"%@/%@%@.log",path,baseName,chr2str(tstr)];
  
  return name;
}

/* ---------------------------------------------------------------- */

- (void)setLogfile:(NSString*)file
{
  char buf[M_FILE],build[64];
   
  NSString *path = [file stringByDeletingLastPathComponent];
  NSFileManager *fm = [[NSFileManager alloc] init]; // MT-safe
  [fm createDirectoryAtPath:path  withIntermediateDirectories:YES
      attributes:nil error:NULL];      // create path if necessary
  [fm release];

  [fileLock lock];
  if (logfile) [logfile release];
  logfile = [file retain];
  [fileLock unlock];
  
  [self message:str2chr(file,buf,sizeof(buf)) File:YES];
  sprintf(buf,"build= %s",str2chr(p_build,build,sizeof(build)));
  [self append:buf];
}

/* ---------------------------------------------------------------- */

- (Logger*)initWithBase:(const char*)base
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%s)\n",PREFUN,base);
#endif

  // Note: NSObject returns _always_ self 
  assert([self superclass] == [NSObject class]);
  // if ((self = [super init]) == nil) return nil;  
  
  NSBundle *bundle = [NSBundle mainBundle];
  p_build  = [[bundle objectForInfoDictionaryKey:@"CFBundleVersion"] retain];
  p_exec   = [[bundle objectForInfoDictionaryKey:@"CFBundleExecutable"] retain];
  if (base) baseName = chr2str(base);
  else      baseName = [p_exec lowercaseString];
  [baseName retain];
  // NSLog(@"%@,%@,%@",p_build,p_exec,baseName);

  messageLock = [[NSLock alloc] init];
  fileLock    = [[NSLock alloc] init];
#ifdef INCLUDE_NOTICE_H
  noticeLock  = [[NSLock alloc] init];
#endif  
  mesgArray = [[NSMutableArray alloc] initWithCapacity:N_MSG];
  timeArray = [[NSMutableArray alloc] initWithCapacity:N_MSG];
  assert(mesg_head == 0);

  [self setLogfile:[self logfileName]];
           
  return self;
}

/* ---------------------------------------------------------------- */

- (void)setField:(NSTextField*)field
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,field);
#endif
  [messageLock lock];
  messageField = field;
  [messageLock unlock];
}

/* ---------------------------------------------------------------- */

- (void)setTable:(NSTableView*)table
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,table);
#endif
  
  [messageLock lock];
  messageTable = table;
  [table setDataSource:self];
  [messageLock unlock];    
}

/* ---------------------------------------------------------------- */

- (BOOL)changeLogfile
{
  BOOL r=FALSE;
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p)\n",PREFUN,self);
#endif

  NSString *string = [self logfileName];
  if ([logfile compare:string] != NSOrderedSame) { // name changed
    [self setLogfile:string];                      // set new name
    r = TRUE;
  }

  return r;
}

/* ---------------------------------------------------------------- */

static void replaceElement(NSMutableArray* array,NSString* string,int head)
{
  if (head < [array count]) {     // NSMutableArary protected by 'messageLock'
    [array replaceObjectAtIndex:head withObject:string];
  } else {
    [array addObject:string];
  }
}

/* ---------------------------------------------------------------- */

- (void)addMessage:(const char*)text
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(): '%s'\n",PREFUN,text);
#endif
  assert(messageLock);

  if (messageField) {                  // overwrite text field
    [messageLock lock];
    if ([NSThread isMainThread]) {
      [messageField setStringValue:chr2str(text)];
    } else {
      [messageField performSelectorOnMainThread:@selector(setStringValue:)
                    withObject:chr2str(text) waitUntilDone:YES];
    }
    [messageLock unlock];
  }
  if (messageTable) {
    assert(timeFormatter);
    [messageLock lock];                // add to table
    replaceElement(timeArray,[timeFormatter stringFromDate:[NSDate date]],mesg_head);
    assert([timeArray count] <= N_MSG);
    replaceElement(mesgArray,chr2str(text),mesg_head);
    assert([mesgArray count] <= N_MSG);
    mesg_head = (mesg_head+1) % N_MSG;
    [messageLock unlock];
    [messageTable reloadData];
  }
}

/* ---------------------------------------------------------------- */

- (void)message:(const char*)text Level:(int)level Time:(int)dt File:(BOOL)file
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%s)\n",PREFUN,text);
#endif

  if (file) [self append:text];

  if (level >= 0) {                    // write to message window
    [self addMessage:text];
#ifdef INCLUDE_NOTICE_H
    assert(noticeLock);
    [noticeLock lock];
    if (level > 0) {                   // open notice?
      NSString *string = chr2str(text);
      if (!last_text || ([last_text compare:string] != NSOrderedSame) || (dt > 0)) { char buf[64];
        [Notice openNotice:text Title:str2chr(p_exec,buf,sizeof(buf)) Time:dt Level:level];
        if (last_text) [last_text release];
        last_text = [string retain];
      }
    }
    [noticeLock unlock];
#endif
  }
}

- (void)message:(const char*)text Level:(int)level Time:(int)dt
{
  [self message:text Level:level Time:dt File:NO];
}

- (void)message:(const char*)text Level:(int)level File:(BOOL)file
{
  [self message:text Level:level Time:0 File:file];
}

- (void)message:(const char*)text File:(BOOL)file
{
  [self message:text Level:0 Time:0 File:file];
}

- (void)message:(const char*)text Level:(int)level
{
  [self message:text Level:level Time:0 File:NO];
}

- (void)message:(const char*)text
{
  [self message:text Level:0 Time:0 File:NO];
}

/* ---------------------------------------------------------------- */

- (void)append:(const char*)text 
{
  char file[M_FILE];
  FILE *fp;
#if (DEBUG > 1)
  fprintf(stderr,"%s(%s)\n",PREFUN,text);
#endif

  [fileLock lock];
  fp = fopen(str2chr(logfile,file,sizeof(file)),"a"); // append to logfile
  if (fp == NULL) {
    fprintf(stderr,"%s(): cannot open file '%s'\n",PREFUN,file);
  } else { char buf[32]; struct tm *now,res; time_t ut;
    ut = time(NULL);  
    now = localtime_r(&ut,&res);
    strftime(buf,sizeof(buf),"%b %d %H:%M:%S",now);
    fprintf(fp,"%s %s\n",buf,text);
    fclose(fp);
  }
  [fileLock unlock];
}


/* ---------------------------------------------------------------- */
/* NSTableView dataSource */
/* ---------------------------------------------------------------- */

- (NSInteger)numberOfRowsInTableView:(NSTableView*)table
{
#if (DEBUG > 2)
  fprintf(stderr,"%s(%p): msg_n=%d\n",PREFUN,table,msg_n);
#endif
  return [mesgArray count];
}

/* ---------------------------------------------------------------- */

- (id)tableView:(NSTableView*)table
      objectValueForTableColumn:(NSTableColumn*)column
      row:(NSInteger)row
{
  int      i;
  NSString *string=nil;

#if (DEBUG > 1)
  fprintf(stderr,"%s(%p,%p,%d)\n",PREFUN,table,column,row);
#endif
  assert(row < [mesgArray count]);
  if (row >= [mesgArray count]) return nil;
  
#if (DEBUG > 0) && !defined(NDEBUG) && defined(INCLUDE_NOTICE_H)
  if (![NSThread isMainThread]) { char buf[128];
    sprintf(buf,"isMainThread=%d, row=%ld\n",[NSThread isMainThread],(long)row);
    [Notice openNotice:buf Title:PREFUN Time:600 Level:1];
  }
#endif

  [messageLock lock];
  i = mesg_head-row-1; if (i < 0) i += N_MSG;
  switch ([[column identifier] intValue]) {
  case 0:                              // time
    string = [timeArray objectAtIndex:i];
    break;
  case 1:                              // text
    string = [mesgArray objectAtIndex:i];
    break;
  default:
    assert(0);
  }
  [messageLock unlock];

  return string;
}

/* ---------------------------------------------------------------- */

@end

/* ---------------------------------------------------------------- */

static NSString* chr2str(const char *text)
{
  assert(text);
  NSString *string = [[NSString alloc] initWithCString:text 
                                       encoding:NSASCIIStringEncoding];
                                       
  return [string autorelease];
}

/* ---------------------------------------------------------------- */

static char* str2chr(NSString* string,char* buf,int maxlen)
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p,%p,%d)\n",PREFUN,string,buf,maxlen);
#endif
  
  assert(string);
  if (![string getCString:buf maxLength:maxlen encoding:NSASCIIStringEncoding]) {
    if (![string getCString:buf maxLength:maxlen encoding:NSUTF8StringEncoding]) {
      assert(0);
    }
  }
  return buf;
}

/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
