//
// Notice.m
//
// Created by Christoph Birk on 2012-10-18
// Copyright 2012 Carnegie Observatories. All rights reserved.
//
// Icons: /System/Library/CoreServices/CoreTypes.bundle/Contents/Resources/
//
// 2013-09-10: use GCD
//

/* ---------------------------------------------------------------- */

#define DEBUG           0              // debug level

/* ---------------------------------------------------------------- */

#import  "Notice.h"

/* ---------------------------------------------------------------- */

static NSLock         *mutex=nil;      // class variables
static NSMutableArray *n_array=nil;
static NSString       *p_exec=nil;
static Notice         *g_notice=nil;
static       float    pos_x,pos_y;
static const float    def_w=380,def_h=100;

/* ---------------------------------------------------------------- */

static int       imax(int,int);
static NSArray*  arrayFromString(const char* string,const char* sep);
static NSString* chr2str(const char *text);
static char*     str2chr(NSString* string,char* buf,int maxlen);

/* ---------------------------------------------------------------- */

@interface Notice ()                   // private class extension
@property (assign)             time_t auto_close;
@property (readonly)           BOOL   closed;
@property (assign,nonatomic)   id     target;
@property (assign,nonatomic)   SEL    selector;
@end

@implementation Notice

@synthesize auto_close;
@synthesize closed;
@synthesize target,selector;

/* ---------------------------------------------------------------- */

+ (void)initialize
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p): mainThread=%d\n",PREFUN,self,[NSThread isMainThread]);
#endif

  if (self != [Notice class]) return;  // avoid call by sub-class
  
  p_exec = [[[NSBundle mainBundle] objectForInfoDictionaryKey:@"CFBundleExecutable"] 
                                   retain];

  mutex = [[NSLock alloc] init];
  n_array = [[NSMutableArray alloc] initWithCapacity:2];

  NSRect rect = [[NSScreen mainScreen] visibleFrame];
  pos_x = (rect.size.width-def_w)/2;
  pos_y = (2*(rect.size.height+def_h))/3;

  assert([NSThread isMainThread]);     // Note: timer on main thread
  [NSTimer scheduledTimerWithTimeInterval:5.0
           target:self selector:@selector(fireTimer:)
           userInfo:nil repeats:YES];
}

/* ---------------------------------------------------------------- */

+ (void)fireTimer:(NSTimer*)theTimer
{
  int    i;
  Notice *note;
#if (DEBUG > 0)
  if ([n_array count]) {
    fprintf(stderr,"%s(%p): n_array.count=%u\n",PREFUN,theTimer,[n_array count]);
  }
#endif
  assert([NSThread isMainThread]);
  
  if (![mutex lockBeforeDate:[NSDate dateWithTimeIntervalSinceNow:0.25]]) return;
  
  for (i=0; i<[n_array count]; i++) {    // check list of open notes
    note = [n_array objectAtIndex:i]; 
    assert(note);
    if (note.closed) {                 // closed by GUI
      [n_array removeObjectAtIndex:i];
    } else {                           // auto-close ?
#ifndef NDEBUG
      [note check_me];
#endif
      if (time(NULL) >= [note auto_close]) {
        [[note window] close];
      }
    }
  }
  
  [mutex unlock];
}

/* ---------------------------------------------------------------- */

+ (void)open_notice:(NSArray*)lines Title:(NSString*)title 
               Time:(int)wait Level:(int)level Target:(id)object Selector:(SEL)sel
{
  int         w=def_w,h=def_h,d,i;
  NSTextField *edit;
  NSString    *string,*file=nil;
  NSSize      size;

#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,lines);
#endif
  assert([NSThread isMainThread]);

  g_notice = [[Notice alloc] init];    // Note: assumes [mutex lock]
  
  NSFont *font0 = [NSFont boldSystemFontOfSize:14];
  NSDictionary *attrs0 = [[NSDictionary alloc] 
                           initWithObjectsAndKeys:font0,NSFontAttributeName,nil];
  NSFont *font1 = [NSFont labelFontOfSize:12];
  NSDictionary *attrs1 = [[NSDictionary alloc] 
                           initWithObjectsAndKeys:font1,NSFontAttributeName,nil];

  for (i=0; i<[lines count]; i++) {    // find max. text width
    string = [lines objectAtIndex:i];
    size = [string sizeWithAttributes:((i==0) ? attrs0 : attrs1)];
    if (10+size.width > def_w-120) {
      w = imax(w,120+size.width+10);
    }
    if (i > 0) h += size.height+4;
  }

  NSRect rect = [[NSScreen mainScreen] visibleFrame]; // lower left corner
  if ([n_array count] == 0) {          // reset default position 
    pos_x = (rect.size.width-def_w)/2;
    pos_y = (2*(rect.size.height+def_h))/3;
  }
  if (pos_x+w > rect.size.width) pos_x = rect.origin.x;
  if (pos_y <= h) pos_y = rect.origin.y+(3*rect.size.height)/4;

  NSWindow *window = [[NSWindow alloc] 
                  initWithContentRect:NSMakeRect(pos_x,pos_y-h,w,h)
                  styleMask:NSTitledWindowMask|NSClosableWindowMask
                  backing:NSBackingStoreBuffered
                  defer:NO];
  [window setLevel:NSFloatingWindowLevel];
  [window setTitle:title];
  [window setDelegate:g_notice];
  g_notice.window = window;            // Note: increases reference count

  for (i=0; i<[lines count]; i++) {
    string = [lines objectAtIndex:i];
    size = [string sizeWithAttributes:((i==0) ? attrs0 : attrs1)];
    d = size.height+4;
    edit = [[NSTextField alloc] initWithFrame:NSMakeRect(95,h-40-d*i,w-120,d-1)];
    [edit setFont:((i==0) ? font0 : font1)];
    [edit setSelectable:NO];
    [edit setBezeled:NO];
    [edit setDrawsBackground:NO];
    [edit setStringValue:[lines objectAtIndex:i]];
    [[g_notice.window contentView] addSubview:edit]; // increases ref.count
    [edit release];
  }
  [attrs0 release];
  [attrs1 release];
 
  NSButton *button = [[NSButton alloc] initWithFrame:NSMakeRect(105+w/4,15,60,25)];
  [button setTitle:@"OK"];             // button almost centered
  [button setBezelStyle:NSRoundedBezelStyle];
  [button setButtonType:NSMomentaryPushInButton];
  [button setTarget:g_notice];
  [button setAction:@selector(action_ok:)];
  [[g_notice.window contentView] addSubview:button]; // increases ref.count
  [button release];

  NSImageView *view = [[NSImageView alloc] initWithFrame:NSMakeRect(17,h-90,70,70)];
  [[g_notice.window contentView] addSubview:view]; // increases ref.count
  NSBundle *bundle = [NSBundle mainBundle];
  switch (level) {                     // set proper icon
  case 1:
    file = [bundle pathForResource:@"AlertCautionIcon" ofType:@"icns"];
    break;
  case 2:
    file = [bundle pathForResource:@"AlertStopIcon" ofType:@"icns"];
    break;
  default:
    file = [bundle pathForResource:[p_exec lowercaseString] ofType:@"icns"];
    if (!file) {
      file = [bundle pathForResource:@"AlertNoteIcon" ofType:@"icns"];
    }
    break;
  }
  if (file) {
    NSImage *image  = [[NSImage alloc] initWithContentsOfFile:file];
    [view setImage:image];
    [image release];
  }
  [view release];

  g_notice.auto_close = (wait) ? time(NULL)+wait : 0x7fffffff;

  [window makeKeyAndOrderFront:self];  
  [window release];

  g_notice.target = object;                // install callback
  g_notice.selector = sel;
  
  [n_array addObject:g_notice];
  [g_notice release];
  pos_x += 20; pos_y -= 20;
}

/* ---------------------------------------------------------------- */

+ (Notice*)openNotice:(const char*)text Title:(const char*)title 
              Time:(int)wait Level:(int)level Target:(id)object Selector:(SEL)sel
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p,%p,%d,%d,%p,%p)\n",PREFUN,text,title,wait,level,object,sel);
#endif

  if (!text) return nil;               // allows initialization
  assert(title);
  
  NSArray *lines = arrayFromString(text,"\r\n");
  NSString *tstr = chr2str(title);

  [mutex lock];                        // block forever

  if ([NSThread isMainThread]) {
    [self open_notice:lines Title:tstr Time:wait Level:level 
          Target:object Selector:sel];
  } else {  
    dispatch_sync(dispatch_get_main_queue(),^(void) {
      [self open_notice:lines Title:tstr Time:wait Level:level 
            Target:object Selector:sel];
    });
  }
  Notice *note = g_notice;
  
  [mutex unlock];

  return note;
}

/* ---------------------------------------------------------------- */

+ (Notice*)openNotice:(const char*)text Title:(const char*)title 
                Level:(int)level Target:(id)object Selector:(SEL)sel
{  
  return [self openNotice:text Title:title Time:0 Level:level 
               Target:object Selector:sel];
}

/* ---------------------------------------------------------------- */

+ (Notice*)openNotice:(const char*)text Title:(const char*)title Time:(int)wait 
                Level:(int)level
{ 
  return [self openNotice:text Title:title Time:wait Level:level 
               Target:nil Selector:nil];
}

/* ---------------------------------------------------------------- */

+ (Notice*)openNotice:(const char*)text Title:(const char*)title Time:(int)wait
{
  return [self openNotice:text Title:title Time:wait Level:0];
}

/* ---------------------------------------------------------------- */

+ (Notice*)openNotice:(const char*)text Title:(const char*)title
{
  return [self openNotice:text Title:title Time:0 Level:0];
}

/* ---------------------------------------------------------------- */

- (void)windowWillClose:(NSNotification*)notification
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p)\n",PREFUN,notification);
#endif
  assert([NSThread isMainThread]);
  if (target) { 
    [target performSelector:selector withObject:nil afterDelay:0];
  }
  closed = TRUE;
}

/* ---------------------------------------------------------------- */

- (void)close
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,self);
#endif
  [mutex lock];                        // block forever
  if ([n_array containsObject:self]) {
    [self.window performSelectorOnMainThread:@selector(close)
                 withObject:nil waitUntilDone:YES];
  }
  [mutex unlock];                      // block forever
}

/* ---------------------------------------------------------------- */

#ifndef NDEBUG
- (void)check_me
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p) isVisible=%d\n",PREFUN,self,[[self window] isVisible]);
#endif
}
#endif

/* ---------------------------------------------------------------- */

- (IBAction)action_ok:(id)sender
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p)\n",PREFUN,sender);
#endif
  [[self window] close];
}

/* ---------------------------------------------------------------- */

@end

/* ---------------------------------------------------------------- */

static int imax(int a,int b)
{
  if (a > b) return(a);
  return(b);
}

/* ---------------------------------------------------------------- */

static NSArray* arrayFromString(const char* string,const char* sep)
{
  char *buffer,*b,*p;
  
  assert(string);
  NSMutableArray *ary = [[NSMutableArray alloc] initWithCapacity:4];
  buffer = b = (char*)malloc(1+strlen(string));
  strcpy(buffer,string);
  while ((p=strsep(&b,sep))) {
    if (*p) [ary addObject:chr2str(p)];
  }  
  free((void*)buffer);
  
  return [ary autorelease];
}

/* ---------------------------------------------------------------- */

static NSString* chr2str(const char *text)
{
  assert(text);
  NSString *string = [[NSString alloc] initWithCString:text 
                                       encoding:NSASCIIStringEncoding];
                                       
  return [string autorelease];
}

/* ---------------------------------------------------------------- */

static char* str2chr(NSString* string,char* buf,int maxlen)
{
#if (DEBUG > 2)
  fprintf(stderr,"%s(%p,%p,%d)\n",PREFUN,string,buf,maxlen);
#endif
  
  assert(string);
  if (![string getCString:buf maxLength:maxlen encoding:NSASCIIStringEncoding]) {
    if (![string getCString:buf maxLength:maxlen encoding:NSUTF8StringEncoding]) {
      assert(0);
    }
  }
  return buf;
}

/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
