//
//  plcio.m
//
//  Created by Christoph Birk on 2013-03-20
//  Copyright 2013 Carnegie Observatories. All rights reserved.
//

/* ---------------------------------------------------------------- */

#define IS_PLCIO_M

#define DEBUG           0              // debug level

/* ---------------------------------------------------------------- */

#import "plcio.h"

/* ---------------------------------------------------------------- */

int o2d(int octal) 
{
  int f=1,decimal=0,sign=1;
  
  if (octal < 0) { sign = -1; octal = abs(octal); }

  while (octal > 0) {
    decimal += f * (octal % 10);
    octal   /= 10;
    f *= 8;
  }
  return(sign*decimal);
}

/* ---------------------------------------------------------------- */

#ifndef NDEBUG
static int is_bigendian(void)
{
  union { long l; char c[sizeof(long)]; } u;
  
  u.l = 1;
  return((u.c[sizeof(long)-1] == 1));
}
#endif

/* ---------------------------------------------------------------- */

@implementation PLC_IO

/* ---------------------------------------------------------------- */

- (int)setCoil:(int)channel Status:(int)status
{
  int err=0,n;
#if (DEBUG > 0)
  fprintf(stderr,"%s(%d,%d)\n",PREFUN,channel,status);
#endif

  n = [self send:0x05 Address:(3072+channel) Value:((status) ? 0xff00 : 0x0000)]; 
  if (online) {
    if (n < 1) err = E_plcio_ioerror;
  }

  return err;
}

/* ---------------------------------------------------------------- */

- (int)getCoil:(int)channel Status:(int*)status
{
  int  err=0,n;
  char buf[256];
#if (DEBUG > 0) 
  fprintf(stderr,"%s(%d,%p)\n",PREFUN,channel,status);
#endif

  n = [self send:0x01 Address:(3072+channel) Value:0x0001 Response:buf];
  if (online) {
    if (n == 10) *status = (int)buf[9];
    else         err = E_plcio_ioerror;
  }

  return err;
}

/* ---------------------------------------------------------------- */

- (int)setOutput:(int)channel Value:(int)value   // set output
{
  int err=0,n;
#if (DEBUG > 0)
  fprintf(stderr,"%s(%d,%d)\n",PREFUN,channel,value);
#endif

  n = [self send:0x05 Address:(2048+channel) Value:((value) ? 0xff00 : 0x0000)];
  if (online) {
    if (n < 1) err = E_plcio_ioerror;
  }
  
  return err;
}

/* ---------------------------------------------------------------- */

- (int)getOutput:(int)channel Value:(int*)value  // read-back output
{
  int  err=0,n;
  char buf[256];
#if (DEBUG > 1)
  fprintf(stderr,"%s(%d,%p)\n",PREFUN,channel,value);
#endif

  n = [self send:0x01 Address:(2048+channel) Value:0x0001 Response:buf];
  if (online) {
    if (n == 10) *value = (int)buf[9];
    else         err = E_plcio_ioerror;
  }

  return err;
}

/* ---------------------------------------------------------------- */

- (int)getInput:(int)channel Value:(int*)value
{
  int  err=0,n;
  char buf[256];
#if (DEBUG > 1)
  fprintf(stderr,"%s(%d,%p)\n",PREFUN,channel,value);
#endif

  n = [self send:0x02 Address:(2048+channel) Value:0x0001 Response:buf];
  if (online) {
    if (n == 10) *value = (int)buf[9];
    else         err = E_plcio_ioerror;
  }

  return err;
}

/* ---------------------------------------------------------------- */

- (int)setValue:(int)addr Value:(int)value 
{
  int  err=0,n;
  char buf[32];

#if (DEBUG > 0)
  fprintf(stderr,"%s(%d,%d)\n",PREFUN,addr,value);
#endif

  sprintf(buf,"%04d",value);           /* BCD conversion */
  value =   0x1000 * (int)(buf[0]-'0')
          + 0x0100 * (int)(buf[1]-'0')
          + 0x0010 * (int)(buf[2]-'0')
          + 0x0001 * (int)(buf[3]-'0');
  n = [self send:0x06 Address:addr Value:value];
  if (online) {  
    if (n < 1) err = E_plcio_ioerror;
  }
  
  return err;
}

/* ---------------------------------------------------------------- */

- (int)getValue:(int)addr Value:(int*)value  /* BCD */
{
  int  err=0,n;
  char res[256];

#if (DEBUG > 0)
  fprintf(stderr,"%s(%d,%p)\n",PREFUN,addr,value);
#endif

  n = [self send:0x04 Address:addr Value:0x0001 Response:res];
  if (online) {
    if (n == 11) { char buf[32];
      sprintf(buf,"%2x%02x",(u_char)res[9],(u_char)res[10]);
      *value = atoi(buf);
    } else { 
      err = E_plcio_ioerror;
    }
  }

  return err;
}

/* ---------------------------------------------------------------- */

- (int)set16bit:(int)addr Value:(u_short)value
{
  int err=0,n;

#if (DEBUG > 1)
  fprintf(stderr,"%s(%d,%04x)\n",PREFUN,addr,value);
#endif
  assert(sizeof(u_short) == 2);

  n = [self send:0x06 Address:addr Value:value];
  if (online) {
    if (n < 1) err = E_plcio_ioerror;
  }
  return err;
}

/* ---------------------------------------------------------------- */

- (int)get16bit:(int)addr Value:(u_short*)value
{
  int  err=0,n;
  char res[256];

#if (DEBUG > 1)
  fprintf(stderr,"%s(%d,%p)\n",PREFUN,addr,value);
#endif
  assert(sizeof(u_short) == 2);

  n = [self send:0x04 Address:addr Value:0x0001 Response:res];
  if (online) {
#ifndef NDEBUG
    assert(!is_bigendian());           // must be little endian
#endif
    if (n == 11) { u_char b1,b2;
      b1 = (u_char)res[9];
      b2 = (u_char)res[10];
      *value =  ((u_short)b1 << 8) + (u_short)b2;
    } else {
      err = E_plcio_ioerror;
    }
  }

  return err;
}

/* ---------------------------------------------------------------- */

- (int)set32bit:(int)addr Value:(u_int)value
{
  int     err;
  u_short lo,hi;
#if (DEBUG > 1)
  fprintf(stderr,"%s(%d,%u)\n",PREFUN,addr,value);
#endif

  assert(sizeof(u_int) == 4);
  err = [self open:lock_timeout]; if (err) return err;

  lo = value & 0xffff;
  if (!err) err = [self set16bit:(addr)   Value:lo];
  hi = (value >> 16) & 0xffff;
  if (!err) err = [self set16bit:(addr+1) Value:hi];
  
  [self close];
  
  return err;
}

/* ---------------------------------------------------------------- */

- (int)get32bit:(int)addr Value:(u_int*)value
{
  int     err;
  u_short lo,hi;
#if (DEBUG > 1)
  fprintf(stderr,"%s(%d,%p)\n",PREFUN,addr,value);
#endif

  assert(sizeof(u_int) == 4);
  err = [self open:lock_timeout]; // printf("err=%d\n",err);
  if (err) return err;

  if (!err) err = [self get16bit:(addr)   Value:&lo];
  if (!err) err = [self get16bit:(addr+1) Value:&hi];
  if (!err && online) *value = (u_int)lo + ((u_int)hi << 16);
  
  [self close];
  
  return err;
}

/* ---------------------------------------------------------------- */

@end

/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
