//
// tcpip.m
//
// Created by Christoph Birk on 2013-03-19
// Copyright 2013 Carnegie Observatories. All rights reserved.
//

/* ---------------------------------------------------------------- */

#define IS_TCPIP_M

#define DEBUG           0              /* debug level */

/* ---------------------------------------------------------------- */

#import  "tcpip.h"

#include <netdb.h>                     /* socket/proto/host stuff */
#include <errno.h>

/* ---------------------------------------------------------------- */

static int TCPIP_Wait4Char(int sock,char* c,int timeout); // [ms]

static NSLock *class_lock=nil;

/* ---------------------------------------------------------------- */

@implementation TCPIP_Connection

@synthesize online;
@synthesize up_and_running;

/* ---------------------------------------------------------------- */

+ (void)initialize
{
  if (self != [TCPIP_Connection class]) return; // prevent sub-classes
#if (DEBUG > 0)                        // from executing 'initialize'
  fprintf(stderr,"%s(%p)\n",PREFUN,self);
#endif  
  // global class lock used when opening a connection
  class_lock = [[NSLock alloc] init]; 
}

/* ---------------------------------------------------------------- */

+ (int)createClientSocket:(const char*)hostname Port:(u_short)port
{
  int                sock,err=0;
  struct hostent     *hostp;           // Warning: static (not MT-safe)
  struct protoent    *proto;           // Warning: static (not MT-safe)
  struct sockaddr_in saddr;

#if (DEBUG > 1)
  fprintf(stderr,"%s(%s,%d)\n",PREFUN,hostname,(int)port);
#endif

  [class_lock lock];                   // protects MT-unsafe functions
  
  hostp = gethostbyname(hostname);     // get host IP address (not MT-safe)
  if (hostp == NULL) { 
    err = E_tcpip_hostbyname;
  } else {
    proto = getprotobyname("tcp");     // get protocol by name (not MT-safe)
    if (proto == NULL) {
      err = E_tcpip_protocol;
    } else { int retry;                // v0.900
      memset((void*)&saddr,0,sizeof(saddr));
      saddr.sin_family = AF_INET;
      saddr.sin_port = htons((u_short)port);
      memcpy(&saddr.sin_addr,hostp->h_addr_list[0],hostp->h_length);
      for (retry=3; retry>0; retry--) {
        if ((sock = socket(AF_INET,SOCK_STREAM,proto->p_proto)) < 0) {
          err = E_tcpip_socket; break;
        }
        (void)fcntl(sock,F_SETFD,FD_CLOEXEC);
        if (connect(sock,(struct sockaddr *)&saddr,sizeof(saddr)) < 0) {
          (void)close(sock);
          err = E_tcpip_connect; if (errno != EINTR) break;
          if (retry > 1) [NSThread sleepForTimeInterval:0.3];
        } else { err=0; break; }
      } // endfor(retry)
    } // endif(proto)
  } // endif(host)

  [class_lock unlock];
  
  return (err) ? -err : sock;
}

/* ---------------------------------------------------------------- */

- (id)initWithHost:(const char*)hstr Port:(int)pint Online:(BOOL)on
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%s,%d,%d)\n",PREFUN,hstr,pint,on);
#endif

  if ((self = [super init]) == nil) return nil;

  host = (char*)malloc((1+strlen(hstr))*sizeof(char));
  strcpy(host,hstr);
  port = pint;
  online = on;
  count = 0;
  sock = -1;
  lock_timeout = 5;
  sock_timeout = 5;
  terminator = '\n';                   // default terminator 
  
  lock = [[NSRecursiveLock alloc] init];
  NSString *name = [NSString stringWithFormat:@"tcpip-lock-%s-%d",host,port];
  // NSLog(@"%@",name);
  [lock setName:name];
  
  up_and_running = TRUE;
  if (online) { int err;
    err = [self open]; if (!err) [self close]; // NEW 0.903
  }
  
  return self;
}

/* ---------------------------------------------------------------- */

- (BOOL)mylock:(double)timeout
{
  if (timeout > 0.0) {                 // timeout interval
    NSDate *date = [NSDate dateWithTimeIntervalSinceNow:timeout];
    return [lock lockBeforeDate:date];
  } else
  if (timeout < 0.0) {                 // infinite wait
    [lock lock];
    return YES;
  }
  
  return [lock tryLock];              // poll
}

/* ---------------------------------------------------------------- */

- (int)open
{
  return [self open:lock_timeout];
}

- (int)open:(int)timeout
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%d): count=%d\n",PREFUN,timeout,count);
#endif
  if (!up_and_running) return E_tcpip_error;

  if (![self mylock:timeout]) return E_tcpip_lock;

  lock_timeout = timeout;
   
  if (count == 0) {
    assert(sock < 0);
    if (online) {
      sock = [TCPIP_Connection createClientSocket:host Port:port];
    } else {
      [NSThread sleepForTimeInterval:0.020];
      sock = 0;
    }
  }
  if (sock >= 0) count += 1;
  else           up_and_running = FALSE; // NEW v0.903
#if (DEBUG > 1)
  fprintf(stderr,"%s(): count=%d, sock=%d\n",PREFUN,count,sock);
#endif
  [lock unlock];
  
  return (sock >= 0) ? 0 : E_tcpip_open;
}

/* ---------------------------------------------------------------- */

- (void)close
{
  BOOL r = [self mylock:lock_timeout];

#if (DEBUG > 1)
  fprintf(stderr,"%s(): count=%d\n",PREFUN,count);
#endif 
  assert(count > 0);
  count--;
  if (count == 0) {
    if (online) close(sock);
    sock = -1;
  }
  
  if (r) [lock unlock];
}

/* ---------------------------------------------------------------- */

- (int)request:(const char*)cmd Response:(char*)res Max:(int)len
{
  int  err;
#if (DEBUG > 2)
  fprintf(stderr,"%s(%p,%d)\n",PREFUN,cmd,len);
#endif

  err = [self open:lock_timeout]; if (err) return err;

  if (![self mylock:lock_timeout]) {
    [self close];
    return E_tcpip_lock;
  }

  if (online) {
    assert(sock > 0);
    if (send(sock,cmd,strlen(cmd),0) < 0) err = E_tcpip_send;
    assert(res);
    if (!err) { int i;
      for (i=0; i<len-1; i++) {
        err = TCPIP_Wait4Char(sock,&res[i],1000*sock_timeout);
        if (err) break;
        if (res[i] == terminator) break;
      }
      res[i] = '\0';                   // remove terminator
    }
  } else {
    assert(sock == 0);
    [NSThread sleepForTimeInterval:0.050]; 
  }
  [lock unlock];
  [self close];

  return err;
}

/* ---------------------------------------------------------------- */

- (int)command:(const char*)cmd
{
  char res[512];
  
  return [self request:cmd Response:res Max:sizeof(res)];
}

/* ---------------------------------------------------------------- */

- (void)shutdown
{
  BOOL r;
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,self);
#endif
  r = [class_lock tryLock];            // stop further connections
  up_and_running = FALSE;
  if (r) [class_lock unlock];

  r = [self mylock:((r) ? lock_timeout : 1)];
  while (count > 0) {                  // active connection
    if (r) [lock unlock];
    [self close];
    r = [self mylock:lock_timeout];
  }
  if (r) [lock unlock];  
}

/* ---------------------------------------------------------------- */

- (void)dealloc
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,self);
#endif

  if (host) free((void*)host);
  [lock release];  
  [super dealloc];
}

/* ---------------------------------------------------------------- */

@end

/* ---------------------------------------------------------------- */
/* local functions */
/* ---------------------------------------------------------------- */

static int TCPIP_Wait4Char(int sock,char* c,int timeout) // [ms]
{
  int    r;
  fd_set rfd;
  struct timeval tout;
#if (DEBUG > 1)
  fprintf(stderr,"%s(%d,%p,%d)\n",PREFUN,sock,c,timeout);
#endif

  do {
    FD_ZERO(&rfd); FD_SET(sock,&rfd); 
    tout.tv_sec = timeout/1000; tout.tv_usec = 1000*(timeout%1000);
    r = select(sock+1,&rfd,NULL,NULL,&tout);
    if (FD_ISSET(sock,&rfd)) break;
  } while (r > 0);
  if (r <= 0) return E_tcpip_timeout;
  r = recv(sock,c,1,0);
  if (r <  0) return E_tcpip_recv;
  if (r == 0) return E_tcpip_nodata;

  return 0;
}

/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
