//
// Logger.m
//
// Created by Christoph Birk on 2013-04-10
// Copyright 2013 Carnegie Observatories. All rights reserved.
//
// 2014-03-07  /Users/Shared/Library/Logs
// 2014-04-02  set permissions when using Shared
// 2014-04-23  use Ystr2tmp()
//

/* ---------------------------------------------------------------- */

#ifndef DEBUG
#define DEBUG      0                   // debug level
#endif

#define N_MSG      128
#define M_FILE     512

/* ---------------------------------------------------------------- */

#import "Logger.h"
#import "Notice.h"
#import "Yutils.h"

/* ---------------------------------------------------------------- */

static NSDateFormatter *timeFormatter=nil;

/* ---------------------------------------------------------------- */

@implementation Logger

@synthesize logfile;

/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

+ (void)initialize
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,self);
#endif

  if (self == [Logger class]) {        // avoid call by sub-class
    timeFormatter = [[NSDateFormatter alloc] init];
    [timeFormatter setDateFormat:@"HH:mm:ss"];
  }
}

/* ---------------------------------------------------------------- */

- (Logger*)initWithBase:(const char*)base
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%s)\n",PREFUN,base);
#endif

  return [self initWithBase:base useShared:NO];
}

/* ---------------------------------------------------------------- */

- (Logger*)initWithBase:(const char*)base useShared:(BOOL)shared
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%s,%d)\n",PREFUN,base,(int)shared);
#endif

  if ((self = [super init]) == nil) return nil;  // Clang complains without
  
  NSBundle *bundle = [NSBundle mainBundle];
  p_build  = [[bundle objectForInfoDictionaryKey:@"CFBundleVersion"] retain];
  p_exec   = [[bundle objectForInfoDictionaryKey:@"CFBundleExecutable"] retain];
  if (base) baseName = Ychr2str(base);
  else      baseName = [p_exec lowercaseString];
  [baseName retain];

  messageLock = [[NSLock alloc] init];
  fileLock = [[NSLock alloc] init];
#ifdef INCLUDE_NOTICE_H
  noticeLock = [[NSLock alloc] init];
#endif  
  mesgArray = [[NSMutableArray alloc] initWithCapacity:N_MSG];
  timeArray = [[NSMutableArray alloc] initWithCapacity:N_MSG];
  assert(mesg_head == 0);

  sharedPath = shared;
  [self setLogfile:[self logfileName]];
           
  return self;
}

/* ---------------------------------------------------------------- */

- (void)dealloc
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,self);
#endif

  [p_build release];
  [p_exec release];
  [baseName release];
  [messageLock release];
  [fileLock release];
#ifdef INCLUDE_NOTICE_H
  [noticeLock release];
#endif
  [mesgArray release];
  [timeArray release];
  
  [super dealloc];
}

/* ---------------------------------------------------------------- */

- (NSString*)logfileName
{
  time_t    ut; 
  struct tm *now,res;
  char      tstr[32];
  NSString  *path,*name;

  ut = time(NULL) - 9*3600;            // 9am UT
  now = localtime_r(&ut,&res);
  strftime(tstr,sizeof(tstr),"%Y_%W",now);

  if (sharedPath) {
    path = [NSString stringWithFormat:@"%s/%s/%@",SHARED_BASE,SHARED_LOGS,p_exec]; // 2014-07-01
  } else { NSArray *ary;
    ary = NSSearchPathForDirectoriesInDomains(NSLibraryDirectory,NSUserDomainMask,YES);
    if ([ary count] > 0) {
      path = [NSString stringWithFormat:@"%@/Logs/%@",[ary objectAtIndex:0],p_exec];
    } else {
      path = NSHomeDirectory();
    }
  }
  name = [NSString stringWithFormat:@"%@/%@%@.log",path,baseName,Ychr2str(tstr)];
  
  return name;
}

/* ---------------------------------------------------------------- */

- (void)setLogfile:(NSString*)file
{
  char buf[M_FILE];
   
  NSString *path = [file stringByDeletingLastPathComponent];
  NSFileManager *fm = [[NSFileManager alloc] init]; // MT-safe
  [fm createDirectoryAtPath:path  withIntermediateDirectories:YES
      attributes:nil error:NULL];      // create path if necessary
  [fm release];
  if (sharedPath) YfilePermissions(Ystr2tmp(file),6,6,6); // rw-,rw-,rw-

  [fileLock lock];
  if (logfile) [logfile release];
  logfile = [file retain];
  [fileLock unlock];
  
  [self message:Ystr2tmp(file) File:YES];
  sprintf(buf,"build= %s",Ystr2tmp(p_build));
  [self append:buf];
}


/* ---------------------------------------------------------------- */

- (void)setField:(NSTextField*)field
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,field);
#endif
  [messageLock lock];
  messageField = field;
  [messageLock unlock];
}

/* ---------------------------------------------------------------- */

- (void)setTable:(NSTableView*)table
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,table);
#endif
  
  [messageLock lock];
  messageTable = table;
  if (table) [table setDataSource:self];
  [messageLock unlock];    
}

/* ---------------------------------------------------------------- */

- (BOOL)changeLogfile
{
  BOOL r=FALSE;
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p)\n",PREFUN,self);
#endif

  NSString *string = [self logfileName];
  if ([logfile compare:string] != NSOrderedSame) { // name changed
    [self setLogfile:string];                      // set new name
    r = TRUE;
  }

  return r;
}

/* ---------------------------------------------------------------- */

static void replaceElement(NSMutableArray* array,NSString* string,int head)
{
  if (head < [array count]) {  // NSMutableArary protected by 'messageLock'
    [array replaceObjectAtIndex:head withObject:string];
  } else {
    [array addObject:string];
  }
}

/* ---------------------------------------------------------------- */

- (void)addMessage:(const char*)text
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(): '%s'\n",PREFUN,text);
#endif
  assert(messageLock);
  assert(text);

  if (messageField) {                  // overwrite text field
    [messageLock lock];
    [messageField performSelectorOnMainThread:@selector(setStringValue:)
                  withObject:Ychr2str(text) waitUntilDone:YES];
    [messageLock unlock];
  }
  if (messageTable) {                  // add to table
    assert(timeFormatter);
    [messageLock lock];
    replaceElement(timeArray,[timeFormatter stringFromDate:[NSDate date]],mesg_head);
    assert([timeArray count] <= N_MSG);
    replaceElement(mesgArray,Ychr2str(text),mesg_head);
    assert([mesgArray count] <= N_MSG);
    mesg_head = (mesg_head+1) % N_MSG;
    [messageLock unlock];
    [messageTable performSelectorOnMainThread:@selector(reloadData)
                  withObject:nil waitUntilDone:NO];
  }
}

/* ---------------------------------------------------------------- */

- (void)message:(const char*)text Level:(int)level Time:(int)dt File:(BOOL)file
{
  char *ptr=NULL;
#if (DEBUG > 1)
  fprintf(stderr,"%s(%s)\n",PREFUN,text);
#endif

  if (file || (level >= 0)) { char *p; // sanitize
    p = ptr = (char*)malloc(1+strlen(text));
    strcpy(ptr,text);
    while (*p) {
      switch (*p) {
      case '\r': case '\n':            // replace <CR>,<LF> with <SPC>
        *p = ' ';
        break;
      }
      p++;
    }
  }

  if (file) [self append:ptr];

  if (level >= 0) {                    // write to message window
    [self addMessage:ptr];
#ifdef INCLUDE_NOTICE_H
    assert(noticeLock);
    [noticeLock lock];
    if (level > 0) {                   // open notice?
      NSString *string = Ychr2str(text);
      if (!last_text || ([last_text compare:string] != NSOrderedSame) || (dt > 0)) {
        [Notice openNotice:text Title:Ystr2tmp(p_exec) Time:dt Level:level];
        if (last_text) [last_text release];
        last_text = [string retain];
      }
    }
    [noticeLock unlock];
#endif
  }

  if (ptr) free((void*)ptr);
}

/* --- */

- (void)message:(const char*)text Level:(int)level Time:(int)dt
{
  [self message:text Level:level Time:dt File:NO];
}

- (void)message:(const char*)text Level:(int)level File:(BOOL)file
{
  [self message:text Level:level Time:0 File:file];
}

- (void)message:(const char*)text File:(BOOL)file
{
  [self message:text Level:0 Time:0 File:file];
}

- (void)message:(const char*)text Level:(int)level
{
  [self message:text Level:level Time:0 File:NO];
}

- (void)message:(const char*)text
{
  [self message:text Level:0 Time:0 File:NO];
}

/* ---------------------------------------------------------------- */

- (void)append:(const char*)text 
{
  char file[M_FILE];
  FILE *fp;
#if (DEBUG > 1)
  fprintf(stderr,"%s(%s)\n",PREFUN,text);
#endif
  
  [fileLock lock];
  fp = fopen(Ystr2chr(logfile,file,sizeof(file)),"a"); // append to logfile
  if (fp == NULL) {
    fprintf(stderr,"%s(): cannot open file '%s'\n",PREFUN,file);
  } else { char buf[32]; struct tm *now,res; time_t ut;
    ut = time(NULL);  
    now = localtime_r(&ut,&res);
    strftime(buf,sizeof(buf),"%b %d %H:%M:%S",now);
    fprintf(fp,"%s %s\n",buf,text);
    fclose(fp);
  }
  [fileLock unlock];
}


/* ---------------------------------------------------------------- */
/* NSTableView dataSource */
/* ---------------------------------------------------------------- */

- (NSInteger)numberOfRowsInTableView:(NSTableView*)table
{
#if (DEBUG > 2)
  fprintf(stderr,"%s(%p): msg_n=%d\n",PREFUN,table,msg_n);
#endif
  return [mesgArray count];
}

/* ---------------------------------------------------------------- */

- (id)tableView:(NSTableView*)table
      objectValueForTableColumn:(NSTableColumn*)column
      row:(NSInteger)row
{
  int      i;
  NSString *string=nil;

#if (DEBUG > 1)
  fprintf(stderr,"%s(%p,%p,%d)\n",PREFUN,table,column,row);
#endif
  assert(row < [mesgArray count]);
  if (row >= [mesgArray count]) return nil;
  
#if (DEBUG > 0) && !defined(NDEBUG) && defined(INCLUDE_NOTICE_H)
  if (![NSThread isMainThread]) { char buf[128];
    sprintf(buf,"isMainThread=%d, row=%ld\n",[NSThread isMainThread],(long)row);
    [Notice openNotice:buf Title:PREFUN Time:600 Level:1];
  }
#endif

  [messageLock lock];
  i = mesg_head-(int)row-1; if (i < 0) i += N_MSG;
  switch ([[column identifier] intValue]) {
  case 0:                              // time
    string = [timeArray objectAtIndex:i];
    break;
  case 1:                              // text
    string = [mesgArray objectAtIndex:i];
    break;
  default:
    assert(0);
  }
  [messageLock unlock];

  return string;
}

/* ---------------------------------------------------------------- */

@end

/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
