//
// Yutils.m
//
// Created by Christoph Birk on 2011-04-19
// Copyright 2011 Carnegie Observatories. All rights reserved.
//
// 2013-07-11  Yutils class
// 2013-09-11  replace Yutils class with GCD
// 2014-03-xx  Shared Preferences
// 2014-04-02  fix permission on Shared Preference
// 2014-04-03  Ystr2tmp()
// 2014-04-07  combine Ypref and Yshared
// 2014-04-10  YprefSharedSetup()
// 2014-04-18  YbundleInfo(char**,...)
// 2014-04-21  YappSupportCreate(char**),const char* YbundleResource
// 2014-05-09  YimageFile() accepts NULL-pointer
// 2014-05-13  YarrayToValues()

/* ---------------------------------------------------------------- */

#ifndef DEBUG
#define DEBUG      1
#endif

/* ---------------------------------------------------------------- */

#import  "yutils.h"

/* ---------------------------------------------------------------- */
/* Bundle */
/* ---------------------------------------------------------------- */

static char* copy_rescource(NSString *string,char** copy)
{
  if (string) {
    const char *temp = Ystr2tmp(string);
    *copy = (char*)malloc((1+strlen(temp))*sizeof(char));
    strcpy(*copy,temp);
  } else {
    *copy = NULL;
  }
  return *copy;
}

/* --- */

BOOL YbundleInfo(char** exec,char** build,char** version,
                 char** info,char** path,char** ident)
{
  BOOL     r=TRUE;
  NSString *string;
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p,%p,%p,%p,%p,%p)\n",PREFUN,exec,build,version,
          info,path,ident);
#endif
  assert([NSThread isMainThread]);
  
  NSBundle *bundle = [NSBundle mainBundle]; if (!bundle) return FALSE;

  if (exec) {
    string = [bundle objectForInfoDictionaryKey:@"CFBundleExecutable"];
    if (!copy_rescource(string,exec)) r = FALSE;
  }
  if (build) {
    string = [bundle objectForInfoDictionaryKey:@"CFBundleVersion"];
    if (!copy_rescource(string,build)) r = FALSE;
  }
  if (version) { 
    string = [bundle objectForInfoDictionaryKey:@"CFBundleShortVersionString"];
    if (!copy_rescource(string,version)) r = FALSE;
  }
  if (info) {
    string = [bundle objectForInfoDictionaryKey:@"NSHumanReadableCopyright"];
    if (!string) string = [bundle objectForInfoDictionaryKey:@"CFBundleGetInfoString"];
    if (!copy_rescource(string,info)) r = FALSE;
  }
  if (path) {
    if (!copy_rescource([bundle resourcePath],path)) r = FALSE;
  }
  if (ident) {
    if (!copy_rescource([bundle bundleIdentifier],ident)) r = FALSE;
  }
  
  return r;
}

/* ---------------------------------------------------------------- */

#if 1 // IDEA: obsolete ?
const char* YbundleResource(const char* name,const char* ext)
{
  assert([NSThread isMainThread]);

  NSString *string = [[NSBundle mainBundle] pathForResource:Ychr2str(name) 
                                            ofType:Ychr2str(ext)];
  if (!string) return NULL;
  
  return Ystr2tmp(string);
}
#endif

/* ---------------------------------------------------------------- */

NSArray* YdirectoryURL(NSUInteger type)
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%lu)\n",PREFUN,(u_long)type);
#endif
  const NSUInteger mask=NSLocalDomainMask|NSUserDomainMask;

  NSFileManager *fm = [[NSFileManager alloc] init];  // MT-safe
  NSArray *array = [fm URLsForDirectory:type inDomains:mask];
  for (NSURL *url in array) NSLog(@"%@",[url path]);
  [fm release];
  
  return array;
}

/* ---------------------------------------------------------------- */

char* YappSupportCreate(char **path)
{
  NSString *name,*dir;
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p)\n",PREFUN,exec);
#endif

  name = [[NSBundle mainBundle] objectForInfoDictionaryKey:@"CFBundleExecutable"];
  
  NSFileManager *fm = [[NSFileManager alloc] init];  // MT-safe

  NSArray *array = [fm URLsForDirectory:NSApplicationSupportDirectory
                       inDomains:NSUserDomainMask];
  
  if ([array count] > 0) {            // ~/Library/Application Support
    dir = [[array firstObject] path];
  } else {                            // fall-back to $HOME directory
    dir = NSHomeDirectory();
  }
  NSString *string = [NSString stringWithFormat:@"%@/%@",dir,name];
  
  if (![fm fileExistsAtPath:string]) {
    [fm createDirectoryAtPath:string withIntermediateDirectories:NO
        attributes:nil error:NULL];
  }
  [fm release];
  
  const char *temp = Ystr2tmp(string);
  *path = (char*)malloc((1+strlen(temp))*sizeof(char));
  strcpy(*path,temp);
  
  return *path;
}

/* ---------------------------------------------------------------- */
/* Strings */
/* ---------------------------------------------------------------- */

char* Ystr2chr(NSString* string,char* buf,int maxlen)
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p,%p,%d)\n",PREFUN,string,buf,maxlen);
#endif
  
  assert(string);                      
  if (![string getCString:buf maxLength:maxlen encoding:NSASCIIStringEncoding]) {
    if (![string getCString:buf maxLength:maxlen encoding:NSUTF8StringEncoding]) {
      assert(0);
      *buf = '\0';
    }
  }
  return buf;
}

/* --- */

const char* Ystr2tmp(NSString* string)
{
  const char *r=NULL;
  assert(string);                      

  if (!(r=[string cStringUsingEncoding:NSASCIIStringEncoding])) {
    if (!(r=[string cStringUsingEncoding:NSUTF8StringEncoding])) {
      static const char *empty="";
      assert(0);
      return empty;
    }
  }
  return r;
}

/* ---------------------------------------------------------------- */

NSString* Ychr2str(const char *text)
{
  assert(text);
  NSString *string = [[NSString alloc] initWithCString:text 
                                       encoding:NSASCIIStringEncoding];
                                       
  return [string autorelease];         // NSString is MT-safe
}

/* ---------------------------------------------------------------- */
// Arrays

NSArray* YarrayFromString(const char* string,const char* sep)
{
  char *buffer,*b,*p;
  
  assert(string);
  NSMutableArray *ary = [NSMutableArray arrayWithCapacity:8];
  buffer = b = (char*)malloc(1+strlen(string));
  strcpy(buffer,string);
  while ((p=strsep(&b,sep))) {         
    if (*p) [ary addObject:Ychr2str(p)];
  }  
  free((void*)buffer);
  
  return [NSArray arrayWithArray:ary];
}

/* ---------------------------------------------------------------- */

NSArray* YarrayFromStringQ(const char* string) // '' separates items
{
  char *buffer,*b,*p;
  
#if (DEBUG > 0)
  fprintf(stderr,"%s(%s)\n",PREFUN,string);
#endif
  assert(string);
  NSMutableArray *ary = [NSMutableArray arrayWithCapacity:8];
  buffer = b = (char*)malloc(1+strlen(string));
  strcpy(buffer,string);
  while (*b) {
    while ((*b) && (*b != '\'')) b++;  // find first quote
    if (!(*b)) break;
    p = ++b;
    while ((*b) && (*b != '\'')) b++;  // find second quote
    if (!(*b)) break;
    *b = '\0'; b++;                    // terminate sub-string
    [ary addObject:Ychr2str(p)];       // append to array
  }  
  free((void*)buffer);
  
  return [NSArray arrayWithArray:ary];
}

/* ---------------------------------------------------------------- */

NSArray* YarrayFromFloats(float* values,int n)
{
  assert(values);
  NSMutableArray *ary = [NSMutableArray arrayWithCapacity:n];
  for (int i=0; i<n; i++) {
    [ary addObject:[NSNumber numberWithFloat:values[i]]];
  }
  
  return [NSArray arrayWithArray:ary];
}

/* --- */

NSArray* YarrayFromInts(int* values,int n)
{
  
  assert(values);
  NSMutableArray *ary = [NSMutableArray arrayWithCapacity:n];
  for (int i=0; i<n; i++) {
    [ary addObject:[NSNumber numberWithInt:values[i]]];
  }
  
  return [NSArray arrayWithArray:ary];
}

/* --- */

int YarrayToValues(NSArray* array,int* values,int n)
{
  int i;
  
  for (i=0; i<[array count]; i++) {
    if (i == n) break;
    values[i] = [[array objectAtIndex:i] intValue];
  }

  return i;
}

/* ---------------------------------------------------------------- */
/* TextField */
/* ---------------------------------------------------------------- */

void YeditText(NSTextField *field,const char* text)
{
  NSString *string;
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p,%p)\n",PREFUN,field,text);
#endif

  string = [NSString stringWithCString:text encoding:NSASCIIStringEncoding];
  if ([NSThread isMainThread]) {
    [field setStringValue:string];
  } else {
    [field performSelectorOnMainThread:@selector(setStringValue:)
           withObject:string waitUntilDone:YES];
  }
}

/* ---------------------------------------------------------------- */

char* YeditGet(NSTextField *field,char* text,int maxlen)
{
  return Ystr2chr([field stringValue],text,maxlen); // MT-safe
}

/* ---------------------------------------------------------------- */

void YeditColor(NSTextField* field,float r,float g,float b,float a)
{  
  NSColor *color = [NSColor colorWithCalibratedRed:r green:g blue:b alpha:a];
  if ([NSThread isMainThread]) {
    [field setTextColor:color]; 
  } else {
    [field performSelectorOnMainThread:@selector(setTextColor:)
           withObject:color waitUntilDone:YES];
  }
}

/* ---------------------------------------------------------------- */

NSTextField* YeditCreate(NSView* view,int x,int y,int w)
{
  assert([NSThread isMainThread]);
  
  NSTextField *edit = [[NSTextField alloc] initWithFrame:NSMakeRect(x,y,w,22)];
  [view addSubview:edit];

  return [edit autorelease];
}

/* ---------------------------------------------------------------- */

NSTextField* YlabelCreate(NSView* view,int x,int y,int w,const char* text)
{
  assert([NSThread isMainThread]);
  
  NSTextField *edit = [[NSTextField alloc] initWithFrame:NSMakeRect(x,y,w,17)];
  [edit setSelectable:NO];
  [edit setBordered:NO];
  [edit setDrawsBackground:NO];
  [edit setAlignment:NSCenterTextAlignment];
  [edit setStringValue:Ychr2str(text)];
  [view addSubview:edit];
  
  return [edit autorelease];
}

/* ---------------------------------------------------------------- */
/* PopupButton */
/* ---------------------------------------------------------------- */

NSArray* YpopupFromXML(NSPopUpButton* popup,NSString* name)
{
  NSString *file = [[NSBundle mainBundle] pathForResource:name ofType:@"xml"];
  NSArray *ary = [NSArray arrayWithContentsOfFile:file];
  YpopupFromArray(popup,ary);
  
  return ary;
}

void YpopupFromArray(NSPopUpButton* popup,NSArray* ary)
{
  assert([NSThread isMainThread]);
  
  [popup addItemsWithTitles:ary];
  NSMenu *menu = [popup menu];
  for (int i=0; i<[menu numberOfItems]; i++) {
    NSMenuItem *item = [menu itemAtIndex:i];  
    NSRange range = [[item title] rangeOfString:@"-empty" 
                                  options:NSCaseInsensitiveSearch];
    if (range.location != NSNotFound) {
      [item setHidden:YES];
    }
  }
}

/* ---------------------------------------------------------------- */

void YpopupSelectItem(NSPopUpButton* popup,int item)
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p,%d)\n",PREFUN,popup,item);
#endif

  if ([NSThread isMainThread]) {
    [popup selectItemAtIndex:item];
  } else {
    dispatch_sync(dispatch_get_main_queue(),^(void) {
      [popup selectItemAtIndex:item];
    });
  }
}

/* ---------------------------------------------------------------- */

void YpopupSetTitle(NSPopUpButton* popup,const char* title)
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p,%s)\n",PREFUN,popup,title);
#endif

  if ([NSThread isMainThread]) {
    [popup setTitle:Ychr2str(title)];
  } else {
    [popup performSelectorOnMainThread:@selector(setTitle:)
           withObject:Ychr2str(title) waitUntilDone:YES];
  }
}

/* ---------------------------------------------------------------- */

void YpopupEnable(NSPopUpButton* popup,BOOL enable)
{
  YcontrolEnable(popup,enable);
}

/* NSControl */

void YcontrolEnable(NSControl* control,BOOL enable)
{
  if ([NSThread isMainThread]) {
    [control setEnabled:enable];
  } else {
    dispatch_sync(dispatch_get_main_queue(),^(void) {
      [control setEnabled:enable];
    });
  }
}

/* ---------------------------------------------------------------- */

void YviewNeedsDisplay(NSView* view)
{
  if ([NSThread isMainThread]) {
    [view setNeedsDisplay:YES];
  } else {
    dispatch_async(dispatch_get_main_queue(),^(void) {
      [view setNeedsDisplay:YES];
    });
  }
}

/* ---------------------------------------------------------------- */
/* Window */
/* ---------------------------------------------------------------- */

void YwindowTitle(NSWindow* window,const char* title)
{
  if ([NSThread isMainThread]) {       // execute on mainThread
    [window setTitle:Ychr2str(title)];
  } else {
    [window performSelectorOnMainThread:@selector(setTitle:)
            withObject:Ychr2str(title) waitUntilDone:NO];
  }
}

/* ---------------------------------------------------------------- */

void YwindowFrameSet(NSWindow* window,const char* name)
{
  assert([NSThread isMainThread]);
  [window setFrameUsingName:Ychr2str(name)];
}

/* ---------------------------------------------------------------- */

void YwindowFrameSave(NSWindow* window,const char* name)
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p,%s) isVisible=%d\n",PREFUN,window,name,[window isVisible]);
#endif
  assert([NSThread isMainThread]);

  if ([window isVisible]) {
    [window saveFrameUsingName:Ychr2str(name)];
    [[NSUserDefaults standardUserDefaults] synchronize];
  }
}

/* ---------------------------------------------------------------- */
/* Menu */
/* ---------------------------------------------------------------- */

void YmenuEnableTitles(const char* list,BOOL e)
{
  char *b,*buffer,*p;
  NSMenuItem *item=nil;
#if (DEBUG > 1)
  fprintf(stderr,"%s(%s,%d)\n",PREFUN,list,(int)e);
#endif
  assert([NSThread isMainThread]);
  
  buffer = b = (char*)malloc(1+strlen(list));
  strcpy(buffer,list);
  NSMenu *menu = [NSApp mainMenu];
  // printf("setAutoenablesItems=%d\n",[menu autoenablesItems]);
  while ((p=strsep(&b,","))) {
    if (*p) {
      item = [menu itemWithTitle:Ychr2str(p)];
      if (e) [item setEnabled:e];
      menu = [item submenu];
    }
  }
  // NSLog(@"%@",[item title]);
  [item setEnabled:e];
  free((void*)buffer);
}

/* ---------------------------------------------------------------- */

NSMenuItem* YmenuItem(const char* list)
{
  char *b,*buffer,*p;
  NSMenuItem *item=nil;
  
  assert([NSThread isMainThread]);
  buffer = b = (char*)malloc(1+strlen(list));
  strcpy(buffer,list);
  NSMenu *menu = [NSApp mainMenu];
  while ((p=strsep(&b,","))) {
    if (*p) {
      item = [menu itemWithTitle:Ychr2str(p)];
      menu = [item submenu];
    }
  }  
  free((void*)buffer);

  return item;
}

/* ---------------------------------------------------------------- */
/* Image */
/* ---------------------------------------------------------------- */

NSImage* YimageFile(NSImageView *view,const char* file)
{
  NSString *string;
  NSImage  *image;                     // not MT-safe
  
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p,%s)\n",PREFUN,view,file);
#endif
  if (!file) return nil;
  
  string = [NSString stringWithCString:file encoding:NSASCIIStringEncoding];
  image  = [[NSImage alloc] initWithContentsOfFile:string];

  if (view) {  
    if ([NSThread isMainThread]) {
      [view setImage:image];
    } else {
      [view performSelectorOnMainThread:@selector(setImage:)
            withObject:image waitUntilDone:YES];   
    }
  }
  return [image autorelease];
}

/* ---------------------------------------------------------------- */

NSImage* YimageData(NSImageView *view,const u_int *data,int w,int h)
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p,%p,%d,%d)\n",PREFUN,view,data,w,h);
#endif
  
  NSBitmapImageRep *bitmap = [NSBitmapImageRep alloc];

  bitmap = [bitmap initWithBitmapDataPlanes:NULL
                   pixelsWide:w
                   pixelsHigh:h
                   bitsPerSample:8
                   samplesPerPixel:3
                   hasAlpha:NO
                   isPlanar:NO
                   colorSpaceName:NSDeviceRGBColorSpace
                   bytesPerRow:sizeof(u_int)*w
                   bitsPerPixel:32];
  memcpy([bitmap bitmapData],data,sizeof(u_int)*w*h);
  NSImage *image = [[NSImage alloc] init];
  [image addRepresentation:bitmap];
  
  if ([NSThread isMainThread]) {
    [view setImage:image];
  } else {
    [view performSelectorOnMainThread:@selector(setImage:)
          withObject:image waitUntilDone:YES];   
  }
  [bitmap release];
  
  return [image autorelease];
}

/* ---------------------------------------------------------------- */
/* Table Views */
/* ---------------------------------------------------------------- */

#if 0 // Note: simpler with [[column identifier] intValue]
// int YgetTableColumn(NSTableView* table,NSTableColumn* column)
{
  NSArray *cols = [table tableColumns];

  return [cols indexOfObject:column];
}
#endif

/* ---------------------------------------------------------------- */
/* Preferences */
/* ---------------------------------------------------------------- */

BOOL YprefSetupShared(void)
{
  BOOL r=TRUE;
  NSString *exec,*path;
  NSDictionary *attr;
  
  NSFileManager *fm = [[NSFileManager alloc] init]; // MT-safe
  
  NSNumber *perm = [NSNumber numberWithShort:((7*64)+(7*8)+7)];
  attr = [NSDictionary dictionaryWithObjectsAndKeys:perm,NSFilePosixPermissions,nil];
  exec = [[NSBundle mainBundle] objectForInfoDictionaryKey:@"CFBundleExecutable"];
  
  if (r) {
    path = [NSString stringWithFormat:@"%s",SHARED_BASE];
    [fm createDirectoryAtPath:path withIntermediateDirectories:NO
        attributes:attr error:nil];
  }
  if (r) {
    path = [NSString stringWithFormat:@"%s/%s",SHARED_BASE,SHARED_PREF];
    r = [fm createDirectoryAtPath:path withIntermediateDirectories:YES
            attributes:attr error:nil];
  }
  if (r) {
    path = [NSString stringWithFormat:@"%s/%s",SHARED_BASE,SHARED_APPS];
    [fm createDirectoryAtPath:path withIntermediateDirectories:NO
        attributes:attr error:nil];
    path = [NSString stringWithFormat:@"%s/%s/%@",SHARED_BASE,SHARED_APPS,exec];
    r = [fm createDirectoryAtPath:path withIntermediateDirectories:YES
            attributes:attr error:nil];
  }
  if (r) {
   path = [NSString stringWithFormat:@"%s/%s",SHARED_BASE,SHARED_LOGS];
   [fm createDirectoryAtPath:path withIntermediateDirectories:NO
       attributes:attr error:nil];
    path = [NSString stringWithFormat:@"%s/%s/%@",SHARED_BASE,SHARED_LOGS,exec];
    r = [fm createDirectoryAtPath:path withIntermediateDirectories:YES
            attributes:attr error:nil];
  }
  
  [fm release];

  return r;
}

/* ---------------------------------------------------------------- */

#define SHARED_FILE  "%s/%s/%@.plist"

static NSMutableDictionary* YsharedRead(void)
{
  NSString *ident = [[NSBundle mainBundle] bundleIdentifier];
  NSString *file = [NSString stringWithFormat:@SHARED_FILE,SHARED_BASE,SHARED_PREF,ident];
  
  NSMutableDictionary *dict = [NSMutableDictionary dictionaryWithContentsOfFile:file];
  
  return dict;
}

/* --- */

static NSObject* YsharedGetObject(NSString* key)
{
  NSMutableDictionary *dict=nil;
  NSObject *object=nil;
  
  @synchronized([NSUserDefaults standardUserDefaults]) {
    dict = YsharedRead();
  }
  if (dict) {
    object = [dict objectForKey:key];
  }
  return object;
}

/* --- */

static BOOL YsharedWrite(NSMutableDictionary* dict)
{
  NSString *ident = [[NSBundle mainBundle] bundleIdentifier];
  NSString *file = [NSString stringWithFormat:@SHARED_FILE,SHARED_BASE,SHARED_PREF,ident];

  BOOL r = [dict writeToFile:file atomically:NO];
  if (r) {
    YfilePermissions(Ystr2tmp(file),6,6,6);  // rw-,rw-,rw-
  }
  
  return r;
}

/* --- */

static BOOL YsharedPutObject(NSString* key,NSObject* object)
{
  BOOL r=FALSE;
  assert(object);
  
  @synchronized([NSUserDefaults standardUserDefaults]) {
    NSMutableDictionary *dict = YsharedRead();
    if (!dict) dict = [NSMutableDictionary dictionaryWithCapacity:1];
    assert(dict);
    [dict setObject:object forKey:key];
    r = YsharedWrite(dict);
  }
  return r;
}

#undef SHARED_FILE

/* ---------------------------------------------------------------- */

void YprefPutInt(const char* key,int value,BOOL shared)
{
#if (DEBUG > 1)
  NSLog(@"%s(%s,%d,%d)",PREFUN,key,value,(int)shared);
#endif
  if (shared) {
    if (!YsharedPutObject(Ychr2str(key),[NSNumber numberWithInt:value])) {
      NSLog(@"%s(%s,%d) failed",PREFUN,key,value);
    }
  } else {
    [[NSUserDefaults standardUserDefaults] setInteger:value forKey:Ychr2str(key)];
  }
}

/* --- */

int YprefGetInt(const char* key,int value,BOOL shared)
{
  NSNumber *number;

  if (shared) {
    number = (NSNumber*)YsharedGetObject(Ychr2str(key));
  } else {
    number = [[NSUserDefaults standardUserDefaults] objectForKey:Ychr2str(key)];
  }
  if (number) value = [number intValue];
  
  return value;
}

/* ---------------------------------------------------------------- */

void YprefPutDouble(const char* key,double value,BOOL shared)
{
  if (shared) {
    if (!YsharedPutObject(Ychr2str(key),[NSNumber numberWithDouble:value])) {
      NSLog(@"%s(%s,%f) failed",PREFUN,key,value);
    }
  } else {
    [[NSUserDefaults standardUserDefaults] setDouble:value forKey:Ychr2str(key)];
  }
}

/* --- */

double YprefGetDouble(const char* key,double value,BOOL shared)
{
  NSNumber *number;

  if (shared) {
    number = (NSNumber*)YsharedGetObject(Ychr2str(key));
  } else {
    number = [[NSUserDefaults standardUserDefaults] objectForKey:Ychr2str(key)];
  }
  if (number) value = [number doubleValue];

  return value;
}

/* ---------------------------------------------------------------- */

void YprefPutChar(const char* key,const char* value,BOOL shared)
{
  YprefPutString(Ychr2str(key),Ychr2str(value),shared);
}

/* --- */

char* YprefGetChar(const char* key,const char* defstr,char* value,int len,BOOL shared)
{
  NSString *string = YprefGetString(Ychr2str(key),Ychr2str(defstr),shared);
  Ystr2chr(string,value,len);

  return value;
}

/* ---------------------------------------------------------------- */

void YprefPutString(NSString* key,NSString* value,BOOL shared)
{
  if (shared) {
    if (!YsharedPutObject(key,value)) {
      NSLog(@"%s(%@,%@) failed",PREFUN,key,value);
    }
  } else {
    [[NSUserDefaults standardUserDefaults] setObject:value forKey:key];
  }
}

/* --- */


NSString* YprefGetString(NSString* key,NSString* defstr,BOOL shared)
{
  NSString *string;
  
  if (shared) {
    string = (NSString*)YsharedGetObject(key);
  } else {
    string = [[NSUserDefaults standardUserDefaults] objectForKey:key];
  }
  if (string) return string;
  return defstr;
}

/* ---------------------------------------------------------------- */

void YprefPutArray(const char* key,NSArray* array,BOOL shared)
{
  if (shared) {
    if (!YsharedPutObject(Ychr2str(key),array)) {
      NSLog(@"%s(%s,%p) failed",PREFUN,key,array);
    }
  } else {
    [[NSUserDefaults standardUserDefaults] setObject:array forKey:Ychr2str(key)];
  }
}

/* --- */

NSArray* YprefGetArray(const char* key,BOOL shared)
{
  NSArray *array;
  
  if (shared) {
    array = (NSArray*)YsharedGetObject(Ychr2str(key));
  } else {
    array = [[NSUserDefaults standardUserDefaults] objectForKey:Ychr2str(key)];
  }
  return array;
}

/* ---------------------------------------------------------------- */
/* Alert Panels */
/* ---------------------------------------------------------------- */

// NSAlertFirstButtonReturn=1000
// NSAlertSecondButtonReturn=1001
// NSAlertThirdButtonReturn=1002

NSInteger YalertPanel(const char* title,const char* message,int level,
                const char* but1,const char* but2,const char* but3)
{ 
  NSInteger i;

  assert([NSThread isMainThread]);

  NSAlert *alert = [[NSAlert alloc] init];
  
  if (level) [alert setAlertStyle:NSCriticalAlertStyle];
  [alert setMessageText:Ychr2str(title)];
  [alert setInformativeText:Ychr2str(message)];
  if (but1) [alert addButtonWithTitle:Ychr2str(but1)];
  if (but2) [alert addButtonWithTitle:Ychr2str(but2)];
  if (but3) [alert addButtonWithTitle:Ychr2str(but3)];
#if 0 // add accessory view
  NSTextView *accessory = [[NSTextView alloc] initWithFrame:NSMakeRect(0,0,200,15)];
  NSFont *font = [NSFont systemFontOfSize:[NSFont systemFontSize]];
  NSDictionary *textAttributes = [NSDictionary dictionaryWithObject:font 
                                               forKey:NSFontAttributeName];
  [accessory insertText:[[NSAttributedString alloc] 
                          initWithString:@"Text in accessory view"
                          attributes:textAttributes]];
  [accessory setEditable:NO];
  [accessory setDrawsBackground:NO];
  [alert setAccessoryView:accessory];
#endif
  i = [alert runModal]; // this blocks the main-thread (application modal)
  [alert release];
                                      
  return i;
}

/* ---------------------------------------------------------------- */

BOOL YalertQuestion(const char* title,const char* message,int level,
                    const char* okbut,const char* ccbut)
{ // Warning: this blocks the main-thread (application modal)
  
  NSInteger i = YalertPanel(title,message,level,okbut,ccbut,NULL);
  return ((i == NSAlertFirstButtonReturn) ? YES : NO);
}

/* ---------------------------------------------------------------- */

void YsheetNotify(NSWindow* window,const char* title,const char* message,
               id delegate,SEL selector,void* context)
{ // keeps the main-thread (timers) running
#if (DEBUG > 0)
  fprintf(stderr,"%s(%s,%s,%p)\n",PREFUN,title,message,context);
#endif
  NSString *tstr=Ychr2str(title);
  NSString *mstr=Ychr2str(message);
  
  NSBeginAlertSheet(tstr,@"OK",nil,nil,window,delegate,selector,nil,context,@"%@",mstr);
} 

/* ---------------------------------------------------------------- */

void YerrorSheet(NSWindow* window,const char* title,const char* text)
{
#if (DEBUG > 0)
  if (![NSThread isMainThread]) {
    fprintf(stderr,"%s(%s) called from secondary thread\n",PREFUN,text);
  }
#endif
  NSDictionary *dict = [NSDictionary dictionaryWithObjectsAndKeys:
                        Ychr2str(title),NSLocalizedDescriptionKey,
                        Ychr2str(text),NSLocalizedRecoverySuggestionErrorKey,
                        nil];
  NSError *error = [NSError errorWithDomain:@"myErrorDomain" code:0 userInfo:dict];
  [window presentError:error modalForWindow:window   // window modal
          delegate:nil didPresentSelector:nil contextInfo:NULL];
}

/* ---------------------------------------------------------------- */
/* Mutex, Semaphore, CriticalSection */
/* ---------------------------------------------------------------- */

Ymutex YmutexCreate(const char* name)
{
  NSRecursiveLock *lock = [[NSRecursiveLock alloc] init];

  [lock setName:Ychr2str(name)];

  return (Ymutex)lock;
}

BOOL YmutexLock(Ymutex mutex,double timeout)
{
  NSRecursiveLock *lock = (NSRecursiveLock*)mutex;

  if (timeout > 0.0) {                 // timeout interval
    NSDate *date = [NSDate dateWithTimeIntervalSinceNow:timeout];
    return [lock lockBeforeDate:date];
  } else
  if (timeout < 0.0) {                 // infinite wait
    [lock lock];
    return YES;
  }
  
  return [lock tryLock];              // poll
}

void YmutexRelease(Ymutex mutex)
{
  [(NSRecursiveLock*)mutex unlock];
}

/* ---------------------------------------------------------------- */

Yglobal YglobalCreate(const char* path)
{
  NSDistributedLock *lock = [[NSDistributedLock alloc] initWithPath:Ychr2str(path)];

  return (Yglobal)lock;
}

BOOL YglobalLock(Yglobal mutex,int timeout,int delta)
{
  BOOL r=NO;
  
  assert((timeout <= 0) || (delta > 0));
  
  do { 
    r = [(NSDistributedLock*)mutex tryLock]; if (r) break;
    timeout -= delta; 
    if (timeout > 0) Ysleep(delta);
  } while (timeout > 0);
  
  return r;
}

void YglobalRelease(Yglobal mutex)
{
  [(NSDistributedLock*)mutex unlock];
}

/* ---------------------------------------------------------------- */

Ysemaphore* YsemaphoreCreate(const char* name)
{
  Ysemaphore *sem;
  
  sem = (Ysemaphore*)malloc(sizeof(Ysemaphore));
  sem->lock = [[NSLock alloc] init];
  [sem->lock setName:Ychr2str(name)];
  sem->count = 0;
  
  return sem;
}

int YsemaphoreInc(Ysemaphore* sem)
{
  return YsemaphoreAdd(sem,+1); 
}

int YsemaphoreDec(Ysemaphore* sem)
{
  return YsemaphoreAdd(sem,-1); 
}

int YsemaphoreAdd(Ysemaphore* sem,int add)
{
  int count;
  
  [sem->lock lock];
  sem->count += add; count = sem->count;
  [sem->lock unlock];
  
  return count;
}

int YsemaphoreGet(Ysemaphore* sem)
{
  int count;
  
  [sem->lock lock];
  count = sem->count;
  [sem->lock unlock];

  return count;
}

int YsemaphoreIncLimit(Ysemaphore* sem,int limit)
{                                      
  int count;
  
  assert(limit > 0);

  while (1) {                          // Warning: no timeout
    [sem->lock lock];
    if (sem->count < limit) {
      sem->count += 1; break;
    }
    [sem->lock unlock];
    Ysleep(250);
  }
  count = sem->count;
  [sem->lock unlock];
  assert(count >  0);
  assert(count <= limit);

  return count;  
}

/* ---------------------------------------------------------------- */
// file operations
/* ---------------------------------------------------------------- */

BOOL YfileExists(const char* file)     // any file or directory
{
  BOOL r;
  
  if (*file == '\0') return YES;       // Note: special case
  
  NSFileManager *fm = [[NSFileManager alloc] init]; // MT-safe
  NSString *path = [fm stringWithFileSystemRepresentation:file length:strlen(file)];
  r = [fm fileExistsAtPath:path];
  [fm release];
  
  return r;
}

/* ---------------------------------------------------------------- */

BOOL YfileReadable(const char* file)   // any file
{
  BOOL r;
    
  NSFileManager *fm = [[NSFileManager alloc] init]; // MT-safe
  NSString *path = [fm stringWithFileSystemRepresentation:file length:strlen(file)];
  r = [fm isReadableFileAtPath:path];
  [fm release];
  
  return r;
}

/* --- */

BOOL YpathReadable(const char* file)   // directory only
{
  BOOL r=NO,e,d;
    
  NSFileManager *fm = [[NSFileManager alloc] init]; // MT-safe
  NSString *path = [fm stringWithFileSystemRepresentation:file length:strlen(file)];
  e = [fm fileExistsAtPath:path isDirectory:&d];
  if (e && d) {                        // exists and is directory
    r = [fm isReadableFileAtPath:path];
  }
  [fm release];
  
  return r;
}

/* ---------------------------------------------------------------- */

BOOL YfileWritable(const char* file)
{
  BOOL r;

  NSFileManager *fm = [[NSFileManager alloc] init]; // MT-safe
  NSString *path = [fm stringWithFileSystemRepresentation:file length:strlen(file)];
  r = [fm isWritableFileAtPath:path];
  [fm release];
  
  return r;
}

/* --- */

BOOL YpathWritable(const char* file)   // directory only
{
  BOOL r=NO,e,d;

  NSFileManager *fm = [[NSFileManager alloc] init]; // MT-safe
  NSString *path = [fm stringWithFileSystemRepresentation:file length:strlen(file)];
  e = [fm fileExistsAtPath:path isDirectory:&d];
  if (e && d) {                        // exists and is directory
    r = [fm isWritableFileAtPath:path];
  }
  [fm release];
  
  return r;
}

/* ---------------------------------------------------------------- */

BOOL YpathCreate(const char* path,BOOL intermediate)
{
  NSFileManager *fm = [[NSFileManager alloc] init]; // MT-safe
  NSString *string = [fm stringWithFileSystemRepresentation:path length:strlen(path)];
  BOOL r = [fm createDirectoryAtPath:string 
               withIntermediateDirectories:intermediate
               attributes:nil error:NULL];
  [fm release];
  
  return r;
}

/* ---------------------------------------------------------------- */

BOOL YfileRemove(const char* path)
{
  NSFileManager *fm = [[NSFileManager alloc] init]; // MT-safe
  NSString *string = [fm stringWithFileSystemRepresentation:path length:strlen(path)];
  BOOL r = [fm removeItemAtPath:string error:NULL];
  [fm release];
  
  return r;
}

/* ---------------------------------------------------------------- */

NSInteger YfileTrash(const char* filename)
{
  NSInteger tag;

  NSString *path = Ychr2str(filename);
  [[NSWorkspace sharedWorkspace] performFileOperation:NSWorkspaceRecycleOperation
                                 source:[path stringByDeletingLastPathComponent]
                                 destination:nil 
                                 files:[NSArray arrayWithObject:[path lastPathComponent]]
                                 tag:&tag];
  return tag;
}

/* ---------------------------------------------------------------- */

BOOL YfilePermissions(const char* file,short u,short g,short o)
{
  NSFileManager *fm = [[NSFileManager alloc] init]; // MT-safe
  NSNumber *perm = [NSNumber numberWithShort:((u*64)+(g*8)+o)];
  NSDictionary *dict = [NSDictionary dictionaryWithObjectsAndKeys:perm,NSFilePosixPermissions,nil];
  BOOL r = [fm setAttributes:dict ofItemAtPath:Ychr2str(file) error:nil];
  [fm release];
  
  return r;
}

/* ---------------------------------------------------------------- */

BOOL YfileCopy(const char* src, const char* trg)
{
  NSFileManager *fm = [[NSFileManager alloc] init]; // MT-safe
  BOOL r = [fm copyItemAtPath:Ychr2str(src) toPath:Ychr2str(trg) error:NULL];
  [fm release];
  
  return r;
}

/* ---------------------------------------------------------------- */
// Miscellaneous

void Ysleep(int msec)                  /* typically 0.2 msec over */
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%d)\n",PREFUN,msec);
#endif
  [NSThread sleepForTimeInterval:(double)msec/1000.0]; 
}

/* ---------------------------------------------------------------- */

NSTask* Ytask(const char* cmd,const char* par)
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%s,%s)\n",PREFUN,cmd,par);
#endif

  NSString *path = Ychr2str(cmd);
  NSArray *args = YarrayFromString(par," ");
  NSTask *task = [NSTask launchedTaskWithLaunchPath:path arguments:args];
  
  return task;
}

/* ---------------------------------------------------------------- */

#if 0 // replaced by 'Sound.m'
void Ysay(const char* phrase)
{
  Ytask("/usr/bin/say",phrase,NO);      // does not wait
}
#endif

/* ---------------------------------------------------------------- */

#if 0 // replaced by 'Sound.m'
void Ysound(const char* file)
{
  NSFileManager *fm = [[NSFileManager alloc] init]; // MT-safe
  NSString *path = [fm stringWithFileSystemRepresentation:file length:strlen(file)];
  NSArray *args = [NSArray arrayWithObject:path];
  [NSTask launchedTaskWithLaunchPath:@"/usr/bin/afplay" arguments:args];
  [fm release];
}
#endif

/* ---------------------------------------------------------------- */

void Ywww(const char* address)
{
  NSString *string = Ychr2str(address);
  NSURL *url = [NSURL URLWithString:string];
  [[NSWorkspace sharedWorkspace] openURL:url];
}

/* ---------------------------------------------------------------- */
/* CoreText */
/* ---------------------------------------------------------------- */

CTFontRef CtextGetFont(const char* family,float size)
{
  NSDictionary *fontAttr = [NSDictionary dictionaryWithObjectsAndKeys:
    Ychr2str(family),(NSString*)kCTFontFamilyNameAttribute,
    [NSNumber numberWithFloat:size],(NSString*)kCTFontSizeAttribute,
    nil];
  CTFontDescriptorRef fontDesc = CTFontDescriptorCreateWithAttributes((CFDictionaryRef)fontAttr);
  CTFontRef font = CTFontCreateWithFontDescriptor(fontDesc,0.0,NULL);
  CFRelease(fontDesc);

  return font;                   // Note: needs CFRelease()
}

/* ---------------------------------------------------------------- */

CTLineRef CtextGetLine(const char* text,CTFontRef font)
{
  CFStringRef keys[] = { kCTFontAttributeName };
  CFTypeRef values[] = { font };
  CFStringRef string = CFStringCreateWithCString(kCFAllocatorDefault,text,kCFStringEncodingASCII);

  CFDictionaryRef attributes =
    CFDictionaryCreate(kCFAllocatorDefault, (const void**)&keys,
        (const void**)&values, sizeof(keys) / sizeof(keys[0]),
        &kCFTypeDictionaryKeyCallBacks,
        &kCFTypeDictionaryValueCallBacks);
 
  CFAttributedStringRef attrString =
      CFAttributedStringCreate(kCFAllocatorDefault, string, attributes);
  CFRelease(string);
  CFRelease(attributes);
 
  CTLineRef line = CTLineCreateWithAttributedString(attrString);
  CFRelease(attrString);
  
  return line;
}

/* ---------------------------------------------------------------- */

CTLineRef CtextCreateColorLine(const char* text,CTFontRef font,
                               float r,float g,float b,float a)
{
  CGFloat components[4];         // IDEA: use [NSColor CGColor] in 10.8
  components[0] = (CGFloat)r;
  components[1] = (CGFloat)g;
  components[2] = (CGFloat)b;
  components[3] = (CGFloat)a;
  CGColorSpaceRef rgbColorSpace = CGColorSpaceCreateDeviceRGB();
  CGColorRef color = CGColorCreate(rgbColorSpace,components);
  CGColorSpaceRelease(rgbColorSpace);

  CFStringRef keys[] = { kCTFontAttributeName,kCTForegroundColorAttributeName };
  CFTypeRef values[] = { font,color };
  CFStringRef string = CFStringCreateWithCString(kCFAllocatorDefault,text,kCFStringEncodingASCII);

  CFDictionaryRef attributes =
    CFDictionaryCreate(kCFAllocatorDefault, (const void**)&keys,
        (const void**)&values, sizeof(keys) / sizeof(keys[0]),
        &kCFTypeDictionaryKeyCallBacks,
        &kCFTypeDictionaryValueCallBacks);
 
  CFAttributedStringRef attrString =
      CFAttributedStringCreate(kCFAllocatorDefault, string, attributes);
  CFRelease(string);
  CFRelease(attributes);
 
  CTLineRef line = CTLineCreateWithAttributedString(attrString);
  CFRelease(attrString);
  CGColorRelease(color);

  return line;
}

void CtextSetColorText(const char* text,CTFontRef font,CGContextRef c,
                       float x,float y,float r,float g,float b,float a)
{
  CTLineRef line = CtextCreateColorLine(text,font,r,g,b,a);
  CGContextSetTextPosition(c,x,y);
  CTLineDraw(line,c);
  CFRelease(line);
}

/* ---------------------------------------------------------------- */

void CtextSetText(const char* text,CTFontRef font,CGContextRef c,float x,float y)
{
  CTLineRef line = CtextGetLine(text,font);
  CGContextSetTextPosition(c,x,y);
  CTLineDraw(line,c);
  CFRelease(line);
}

/* ---------------------------------------------------------------- */

void CtextCenterText(const char* text,CTFontRef font,CGContextRef c,float x,float y)
{
  CTLineRef line = CtextGetLine(text,font);
  CGRect rect = CTLineGetImageBounds(line,c);
  CGContextSetTextPosition(c,x-rect.size.width/2,y-rect.size.height/2);
  CTLineDraw(line,c);
  CFRelease(line);
}

/* ---------------------------------------------------------------- */

void CtextRightText(const char* text,CTFontRef font,CGContextRef c,float x,float y)
{
  CTLineRef line = CtextGetLine(text,font);
  CGRect rect = CTLineGetImageBounds(line,c);
  CGContextSetTextPosition(c,x-rect.size.width,y-rect.size.height/2);
  CTLineDraw(line,c);
  CFRelease(line);
}

/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
