/* ---------------------------------------------------------------- */
//
// MTwister.m
//
// Created by Christoph Birk on 2013-04-12
// Copyright 2013 Carnegie Observatories. All rights reserved.
//
// Algorithm: Makoto Matsumoto and Takuji Nishimura (2002 version)
//            http://www.math.sci.hiroshima-u.ac.jp/~m-mat/MT/emt.html
//
// 2014-04-01  use [NSProcessInfo globallyUniqueString]
// 2014-04-18  use 'instanceLock' on defaultTwister
// 2014-10-08  -initWithBytes:Len:
// 2018-01-03  create 'classTwister' on demand (was @ initialize)
// 2018-11-01  move iVARs into implementation
//
/* ---------------------------------------------------------------- */

#ifndef DEBUG
#define DEBUG           0              // debug level
#endif

#define MT_M            397
#define MT_N            624
#define MATRIX_A        0x9908b0df     // constant vector
#define UPPER_MASK      0x80000000     // most significant w-r bits
#define LOWER_MASK      0x7fffffff     // least significant r bits

#define SAFETY_NET      0x5afeb10c     // safe block
#define SECURE_FACTOR   8              // secure uint

/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

#import  "MTwister.h"

#include <CommonCrypto/CommonDigest.h>

/* ---------------------------------------------------------------- */

static MTwister *classTwister=nil;
static u_short fallbackSeed=0;

/* --- */

u_int* hashBytes(u_char* bytes,int size) // execTime=0.5 usec
{
  assert(size >= CC_SHA256_DIGEST_LENGTH);
  CC_SHA256_CTX c;
  CC_SHA256_Init(&c);
  CC_SHA256_Update(&c,bytes,size);
  CC_SHA256_Final(bytes,&c);
  return (u_int*)bytes;
}

void addHashedUniqueString(NSMutableArray* array) // execTime = 7.5 usec
{
  CC_SHA256_CTX c;
  unichar u;
  u_char md[CC_SHA256_DIGEST_LENGTH];
  
  NSString *unique = [[NSProcessInfo processInfo] globallyUniqueString];
  CC_SHA256_Init(&c);
  for (int i=0; i<[unique length]; i++) {
    u = [unique characterAtIndex:i];  // printf("%08x\n",ubuf[i]);
    CC_SHA256_Update(&c,&u,sizeof(unichar));
  }
  CC_SHA256_Final(md,&c);

  u_int *data = (u_int*)md;
  for (int i=0; i<CC_SHA256_DIGEST_LENGTH/sizeof(u_int); i++) {
    [array addObject:[NSNumber numberWithUnsignedInt:data[i]]];
  }
}

/* --- */

void addRandomNumbers(NSMutableArray* array,int num)
{
  u_int r;
  BOOL  suspicious=FALSE;
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p,%d)\n",PREFUN,array,num);
#endif

  int fd = open("/dev/urandom",O_RDONLY);
  if (fd != -1) {
    for (int i=0; i<num; i++) {
      read(fd,&r,sizeof(unsigned int));      // r = 0xdeadbeaf; // TEST
      NSNumber *number = [NSNumber numberWithUnsignedInt:r];
      if ([array containsObject:number]) {
        suspicious = TRUE; NSLog(@"suspicious ... suspicious ... suspicious");
      }
      [array addObject:number];
    }
    close(fd);
  }
  if ((fd < 0) || suspicious) { u_int s,p,t; // Numerical Recipes: 1664525 1013904223
    @synchronized([MTwister class]) {
      s = (u_int)(++fallbackSeed & 0x007f);  // 7 bits
      p = (u_int)(getpid() & 0x07ff);        // 11 bits
      t = (time(NULL) & 0x3fff);             // 14 bits
    }
    r = (s << 25) + (p << 14) + t;
    // printf("s=%u (%x), p=%u (%x), t=%u (%x)\n",s,s,p,p,t,t);
    // printf("%08x\n%08x\n%08x\n%08x\n",s<<25,p<<14,t,r);
    for (int i=0; i<num; i++) {
      r = (r*1664525) + 1013904223; // printf("%08x\n",r);
      [array addObject:[NSNumber numberWithUnsignedInt:r]];
    }
  }
  // printf("array.count=%lu\n",[array count]);
}

/* ---------------------------------------------------------------- */

@implementation MTwister
{
  unsigned int *mt_state;
  volatile int  mt_index;
  NSLock *instanceLock;
}

/* ---------------------------------------------------------------- */
#pragma mark "Class"
/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

#if (DEBUG > 1)
+ (void)initialize
{
#if (DEBUG > 0)
  fprintf(stderr,"%s:%p: class=%p\n",PREFUN,self,[MTwister class]);
#endif

  if (self != [MTwister class]) return;  // avoid call by sub-class
}
#endif

/* --- */

#if (DEBUG > 1)                          // called before main()
+ (void)load
{
#if (DEBUG > 0)
  fprintf(stderr,"%s:%p\n",PREFUN,self);
#endif
}
#endif

/* ---------------------------------------------------------------- */

+ (MTwister*)sharedTwister
{
  if (!classTwister) {
    classTwister = [[MTwister alloc] initWithLock:YES];
  }
  return classTwister;
}

/* ---------------------------------------------------------------- */

+ (u_int)hash32FromFile:(const char*)file
{
  u_int r;
  
  MTwister *twister = [[MTwister alloc] initWithFile:file];
  r = [twister uintValue];
  [twister release];

  return r;
}

/* ---------------------------------------------------------------- */

+ (u_int)hash32FromChars:(const char*)c
{
  u_int r;
  
  MTwister *twister = [[MTwister alloc] initWithBytes:c length:(u_int)strlen(c)];
  r = [twister uintValue];
  [twister release];

  return r;
}

/* ---------------------------------------------------------------- */

+ (int)intValue:(int)x
{
  return [[self sharedTwister] intValue:x];
}

/* ---------------------------------------------------------------- */

+ (double)doubleValue:(double)x
{
  return [[self sharedTwister] doubleValue:x];
}

/* ---------------------------------------------------------------- */

+ (double)gaussMean:(double)mean sigma:(double)sigma
{
  return [[self sharedTwister] gaussMean:mean sigma:sigma];
}

/* ---------------------------------------------------------------- */
#pragma mark "Instance"
/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

- (id)init
{
  return [self initWithLock:NO];
}

/* --- */

- (id)initWithLock:(BOOL)useLock
{
  return [self initWithLock:useLock urand:0];
}

/* --- */

- (id)initWithLock:(BOOL)useLock urand:(int)n
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%d,%d,%p)\n",PREFUN,(int)useLock,n,self);
#endif
  assert(sizeof(unsigned int) == 4);

  if (!(self = [super init])) return nil;

  mt_state = (unsigned int*)malloc(MT_N*sizeof(unsigned int));
  
  if (useLock) instanceLock = [[NSLock alloc] init];

  NSMutableArray *key = [NSMutableArray arrayWithCapacity:(8+n)];
  addHashedUniqueString(key);
  if (n > 0) addRandomNumbers(key,n);
  assert([key count] > 0);

  [self setupState:key];
  assert(mt_index == 0);
  
  return self;
}

/* ---------------------------------------------------------------- */

- (id)initWithFile:(const char*)file
{
  NSMutableArray *key;
  FILE *fp;
#if (DEBUG > 0)
  fprintf(stderr,"%s(%s,%p)\n",PREFUN,file,self);
#endif
  assert(sizeof(unsigned int) == 4);

  if (!(self = [super init])) return nil;

  mt_state = (unsigned int*)malloc(MT_N*sizeof(unsigned int));

  key = [NSMutableArray arrayWithCapacity:512];

  [key addObject:[NSNumber numberWithUnsignedInt:SAFETY_NET]];
  if ((fp = fopen(file,"r")) != NULL) { u_int value=0;
    while (fread(&value,1,sizeof(u_int),fp) > 0) {
      [key addObject:[NSNumber numberWithUnsignedInt:value]];
      value = 0;
    }
    fclose(fp);
  }

  [self setupState:key];
  assert(mt_index == 0);
  
  return self;
}

/* ---------------------------------------------------------------- */

- (id)initWithBytes:(const char*)bytes length:(u_int)len
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p,%u)\n",PREFUN,bytes,len);
#endif
  assert(sizeof(unsigned int) == 4);

  if ((self = [super init]) == nil) return nil;

  mt_state = (unsigned int*)malloc(MT_N*sizeof(unsigned int));

  [self setupWithBytes:bytes length:len];
  assert(mt_index == 0);
  
  return self;
}

/* --- */

- (void)setupWithBytes:(const char*)bytes length:(u_int)len
{
  NSNumber *num;
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p,%u)\n",PREFUN,bytes,len);
#endif
  NSMutableArray *key = [NSMutableArray arrayWithCapacity:len];

  [key addObject:[NSNumber numberWithUnsignedInt:SAFETY_NET]];
  for (u_int u=0; u<len; u++) {
    if      (u+3 < len) { num=[NSNumber numberWithUnsignedInt:*(u_int*)&bytes[u]]; u+=3; }
    else if (u+1 < len) { num=[NSNumber numberWithUnsignedInt:*(u_short*)&bytes[u]]; u+=1; }
    else                { num=[NSNumber numberWithUnsignedInt:(u_int)bytes[u]]; }
    [key addObject:num];
  }

  [self setupState:key];
  assert(mt_index == 0);
}

/* ---------------------------------------------------------------- */

- (void)setupState:(NSArray*)key
{
  int i,j,k,key_len;
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p,%u,%p)\n",PREFUN,key,(u_int)[key count],self);
#endif

  mt_state[0]= 19650218 & 0xffffffff;
  for (j=1; j<MT_N; j++) {
    mt_state[j] = (1812433253 * (mt_state[j-1] ^ (mt_state[j-1] >> 30)) + j); 
    /* See Knuth TAOCP Vol2. 3rd Ed. P.106 for multiplier. */
    /* In the previous versions, MSBs of the seed affect   */
    /* only MSBs of the array state[].                     */
    /* 2002-01-09 modified by Makoto Matsumoto             */
  }

  i=1; j=0; key_len = (int)[key count];
  k = (MT_N>key_len ? MT_N : key_len);
  for (; k; k--) {
    mt_state[i] = (mt_state[i] ^ ((mt_state[i-1] ^ (mt_state[i-1] >> 30)) * 1664525))
            + [[key objectAtIndex:j] unsignedIntValue] + j;  /* non linear */
    i++; j++;
    if (i >= MT_N) { mt_state[0] = mt_state[MT_N-1]; i=1; }
    if (j >= key_len) j=0;
  }
  for (k=MT_N-1; k; k--) {    /* non linear */
    mt_state[i] = (mt_state[i] ^ ((mt_state[i-1] ^ (mt_state[i-1] >> 30)) * 1566083941)) - i;
    i++;
    if (i>=MT_N) { mt_state[0] = mt_state[MT_N-1]; i=1; }
  }
  mt_state[0] = 0x80000000;   /* MSB is 1; assuring non-zero initial array */

  [self shuffle];
  assert(mt_index == 0);
}

/* ---------------------------------------------------------------- */

- (void)shuffle
{
  auto  int          kk;
  auto  unsigned int y;
  const unsigned int mag01[2]={0x0,MATRIX_A};
#if (DEBUG > 2)
  fprintf(stderr,"%s(%p)\n",PREFUN,self);
#endif

  mt_index = 0;

  for (kk=0; kk<MT_N-MT_M; kk++) {
    // assert(kk < MT_N-1);
    y = (mt_state[kk]&UPPER_MASK) | (mt_state[kk+1]&LOWER_MASK);
    // assert(kk+MT_M < MT_N);
    mt_state[kk] = mt_state[kk+MT_M] ^ (y >> 1) ^ mag01[y & 0x1];
  }
  for ( ; kk<MT_N-1; kk++) {
    // assert(kk < MT_N-1);
    y = (mt_state[kk]&UPPER_MASK) | (mt_state[kk+1]&LOWER_MASK);
    // assert(kk+MT_M-MT_N < MT_N);
    mt_state[kk] = mt_state[kk+(MT_M-MT_N)] ^ (y >> 1) ^ mag01[y & 0x1];
  }
  y = (mt_state[MT_N-1]&UPPER_MASK) | (mt_state[0]&LOWER_MASK);
  mt_state[MT_N-1] = mt_state[MT_M-1] ^ (y >> 1) ^ mag01[y & 0x1];

  assert(mt_index == 0);
}

/* ---------------------------------------------------------------- */

- (unsigned int)mt_19937
{
  int          i;
  unsigned int y;

  if (instanceLock) [instanceLock lock]; // 10% faster
  while ((i = mt_index++) >= MT_N) { [self shuffle]; }
  if (instanceLock) [instanceLock unlock];
  
  assert(i < MT_N);
  y = mt_state[i];

  y ^= (y >> 11);                      /* tempering */
  y ^= (y <<  7) & 0x9d2c5680;
  y ^= (y << 15) & 0xefc60000;
  y ^= (y >> 18);

  return y;
}

/* ---------------------------------------------------------------- */

- (int)intValue:(int)x                 // execTime = 0.025 usec
{
  assert(x > 0);
#if (DEBUG > 0) && !defined(NDEBUG)    // Warning: higher values less likely
  if (x > 0x7fff) NSLog(@"%s: modulus too large (%d)",PREFUN,x);
#endif
  return (int)([self mt_19937] % x);
}

/* ---------------------------------------------------------------- */

- (u_int)uintValue
{
  return [self mt_19937];
}

/* ---------------------------------------------------------------- */

- (u_int)secureUint                  // execTime = 0.5 usec
{
  u_int uvals[SECURE_FACTOR];
  assert(SECURE_FACTOR*sizeof(u_int) >= CC_SHA256_DIGEST_LENGTH);

  for (int i=0; i<SECURE_FACTOR; i++) uvals[i] = [self mt_19937];
  hashBytes((u_char*)uvals,SECURE_FACTOR*sizeof(u_int));

  return uvals[0];
}


/* ---------------------------------------------------------------- */

- (double)doubleValue:(double)x
{
  return (x * (double)[self mt_19937] * (1.0/4294967296.0));
}

/* --- */

- (double)quadValue:(double)x
{
  u_int a = [self mt_19937] >> 5;
  u_int b = [self mt_19937] >> 6;
 
   return (x * ((double)a*67108864.0+(double)b) * (1.0/9007199254740992.0));
}

/* ---------------------------------------------------------------- */

- (double)gaussMean:(double)mean sigma:(double)sigma
{
  double x,y,r;

  do {
    x = [self doubleValue:2.0]-1.0;
    y = [self doubleValue:2.0]-1.0;
    r = x*x + y*y;
  } while (r >= 1.0);

  return (mean+(sigma*x*sqrt(-2.0*log(r)/r)));
}

/* --- */

- (double)gaussQuad:(double)mean sigma:(double)sigma
{
  double x,y,r;

  do {
    x = [self quadValue:2.0]-1.0;
    y = [self quadValue:2.0]-1.0;
    r = x*x + y*y;
  } while (r >= 1.0);

  return (mean+(sigma*x*sqrt(-2.0*log(r)/r)));
}

/* ---------------------------------------------------------------- */

- (void)dealloc
{
#if (DEBUG > 1)
  fprintf(stderr,"%s:%p\n",PREFUN,self);
#endif

  [instanceLock release];

  if (mt_state) free((void*)mt_state);
  
  [super dealloc];
}

/* ---------------------------------------------------------------- */

@end

/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
