/* ---------------------------------------------------------------- */
//
// FFmodel.m
// DCU
//
// Created by Christoph Birk on 2013-03-15
// Copyright 2013 Carnegie Observatories. All rights reserved.
//
/* ---------------------------------------------------------------- */

#define DEBUG      0                   // debug level

/* ---------------------------------------------------------------- */

#import "FFmodel.h"
#import "Notice.h"
#import "AppDelegate.h"

/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

@interface FFbase ()                   // private class extension
@property (readwrite,nonatomic) int status;
@end

@implementation FFbase

@synthesize name,status;

/* --- */

+ (NSString*)errorString:(int)err      // b0201
{
  NSString *string=nil;
  
  switch (err) {
    case E_FF_error: string = @"FFmodel: error"; break;
    case E_FF_coil: string = @"DCU: coil/interlock"; break;
    case E_FF_lamp: string = @"DCU: lamp output not set"; break;
    case E_FF_varLamp: string = @"DCU: varLamp dimmer not set"; break;
    case E_FF_brake: string = @"DCU: brake sensor failed"; break;
    case E_FF_screen: string = @"DCU: screen sensor failed"; break;
  }
  return string;
}

/* --- */

- (id)initWithPLC:(PLC_IO*)con Name:(NSString*)nam;
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p,%s)\n",PREFUN,con,[nam UTF8String]);
#endif
  if (!(self = [super init])) return nil;

  connection = [con retain];
  assert(nam);
  name = [nam retain];
  
  return self;
}

/* ---------------------------------------------------------------- */

- (void)dealloc
{
#if (DEBUG > 2)
  fprintf(stderr,"%s(%s)\n",PREFUN,[name UTF8String]);
#endif
  [connection release];
  [name release];
  [super dealloc];
}

@end

/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

@implementation FFcoil

- (id)initWithPLC:(PLC_IO*)con XML:(NSDictionary*)dict prefix:(NSString*)pref
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p,%s)\n",PREFUN,dict,[pref UTF8String]);
#endif

  if (!(self = [super initWithPLC:con Name:pref])) return nil;

  assert(dict);
  if (dict) { NSString *key,*obj; 
    key = [NSString stringWithFormat:@"%@_Coil",pref];
    addr_coil = o2d([[dict objectForKey:key] intValue]);
    key = [NSString stringWithFormat:@"%@_Output",pref];
    obj = [dict objectForKey:key];
    addr_out = (obj) ? o2d([obj intValue]) : -1;
#if (DEBUG > 1) 
    printf("addr_coil=%d, addr_out=%d\n",addr_coil,addr_out);
#endif
  }
  
  return self;  
}

/* ---------------------------------------------------------------- */

- (int)get                             // FFcoil:
{
  int err,value=0;
#if (DEBUG > 1)
  fprintf(stderr,"%s(%s)\n",PREFUN,[name UTF8String]);
#endif
  
  value = status;                      // Simulator preset
  if (addr_out < 0) {                  // no output address
    assert(addr_coil >= 0);            // just read-back coil
    err = [connection getCoil:addr_coil status:&value];
  } else {                             // associated output (Y)
    err = [connection getOutput:addr_out value:&value];
  }
  if (!err) self.status = (value) ? 1 : 0;  // KVO compliant

  return err;
}

/* ---------------------------------------------------------------- */

- (int)set:(int)state                  // FFcoil:
{
  int err;
#if (DEBUG > 1)
  fprintf(stderr,"%s(%s,%d)\n",PREFUN,[name UTF8String],state);
#endif
  assert(addr_coil >= 0);
  
  err = [connection setCoil:addr_coil status:state];
  if (!err) self.status = state;       // KVO requires property
    
  return err;
}

@end

/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

@implementation FFinput

- (id)initWithPLC:(PLC_IO*)con XML:(NSDictionary*)dict prefix:(NSString*)pref
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p,%s)\n",PREFUN,dict,[pref UTF8String]);
#endif

  if (!(self = [super initWithPLC:con Name:pref])) return nil;

  assert(dict);
  if (dict) { NSString *key; 
    key = [NSString stringWithFormat:@"%@_Input",pref];
    addr_input = o2d([[dict objectForKey:key] intValue]);
#if (DEBUG > 1)
    printf("addr_input=%d\n",addr_input);
#endif
  }
  
  return self;  
}

/* ---------------------------------------------------------------- */

- (int)get                             // FFinput:
{
  int err,value;
#if (DEBUG > 1)
  fprintf(stderr,"%s(%s)\n",PREFUN,[name UTF8String]);
#endif
  assert(addr_input >= 0);

  value = status;                      // Simulator preset
  err = [connection getInput:addr_input value:&value];
  if (!err) self.status = (value) ? 1 : 0; // KVO compliant
  
  return err;
}

- (int)get:(int)value                  // FFinput:
{
#if (DEBUG > 2)
  fprintf(stderr,"%s(%s,%d)\n",PREFUN,[name UTF8String],value);
#endif
  status = value;                      // Simulator preset
  return [self get];
}

@end

/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

@implementation FFsensor 

- (int)get
{
  int err,value;
#if (DEBUG > 1)
  fprintf(stderr,"%s(%s)\n",PREFUN,[name UTF8String]);
#endif
  assert(addr_input >= 0);

  value = status;                       // Simulator preset
  err = [connection getValue:addr_input value:&value];
  if (!err) self.status = value;        // KVO compliant

  return err;
}

@end

/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */

@interface FFdevice ()                 // private class extension
@property (readwrite)  BOOL moving;
@end

@implementation FFdevice
@synthesize     request,moving;

- (int)get
{
  assert(0);
  
  self.status = request;
  
  return 0;
}

/* --- */

- (void)run:(NSNumber*)target
{
  assert(0);
  
  NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];

  request = [target intValue];         

  self.moving = TRUE;                  // KVO requires 'property'
  [NSThread sleepForTimeInterval:0.500];
  [self get];
  self.moving = FALSE;

  [pool drain];
}

@end

/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */

@implementation FFlamp

- (id)initWithPLC:(PLC_IO*)con XML:(NSDictionary*)dict prefix:(NSString*)pref
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p,%s)\n",PREFUN,dict,[pref UTF8String]);
#endif

  if (!(self = [super initWithPLC:con Name:pref])) return nil;

  coil = [[FFcoil alloc] initWithPLC:con XML:dict prefix:pref];
  
  return self;  
}

/* ---------------------------------------------------------------- */

- (int)get                             // FFlamp:
{
  int err;
#if (DEBUG > 1)
  fprintf(stderr,"%s(%s)\n",PREFUN,[name UTF8String]);
#endif
  assert(coil);
  
  err = [coil get];
  if (!err) self.status = (coil.status) ? 1 : 0; // KVO

  return err;
}

/* ---------------------------------------------------------------- */

- (void)run:(NSNumber*)target          // FFlamp:
{
  int err;
#if (DEBUG > 0)
  fprintf(stderr,"%s(%s,%d)\n",PREFUN,[name UTF8String],[target intValue]);
#endif
  NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];

  request = [target intValue];         
  self.moving = TRUE;                  // KVO requires 'property'

  err = [connection open];
  if (!err) {
    err = [coil set:request];
    if (!err) {                        // read-back output
      [NSThread sleepForTimeInterval:PLC_DELAY];
      err = [coil get];
    }
    [connection close];
  }
  if (!err) { 
    self.status = coil.status;         // KVO
    if (status != request) err = E_FF_lamp;
  }
  self.moving = FALSE;

  if (err) {                           // write to logfile NEW b0307
    dispatch_sync(dispatch_get_main_queue(),^(void) { NSString *str; // b0308
      str = [NSString stringWithFormat:@"%@: request failed (err=%d)\r%@",
               name,err,[(AppDelegate*)[NSApp delegate] errorString:err]];
      [main_logger message:[str UTF8String] level:1 file:YES];
    });
  }

#if (DEBUG > 0)
  fprintf(stderr,"%s(%p) ... done\n",PREFUN,self);
#endif
  [pool drain];
}

/* ---------------------------------------------------------------- */

- (void)dealloc                        // FFlamp:
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%s)\n",PREFUN,[name UTF8String]);
#endif

  [coil release];                      // release sub-devices
  [super dealloc];
}

@end

/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

@interface FFvarLamp ()                 // private class extension
@property (readwrite,nonatomic) int  level;
@end

@implementation FFvarLamp

@synthesize level;

- (id)initWithPLC:(PLC_IO*)con XML:(NSDictionary*)dict prefix:(NSString*)pref
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p,%s)\n",PREFUN,dict,[pref UTF8String]);
#endif

  if (!(self = [super initWithPLC:con XML:dict prefix:pref])) return nil;
  assert(coil);
  
  assert(dict);
  if (dict) { NSString *key; 
    key = [NSString stringWithFormat:@"%@_Analog",pref];
    addr_analog = o2d([[dict objectForKey:key] intValue]);
#if (DEBUG > 1)
    printf("addr_analog=%d\n",addr_analog);
#endif
  }
  
  return self;  
}

/* ---------------------------------------------------------------- */

- (int)get                             // FFvarLamp:
{
  int err;
#if (DEBUG > 0)
  fprintf(stderr,"%s(%s)\n",PREFUN,[name UTF8String]);
#endif

  err = [connection open];
  if (!err) {
    err = [super get];                 // get coil/output
    if (!err) { int value=request;     // Simulator preset
      err = [connection getValue:addr_analog value:&value];
      if (!err) self.level = value;    // KVO      
    }
    [connection close];
  }

  return err;
}

/* ---------------------------------------------------------------- */

- (void)run:(NSNumber*)target          // FFvarLamp:
{
  int err;
#if (DEBUG > 0)
  fprintf(stderr,"%s(%s,%d)\n",PREFUN,[name UTF8String],[target intValue]);
#endif
  NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];

  request = [target intValue];
  self.moving = TRUE;                  // KVO requires 'property'

  err = [connection open];
  if (!err) {
    if (!err) err = [coil set:((request) ? 1 : 0)]; // set coil
    if (!err) err = [connection setValue:addr_analog value:request]; // set dimmer
    if (!err) {
      [NSThread sleepForTimeInterval:PLC_DELAY];
      err = [self get];
    }
    [connection close];
  }
  if (!err) { 
    if (status != ((request) ? 1 : 0)) err = E_FF_lamp;
    if (level != request) err = E_FF_varLamp;
  }
  self.moving = FALSE;  

  if (err) {                            // write to logfile NEW b0307
    dispatch_sync(dispatch_get_main_queue(),^(void) { NSString *str; // b0308
      str = [NSString stringWithFormat:@"%@: request failed (err=%d)\r%@",
               name,err,[(AppDelegate*)[NSApp delegate] errorString:err]];
      [main_logger message:[str UTF8String] level:1 file:YES];
    });
  }

#if (DEBUG > 0)
  fprintf(stderr,"%s(%p) ... done\n",PREFUN,self);
#endif
  [pool drain];
}

@end

/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

@implementation FFbrake  

@synthesize pswitch;

- (id)initWithPLC:(PLC_IO*)con XML:(NSDictionary*)dict prefix:(NSString*)pref
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p,%s)\n",PREFUN,dict,[pref UTF8String]);
#endif

  if (!(self = [super initWithPLC:con Name:pref])) return nil;

  coil = [[FFcoil alloc] initWithPLC:con XML:dict prefix:pref];
  pswitch = [[FFinput alloc] initWithPLC:con XML:dict prefix:pref];

  return self;  
}

/* ---------------------------------------------------------------- */

- (int)get                             // FFbrake:
{
  int err,value;
#if (DEBUG > 1)
  fprintf(stderr,"%s(%s)\n",PREFUN,[name UTF8String]);
#endif
  assert(pswitch);

  value = (status) ? 1 : 0;            // Simulator preset
  err = [pswitch get:value];
  if (!err) self.status = (pswitch.status) ? 1 : 0;  // KVO

  return err;
}

/* ---------------------------------------------------------------- */

- (void)run:(NSNumber*)target          // FFbrake:
{
  int err;
#if (DEBUG > 0)
  fprintf(stderr,"%s(%s,%d)\n",PREFUN,[name UTF8String],[target intValue]);
#endif
  NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];

  request = [target intValue];
  self.moving = TRUE;                  // KVO requires 'property'

  err = [connection open];
  if (!err) {
    err = [coil set:request];          // set coil
    if (!err) {                        // read-back output
      [NSThread sleepForTimeInterval:PLC_DELAY];
      err = [coil get];
      if (!err) {  
        if (coil.status != request) {  // verify brake sensor v0.802
          err = E_FF_coil;
          [coil set:coil.status];      // reset coil    
        }
      }
    }
    if (!err) {
      [NSThread sleepForTimeInterval:1.00];
      for (int i=0; i<9; i++) {        // timeout = 10 seconds
        [NSThread sleepForTimeInterval:1.00];
        if (!connection.online) status = request; // Simulator preset
        err = [self get]; if (err) break;
        if (status == request) break;
      }      
      if (status != request) [coil set:BRAKE_ON]; // Note: brake coil off
    }
    [connection close];
  }
  if (!err) { 
    if (status != request) err = E_FF_brake;
  }
  self.moving = FALSE;  

  if (err) {                          // write to logfile NEW b0307
    dispatch_sync(dispatch_get_main_queue(),^(void) { NSString *str; // b0308
      str = [NSString stringWithFormat:@"%@: move failed (err=%d)\r%@",
               name,err,[(AppDelegate*)[NSApp delegate] errorString:err]];
      [main_logger message:[str UTF8String] level:1 file:YES];
    });
  }

#if (DEBUG > 0)
  fprintf(stderr,"%s(%p) ... done\n",PREFUN,self);
#endif
  [pool drain];
}

/* ---------------------------------------------------------------- */

- (void)dealloc                        // FFbrake:
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%s)\n",PREFUN,[name UTF8String]);
#endif
  [pswitch release];
  [coil release];
  [super dealloc];
}

@end

/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

@implementation FFscreen

@synthesize ins_limit,ret_limit;
@synthesize brake; 

- (id)initWithPLC:(PLC_IO*)con XML:(NSDictionary*)dict prefix:(NSString*)pref
{
  NSString *pfx;
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p,%p)\n",PREFUN,dict,pref);
#endif

  request = SCREEN_RET;

  if (!(self = [super initWithPLC:con Name:pref])) return nil;

  pfx = [NSString stringWithFormat:@"%@_INS",pref];
  ins_coil  = [[FFcoil alloc] initWithPLC:con XML:dict prefix:pfx];
  ins_limit = [[FFinput alloc] initWithPLC:con XML:dict prefix:pfx];
  
  pfx = [NSString stringWithFormat:@"%@_RET",pref];
  ret_coil  = [[FFcoil alloc] initWithPLC:con XML:dict prefix:pfx];
  ret_limit = [[FFinput alloc] initWithPLC:con XML:dict prefix:pfx];
  
  pfx = [NSString stringWithFormat:@"%@_BRAKE",pref];
  brake     = [[FFbrake alloc] initWithPLC: con XML:dict prefix:pfx];
  
  if (!connection.online) status = SCREEN_RET;  // Simulator
  
  return self;  
}

/* ---------------------------------------------------------------- */

- (int)get                             // FFscreen:
{
  int err=0;
#if (DEBUG > 1)
  fprintf(stderr,"%s(%s)\n",PREFUN,[name UTF8String]);
#endif

  err = [connection open]; if (err) return err;

  if (!err) err = [ins_limit get:((status & SCREEN_INS) ? 1 : 0)];
  if (!err) err = [ret_limit get:((status & SCREEN_RET) ? 1 : 0)];
#if (DEBUG > 2)
  printf("inslim=%d, retlim=%d\n",ins_limit.status,ret_limit.status);
#endif
  status = 0;
  if (!err) {
    if (ins_limit.status) status |= SCREEN_INS;
    if (ret_limit.status) status |= SCREEN_RET;
  }
  if (!err) self.status = status;      // KVO
  if (!err) err = [brake get]; 
#if (DEBUG > 2)
  printf("brake=%d\n",brake.status);
#endif
  [connection close];
  
  return err;
}

/* ---------------------------------------------------------------- */

- (void)run:(NSNumber*)target          // FFscreen:
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%s,%d)\n",PREFUN,[name UTF8String],[target intValue]);
#endif
  NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];

  request = [target intValue];
  status = 0;                          // transition
  if (!connection.online) [self get];  // Simulator
  
  self.moving = TRUE;                  // KVO requires 'property'
  int err = [connection open];
  if (!err) { NSNumber *pos;
    if (!err) err = [ins_coil set:1];  // open both valves
    if (!err) err = [ret_coil set:1];
    if (!err) {                        // readback output
      [NSThread sleepForTimeInterval:PLC_DELAY];
      if (!err) err = [ins_coil get];
      if (!err) err = [ret_coil get];
      if (!err) {                      // verify coils vs. output v0.802
        if ((ins_coil.status != 1) || (ret_coil.status != 1)) err = E_FF_coil;
      }
    }
    if (!err) {                        // open brake
      [NSThread sleepForTimeInterval:3.00]; // v0.900 
      pos = [NSNumber numberWithInt:BRAKE_OFF]; [brake run:pos]; 
      if (brake.status != BRAKE_OFF) err = E_FF_brake;
    }
    if (!err) {                        // close 'other' valve
      [NSThread sleepForTimeInterval:0.50];
      if (request == SCREEN_RET) err = [ins_coil set:0];
      else                       err = [ret_coil set:0];
    }
    if (!err) {
      [NSThread sleepForTimeInterval:4.00];
      for (int i=0; i<18; i++) {       // timeout = 40 seconds
        [NSThread sleepForTimeInterval:2.00];
        if (!connection.online) status = request; // Simulator preset
        err = [self get]; if (err) break;
        if (status == request) {       // we are there
          if (connection.online) sleep(10); // wait to finish move
          break;
        }
      }      
    }
    pos = [NSNumber numberWithInt:BRAKE_ON]; [brake run:pos];
    if (!err) { 
      if (brake.status != BRAKE_ON) err = E_FF_brake;
    }
    [ret_coil set:0]; [ins_coil set:0]; // close both valves
    [connection close];
  }
  if (!err) { 
    if (status != request) err = E_FF_screen; 
  }
  self.moving = FALSE;
  
  if (err) {                      // write to logfile NEW b0307
    dispatch_sync(dispatch_get_main_queue(),^(void) { NSString *str; // b0308
      str = [NSString stringWithFormat:@"%@: move failed (err=%d)\r%@",
               name,err,[(AppDelegate*)[NSApp delegate] errorString:err]];
      [main_logger message:[str UTF8String] level:1 file:YES];
    });
  }

#if (DEBUG > 0)
  fprintf(stderr,"%s(%p) ... done\n",PREFUN,self);
#endif
  [pool drain];
}

/* ---------------------------------------------------------------- */

- (void)dealloc
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%s)\n",PREFUN,[name UTF8String]);
#endif

  [ins_limit release];       // release sub-devices
  [ret_limit release];
  [ins_coil release];
  [ret_coil release];
  [brake release];

  [super dealloc];
}

@end

/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
