/* ---------------------------------------------------------------- */
//
// MTwister.m
//
// Created by Christoph Birk on 2013-04-12
// Copyright 2013 Carnegie Observatories. All rights reserved.
//
// Algorithm: Makoto Matsumoto and Takuji Nishimura (2002 version)
//            http://www.math.sci.hiroshima-u.ac.jp/~m-mat/MT/emt.html
//
// 2014-04-01  use [NSProcessInfo globallyUniqueString]
// 2014-04-18  use 'instanceLock' on defaultTwister
// 2014-10-08  -initWithBytes:Len:
// 2018-01-03  create 'classTwister' on demand (was @ initialize)
// 2018-11-01  move iVARs into implementation
// 2020-11-16  replace /dev/urandom by getentropy(),arc4random()
//             avoid 'modulo bias' when creating 'intValue'
//
/* ---------------------------------------------------------------- */

#ifndef DEBUG
#define DEBUG           0              // debug level
#endif

#define MT_M            397
#define MT_N            624
#define MATRIX_A        0x9908b0df     // constant vector
#define UPPER_MASK      0x80000000     // most significant w-r bits
#define LOWER_MASK      0x7fffffff     // least significant r bits

#define SAFETY_NET      0x5afeb10c     // safe bloc[k]
#define SECURE_FACTOR   8              // secure uint

/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

#import  "MTwister.h"

#include <CommonCrypto/CommonDigest.h>
#include <sys/random.h>

/* ---------------------------------------------------------------- */

static MTwister *classTwister=nil;

/* --- */

u_int* hashBytes(u_char* bytes,int size) // execTime=0.5 usec
{
  assert(size >= CC_SHA256_DIGEST_LENGTH);
  CC_SHA256_CTX c;
  CC_SHA256_Init(&c);
  CC_SHA256_Update(&c,bytes,size);
  CC_SHA256_Final(bytes,&c);
  return (u_int*)bytes;
}

/* --- */

void addHashedUniqueString(NSMutableArray* array) // execTime = 7.5 usec
{
  CC_SHA256_CTX c;
  unichar u;
  u_char md[CC_SHA256_DIGEST_LENGTH];
  
  NSString *unique = [[NSProcessInfo processInfo] globallyUniqueString];
  CC_SHA256_Init(&c);
  for (int i=0; i<[unique length]; i++) {
    u = [unique characterAtIndex:i];  // printf("%08x\n",ubuf[i]);
    CC_SHA256_Update(&c,&u,sizeof(unichar));
  }
  CC_SHA256_Final(md,&c);

  u_int *data = (u_int*)md;
  for (int i=0; i<CC_SHA256_DIGEST_LENGTH/sizeof(u_int); i++) {
    [array addObject:[NSNumber numberWithUnsignedInt:data[i]]];
  }
}

/* --- */

void addRandomNumbers(NSMutableArray* array,int num)
{
  u_int r;
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p,%d)\n",__func__,array,num);
#endif

  for (int i=0; i<num; i++) {
    if (getentropy(&r,sizeof(r)) < 0) r = arc4random();
    [array addObject:@(r)];
  }
}

/* ---------------------------------------------------------------- */

@implementation MTwister
{
  unsigned int *mt_state;
  volatile int  mt_index;
  NSLock *instanceLock;
}

/* ---------------------------------------------------------------- */
#pragma mark "Class"
/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

#if (DEBUG > 1)
+ (void)initialize
{
#if (DEBUG > 0)
  fprintf(stderr,"%s:%p: class=%p\n",__func__,self,[MTwister class]);
#endif

  if (self != [MTwister class]) return;  // avoid call by sub-class
}
#endif

/* --- */

#if (DEBUG > 1)                          // called before main()
+ (void)load
{
#if (DEBUG > 0)
  fprintf(stderr,"%s:%p\n",__func__,self);
#endif
}
#endif

/* ---------------------------------------------------------------- */

+ (MTwister*)sharedTwister
{
  if (!classTwister) {
    classTwister = [[MTwister alloc] initWithLock:YES];
  }
  return classTwister;
}

/* ---------------------------------------------------------------- */

+ (u_int)hash32FromFile:(const char*)file
{
  u_int r;
  
  MTwister *twister = [[MTwister alloc] initWithFile:file];
  r = [twister uintValue];
  [twister release];

  return r;
}

/* ---------------------------------------------------------------- */

+ (u_int)hash32FromChars:(const char*)c
{
  u_int r;
  
  MTwister *twister = [[MTwister alloc] initWithBytes:c length:(u_int)strlen(c)];
  r = [twister uintValue];
  [twister release];

  return r;
}

/* ---------------------------------------------------------------- */

+ (int)intValue:(int)x
{
  return [[self sharedTwister] intValue:x];
}

/* ---------------------------------------------------------------- */

+ (double)doubleValue:(double)x
{
  return [[self sharedTwister] doubleValue:x];
}

/* ---------------------------------------------------------------- */

+ (double)gaussMean:(double)mean sigma:(double)sigma
{
  return [[self sharedTwister] gaussMean:mean sigma:sigma];
}

/* ---------------------------------------------------------------- */
#pragma mark "Instance"
/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

- (instancetype)init
{
  return [self initWithLock:NO];
}

/* --- */

- (instancetype)initWithLock:(BOOL)useLock
{
  return [self initWithLock:useLock urand:0];
}

/* --- */

- (instancetype)initWithLock:(BOOL)useLock urand:(int)n
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%d,%d,%p)\n",__func__,(int)useLock,n,self);
#endif
  assert(sizeof(unsigned int) == 4);

  if (!(self = [super init])) return nil;

  mt_state = (unsigned int*)malloc(MT_N*sizeof(unsigned int));
  
  if (useLock) instanceLock = [[NSLock alloc] init];

  NSMutableArray *key = [NSMutableArray arrayWithCapacity:(8+n)];
  addHashedUniqueString(key);
  if (n > 0) addRandomNumbers(key,n);
  assert([key count] > 0);

  [self setupState:key];
  assert(mt_index == 0);
  
  return self;
}

/* ---------------------------------------------------------------- */

- (instancetype)initWithFile:(const char*)file
{
  NSMutableArray *key;
  FILE *fp;
#if (DEBUG > 1)
  fprintf(stderr,"%s(%s,%p)\n",__func__,file,self);
#endif
  assert(sizeof(unsigned int) == 4);

  if (!(self = [super init])) return nil;

  mt_state = (unsigned int*)malloc(MT_N*sizeof(unsigned int));

  key = [NSMutableArray arrayWithCapacity:512];
  [key addObject:[NSNumber numberWithUnsignedInt:SAFETY_NET]];
  if ((fp = fopen(file,"r")) != NULL) { u_int value=0;
    while (fread(&value,1,sizeof(u_int),fp) > 0) {
      [key addObject:[NSNumber numberWithUnsignedInt:value]];
      value = 0;
    }
    fclose(fp);
  }

  [self setupState:key];
  assert(mt_index == 0);
  
  return self;
}

/* ---------------------------------------------------------------- */

- (instancetype)initWithBytes:(const char*)bytes length:(u_int)len
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p,%u)\n",__func__,bytes,len);
#endif
  assert(sizeof(unsigned int) == 4);

  if ((self = [super init]) == nil) return nil;

  mt_state = (unsigned int*)malloc(MT_N*sizeof(unsigned int));

  [self setupWithBytes:bytes length:len];
  assert(mt_index == 0);
  
  return self;
}

/* --- */

- (void)setupWithBytes:(const char*)bytes length:(u_int)len
{
  NSNumber *num;
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p,%u)\n",__func__,bytes,len);
#endif

  NSMutableArray *key = [NSMutableArray arrayWithCapacity:len];

  [key addObject:[NSNumber numberWithUnsignedInt:SAFETY_NET]];
  for (u_int u=0; u<len; u++) {
    if      (u+3 < len) { num=[NSNumber numberWithUnsignedInt:*(u_int*)&bytes[u]]; u+=3; }
    else if (u+1 < len) { num=[NSNumber numberWithUnsignedInt:*(u_short*)&bytes[u]]; u+=1; }
    else                { num=[NSNumber numberWithUnsignedInt:(u_int)bytes[u]]; }
    [key addObject:num];
  }

  [self setupState:key];
  assert(mt_index == 0);
}

/* ---------------------------------------------------------------- */

- (void)setupState:(NSArray*)key
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p,%lu,%p)\n",__func__,key,[key count],self);
#endif

  mt_state[0]= 19650218 & 0xffffffff;
  for (int j=1; j<MT_N; j++) {
    mt_state[j] = (1812433253 * (mt_state[j-1] ^ (mt_state[j-1] >> 30)) + j); 
    /* See Knuth TAOCP Vol2. 3rd Ed. P.106 for multiplier. */
    /* In the previous versions, MSBs of the seed affect   */
    /* only MSBs of the array state[].                     */
    /* 2002-01-09 modified by Makoto Matsumoto             */
  }
  [self addEntropy:key];
  assert(mt_index == 0);
}

/* ---------------------------------------------------------------- */

- (void)addEntropy:(NSArray*)key
{
  int i=1,j=0,k,key_len;
#if (DEBUG > 0)
  fprintf(stderr,"%s(%x...)\n",__func__,[key[0] unsignedIntValue]);
#endif
  
  key_len = (int)[key count];

  if (instanceLock) [instanceLock lock];
  k = (MT_N > key_len) ? MT_N : key_len;
  for (; k; k--) {
    mt_state[i] = (mt_state[i] ^ ((mt_state[i-1] ^ (mt_state[i-1] >> 30)) * 1664525))
            + [key[j] unsignedIntValue] + j;  /* non linear */
    i++; j++;
    if (i >= MT_N) { mt_state[0] = mt_state[MT_N-1]; i=1; }
    if (j >= key_len) j=0;
  }
  for (k=MT_N-1; k; k--) {    /* non linear */
    mt_state[i] = (mt_state[i] ^ ((mt_state[i-1] ^ (mt_state[i-1] >> 30)) * 1566083941)) - i;
    i++;
    if (i>=MT_N) { mt_state[0] = mt_state[MT_N-1]; i=1; }
  }
  mt_state[0] = 0x80000000;   /* MSB is 1; assuring non-zero initial array */
  [self shuffle];
  if (instanceLock) [instanceLock unlock];
}

/* ---------------------------------------------------------------- */

- (void)shuffle
{
  auto  int          kk;
  auto  unsigned int y;
  const unsigned int mag01[2]={0x0,MATRIX_A};
#if (DEBUG > 2)
  fprintf(stderr,"%s(%p)\n",__func__,self);
#endif

  mt_index = 0;

  for (kk=0; kk<MT_N-MT_M; kk++) {
    // assert(kk < MT_N-1);
    y = (mt_state[kk]&UPPER_MASK) | (mt_state[kk+1]&LOWER_MASK);
    // assert(kk+MT_M < MT_N);
    mt_state[kk] = mt_state[kk+MT_M] ^ (y >> 1) ^ mag01[y & 0x1];
  }
  for ( ; kk<MT_N-1; kk++) {
    // assert(kk < MT_N-1);
    y = (mt_state[kk]&UPPER_MASK) | (mt_state[kk+1]&LOWER_MASK);
    // assert(kk+MT_M-MT_N < MT_N);
    mt_state[kk] = mt_state[kk+(MT_M-MT_N)] ^ (y >> 1) ^ mag01[y & 0x1];
  }
  y = (mt_state[MT_N-1]&UPPER_MASK) | (mt_state[0]&LOWER_MASK);
  mt_state[MT_N-1] = mt_state[MT_M-1] ^ (y >> 1) ^ mag01[y & 0x1];

  assert(mt_index == 0);
}

/* ---------------------------------------------------------------- */

- (unsigned int)mt_19937
{
  int          i;
  unsigned int y;

  if (instanceLock) [instanceLock lock]; // 10% faster
  while ((i = mt_index++) >= MT_N) { [self shuffle]; }
  assert(i < MT_N);
  y = mt_state[i];
  if (instanceLock) [instanceLock unlock];
  
  y ^= (y >> 11);                      /* tempering */
  y ^= (y <<  7) & 0x9d2c5680;
  y ^= (y << 15) & 0xefc60000;
  y ^= (y >> 18);

  return y;
}

/* ---------------------------------------------------------------- */

/* Credits to David Mazieres, Damien Miller */

- (int)intValue:(int)upper_bound       // execTime = 0.7 usec
{
  if (upper_bound < 2) return 0;

  int zeros = __builtin_clz(upper_bound);
  int bits = CHAR_BIT * sizeof(u_int) - zeros;
  u_int mask = 0xFFFFFFFFU >> zeros;   // find smallest 2**n -1 >= upper_bound
  u_int result,value;

  do {
    value = [self mt_19937];           // If low 2**n-1 bits satisfy the
    result = value & mask;             //   requested condition
    if (result < upper_bound) break;   // then return result
    int bits_left = zeros;             // otherwise consume remaining bits of
    while (bits_left >= bits) {        //   randomness looking for a
      value >>= bits;                  //   satisfactory result
      result = value & mask;
      if (result < upper_bound) return result;
      bits_left -= bits;
    }
  } while (1);

  return result;
}

/* ---------------------------------------------------------------- */

- (u_int)uintValue
{
  return [self mt_19937];
}

/* ---------------------------------------------------------------- */

- (u_int)secureUint                  // execTime = 0.5 usec
{
  u_int uvals[SECURE_FACTOR];
  assert(SECURE_FACTOR*sizeof(u_int) >= CC_SHA256_DIGEST_LENGTH);

  for (int i=0; i<SECURE_FACTOR; i++) uvals[i] = [self mt_19937];
  hashBytes((u_char*)uvals,SECURE_FACTOR*sizeof(u_int));

  return uvals[0];
}

/* ---------------------------------------------------------------- */

- (double)doubleValue:(double)x
{
  return (x * (double)[self mt_19937] * (1.0/4294967296.0));
}

/* --- */

- (double)quadValue:(double)x
{
  u_int a = [self mt_19937] >> 5;
  u_int b = [self mt_19937] >> 6;
 
   return (x * ((double)a*67108864.0+(double)b) * (1.0/9007199254740992.0));
}

/* ---------------------------------------------------------------- */

- (double)gaussMean:(double)mean sigma:(double)sigma
{
  double x,y,r;

  do {
    x = [self doubleValue:2.0]-1.0;
    y = [self doubleValue:2.0]-1.0;
    r = x*x + y*y;
  } while (r >= 1.0);

  return (mean+(sigma*x*sqrt(-2.0*log(r)/r)));
}

/* --- */

- (double)gaussQuad:(double)mean sigma:(double)sigma
{
  double x,y,r;

  do {
    x = [self quadValue:2.0]-1.0;
    y = [self quadValue:2.0]-1.0;
    r = x*x + y*y;
  } while (r >= 1.0);

  return (mean+(sigma*x*sqrt(-2.0*log(r)/r)));
}

/* ---------------------------------------------------------------- */

- (void)dealloc
{
#if (DEBUG > 1)
  fprintf(stderr,"%s:%p\n",__func__,self);
#endif

  [instanceLock release];

  if (mt_state) free((void*)mt_state);
  
  [super dealloc];
}

/* ---------------------------------------------------------------- */

@end

/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
