/* ---------------------------------------------------------------- */
//
// DCUcontroller.m
// DCU
//
// Created by Christoph Birk on 2013-02-27
// Copyright 2013 Carnegie Observatories. All rights reserved.
//
/* ---------------------------------------------------------------- */

#define DEBUG              1

/* ---------------------------------------------------------------- */

#import "DCUcontroller.h"
#import "AppDelegate.h"                  // main.h
#import "Notice.h"
#import "Yutils.h"
#import "MTwister.h"

/* ---------------------------------------------------------------- */

enum DCU_error_enum { E_DCU_error=1,E_DCU_config,E_DCU_busy,
                      E_DCU_disabled,E_DCU_ret,E_DCU_air,
                      E_DCU_selector,E_DCU_failed,E_DCU_last };

/* ---------------------------------------------------------------- */

static float level2volt(int level)
{
  return (float)((24.0*level)/5000.0);
}

/* --- */

static char* get_ut_timestr(char* string,time_t ut)
{
  struct tm *now,res;

  if (ut == 0) ut = time(NULL);
  now = gmtime_r(&ut,&res);
  strftime(string,16,"%H:%M:%S",now);
  return string;
}

/* ---------------------------------------------------------------- */

@implementation DCUcontroller
{
  AppDelegate *app;
  TCPIP_Connection *tcs;
  NSArray* buttonTitles;            // NEW b0311
}

@synthesize drop_ffs,drop_mcal;
@synthesize error_state;

/* ---------------------------------------------------------------- */
#pragma mark "Class"
/* ---------------------------------------------------------------- */

+ (NSString*)errorString:(int)err   // b0201
{
  NSString *string=nil;

  switch (err) {
  case E_DCU_error: string = @"DCUcontroller: error"; break;
  case E_DCU_config: string = @"DCU: not configured"; break;
  case E_DCU_busy: string = @"DCU: busy"; break;
  case E_DCU_disabled: string = @"DCU: disabled (requires PLC reset)"; break;
  case E_DCU_ret: string = @"DCU: retract-sensor not active"; break;
  case E_DCU_air: string = @"DCU: air pressure too low"; break;
  case E_DCU_selector: string = @"DCU: selector failed (PLC interlock)"; break;
  case E_DCU_failed: string = @"DCU: sensor failed"; break;
  default:
    if (!string) string = [PLC_IO errorString:err];
    if (!string) string = [Modbus_IO errorString:err];
    if (!string) string = [FFbase errorString:err];
    if (!string) string = [TCPIP_Connection errorString:err];
    break;
  }
  
  return string;
}

/* ---------------------------------------------------------------- */
#pragma mark "Instance"
/* ---------------------------------------------------------------- */

- (id)init
{
  NSString *file;
  char     buf[128];
#if (DEBUG > 0)
  fprintf(stderr,"%s:%p\n",PREFUN,self);
#endif

  if ((self = [super init]) == nil) return nil;
  
  configured = FALSE;
  cookie_continue = YES;
  
  NSBundle *bundle = [NSBundle mainBundle];
  file = [bundle pathForResource:@"bonjour" ofType:@"png"]; // b0208
  icon_on  = [[NSImage alloc] initWithContentsOfFile:file];
  file = [bundle pathForResource:@"AlertCautionIcon" ofType:@"icns"];
  icon_err = [[NSImage alloc] initWithContentsOfFile:file];
  icon_off = nil;
  file = [bundle pathForResource:@"BurningIcon" ofType:@"icns"];
  icon_in = [[NSImage alloc] initWithContentsOfFile:file];  // b0200
  s_title = [[bundle objectForInfoDictionaryKey:@"CFBundleExecutable"] retain];
  Ystr2chr(s_title,p_title,sizeof(p_title));
  
  for (int i=0; i<N_LAMPS; i++) lamps[i] = nil;

  NSArray *users = [NSArray arrayWithObjects:@"birk",nil];
  if ([users indexOfObject:NSUserName()] == NSNotFound) {
    use_shared = YprefSetupShared();   // use Shared preferences b0204
  } else {
    use_shared = NO;
  }
  // not used (void)YappSupportCreate(&s_path,use_shared);

  main_logger = [[Logger alloc] initWithBase:nil useShared:use_shared]; // b0200

  file = [NSString stringWithFormat:@"user= %@ , sharedSetup= %d , UT= %s",
                   NSUserName(),use_shared,get_ut_timestr(buf,0)]; // NEW b0307
  [main_logger append:[file UTF8String]];

  interval_dog = YprefGetInt("watchdog_interval",60,use_shared);
  sprintf(buf,"watchdog_interval= %d",interval_dog); [main_logger append:buf];
  next_dog = time(NULL) + interval_dog;

  min_pressure = (float)YprefGetDouble("min_pressure",90.0,use_shared);
  sprintf(buf,"min_pressure= %.1f",min_pressure); [main_logger append:buf];
  air_pressure = min_pressure+5.0f;    // Simulator preset

  return self;
}

/* ---------------------------------------------------------------- */

- (void)awakeFromNib
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,self);
#endif
  assert([NSThread isMainThread]);     // Note: timer on main thread

  app = [NSApp delegate];              // MT-only NEW b0308
  assert(app);

  but_lamps = [[NSArray alloc] initWithObjects:but_lamp1,but_lamp2,
    but_lamp3,but_lamp4,but_lamp5,but_lamp6,but_lamp7,but_lamp8,nil];
#ifndef NDEBUG
  { NSSet *set = [NSSet setWithArray:but_lamps]; assert([set count] >= N_LAMPS); }
#endif

  [NSTimer scheduledTimerWithTimeInterval:2.0
           target:self selector:@selector(fireTimer:)
           userInfo:nil repeats:YES];
}

/* ---------------------------------------------------------------- */

- (void)terminate:(const char*)text
{
  if (text) {
    [main_logger message:text level:2 file:YES];
    Ysleep(3000);
    YnotifyNotice(P_TITLE,text,0,2);   // b0301
  }
  [main_logger append:"terminate"];
  dispatch_async(MAIN_QUEUE,^(void) {  // NEW b0309
    [NSApp terminate:self];
  });
}

/* ---------------------------------------------------------------- */

- (BOOL)configure
{
  int  i;
  char buf[256];
#if (DEBUG > 0)
  fprintf(stderr,"%s:%p\n",PREFUN,self);
#endif
  assert([NSThread isMainThread]);
  
  BOOL was_configured = configured;
  configured = FALSE;
  [self updateState:NO stage:NO];      // disable everything
  if (was_configured) [self shutdown]; // close all connections

#ifdef NDEBUG
  use_interlocks = YES;
#else
  use_interlocks = (YprefGetInt("no_interlocks",0,use_shared)) ? NO : YES; 
#endif
  sprintf(buf,"use_interlocks= %d",(int)use_interlocks); [main_logger append:buf];
  eng_mode = (YprefGetInt(KEY_ENGMODE,0,use_shared)) ? YES : NO;
  sprintf(buf,"eng_mode= %d",(int)eng_mode); [main_logger append:buf];
  
  /* --- determine telescope and TCSIS host --- */
  
  if (telname) { [telname release]; telname = nil; }
  telname = YprefGetString(@KEY_TELNAME,nil,use_shared);
  if (!telname) { [self terminate:"no telescope defined"]; return FALSE; }
#ifdef SIM_ONLY
#ifdef NDEBUG
  strcpy(telhost,"localhost");
#else
  strcpy(telhost,"andromeda");
#endif
#else // !SIM_ONLY
  *telhost = '\0';
  if      ([telname compare:@"Baade"] == NSOrderedSame) strcpy(telhost,"mag1tcs");
  else if ([telname compare:@"Clay"]  == NSOrderedSame) strcpy(telhost,"mag2tcs");
#endif
  if (*telhost == '\0') { 
    [self terminate:"invalid telescope\rdelete Preferences and re-start"]; 
    return FALSE;
  }
  [telname retain];

  NSString *title = [NSString stringWithFormat:@"%@/%@ ... connecting to PLC ...",
                                      s_title,telname];
  [window setTitle:title];

  /* --- setup button/lamp labels --- */
  
  NSString *path  = [[NSBundle mainBundle] pathForResource:telname ofType:@"xml"];
  NSDictionary *dict = [NSDictionary dictionaryWithContentsOfFile:path];
  if (dict) { NSString *key,*label; NSButton *button;
    NSMutableArray* array = [NSMutableArray arrayWithCapacity:N_LAMPS];
    for (i=0; i<N_LAMPS; i++) {
      key = [NSString stringWithFormat:@"Lamp%d_Name",(1+i)];
      label = [dict objectForKey:key];
      [array addObject:label];
      assert(label);
      button = [but_lamps objectAtIndex:i];
      [button setTitle:label];
    }
    label = [dict objectForKey:@"varLamp_Name"];
    assert(label);
    [label_quartz setStringValue:label];
    [array addObject:label];         // var. quartz is the last label
    buttonTitles = [[NSArray alloc] initWithArray:array];
#if 1 // hide controls
    [label_quartz setHidden:([label compare:@"-"] == NSOrderedSame)];
    [popup_quartz setHidden:([label compare:@"-"] == NSOrderedSame)];
    [edit_quartz  setHidden:([label compare:@"-"] == NSOrderedSame)];
#endif
  }
  
  /* --- get some PLC addresses --- */

  cookie_addr = o2d([[dict objectForKey:@"Cookie_Address"] intValue]);
#if (DEBUG > 1)
  fprintf(stderr,"cookie_addr=%d\n",cookie_addr);
#endif
  assert(cookie_addr > 0);
  
  selector_addr = o2d([[dict objectForKey:@"Selector_Output"] intValue]);
#if (DEBUG > 1)
  fprintf(stderr,"selector_addr=%d\n",selector_addr);
#endif
  assert(selector_addr > 0);
  
  error_addr = o2d([[dict objectForKey:@"Error_Address"] intValue]);
#if (DEBUG > 1)
  fprintf(stderr,"error_addr=%d\n",error_addr);
#endif
  assert(error_addr > 0);
  
  /* --- release FFmodel devices --- */
  
  if (tcs) { [tcs release]; tcs=nil; }
  if (plc) { [plc release]; plc=nil; }
  
  for (i=0; i<N_LAMPS; i++) {
    if (lamps[i]) {                    // remove old devices
      [lamps[i] removeObserver:self forKeyPath:KEY_MOVING];
      [lamps[i] release]; lamps[i] = nil;
    }
  }
  if (q_lamp) {                        // remove old device
    [q_lamp removeObserver:self forKeyPath:KEY_MOVING];
    [q_lamp release]; q_lamp = nil;
  }
  if (ff_enable) { [ff_enable release]; ff_enable = nil; }
  if (ff_screen) {                     // remove old device
    [ff_screen.brake.pswitch removeObserver:self forKeyPath:KEY_STATUS];
    [ff_screen.brake removeObserver:self forKeyPath:KEY_MOVING];
    [ff_screen removeObserver:self forKeyPath:KEY_MOVING];
    [ff_screen release]; ff_screen = nil;
  }
  if (mc_enable) { [mc_enable release]; mc_enable = nil; }
  if (mc_screen) {                     // remove old device
    [mc_screen.brake.pswitch removeObserver:self forKeyPath:KEY_STATUS];
    [mc_screen.brake removeObserver:self forKeyPath:KEY_MOVING];
    [mc_screen removeObserver:self forKeyPath:KEY_MOVING];
    [mc_screen release]; mc_screen = nil;
  }
  if (air_sensor) {
    [air_sensor release]; air_sensor = nil;
  }
  
  /* ---  the remainder of the configuration (I/O) runs in a thread */
  
  [self performSelectorInBackground:@selector(run_conf) withObject:nil];
  
  return YES;
}

/* ---------------------------------------------------------------- */

- (void)run_conf
{
  NSString *name,*string;
  int      status;
  char     host[128],buf[256];
  int      port,err=0;
#if (DEBUG > 0)
  fprintf(stderr,"%s:%p\n",PREFUN,self);
#endif

  NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
  
  /* --- setup DCU PLC connection, check cookie --- */
  
  port = 502; *host = '\0';
  name = YprefGetString(@KEY_DCUNAME,nil,use_shared); // Note: 192.91.178.187 at SBS
#ifdef SIM_ONLY
  port = 0xffff-port;
#ifdef NDEBUG
  if ([name compare:@"localhost"] != NSOrderedSame) name = @"localhost";
#else
  if ([name compare:@"localhost"] != NSOrderedSame) {
    if ([name compare:@"andromeda"] != NSOrderedSame) name = @"andromeda";
  }
#endif
#else  // !SIM_ONLY
  if ([name compare:@"andromeda"] == NSOrderedSame) port = 0xffff-port;
  if ([name compare:@"localhost"] == NSOrderedSame) port = 0xffff-port;
#endif
#if defined(SIM_ONLY) && defined(NDEBUG)
  status = 0;
#else
  status = YprefGetInt(KEY_DCUSTAT,1,use_shared);
#endif
  Ystr2chr(name,host,sizeof(host));
  plc = [[PLC_IO alloc] initWithHost:host port:port online:status];
  sprintf(buf,"PLC[%s:%d,%d]= %p",host,port,status,plc); [main_logger append:buf];
  if (plc.up_and_running) {            // set/check cookie
    if (!plc.online) { cookie = 0; Ysleep(1000); }
    err = [plc get32bit:cookie_addr value:&cookie];
    if (!err) {
      sprintf(buf,"get cookie=0x%x",cookie); [main_logger append:buf];
      if (cookie != 0) {
        assert(![NSThread isMainThread]);
        dispatch_sync(MAIN_QUEUE,^(void) { char buf[128]; // b0201
          sprintf(buf,"The DCU-PLC might already be in use.\rContinue anyway?");
          cookie_continue = YalertQuestion(p_title,buf,1,"Yes","No");
          if (!cookie_continue) [self terminate:NULL];
        });
        if (!cookie_continue) return;
      }
      cookie = [[MTwister sharedTwister] secureUint];    // b0206
      sprintf(buf,"set cookie=0x%x",cookie); [main_logger append:buf];
      err = [plc set32bit:cookie_addr value:cookie];
    }
    if (err) { NSString *str;
      str = [NSString stringWithFormat:@"setting cookie failed (err=%d)\r%@",
               err,[app errorString:err]];
      [self terminate:[str UTF8String]]; return;
    }
  } else err = E_plcio_error;
  if (err) {                           // terminate application
    [self terminate:"connecting to PLC failed"]; return;
  }

  string = [NSString stringWithFormat:@"%@/%@",s_title,telname];
  [window performSelectorOnMainThread:@selector(setTitle:) 
          withObject:string waitUntilDone:YES];

  /* --- create FFmodel devices (lamps, screens) --- */

  NSString *path = [[NSBundle mainBundle] pathForResource:telname ofType:@"xml"];
  NSDictionary *dict = [NSDictionary dictionaryWithContentsOfFile:path];

  for (int i=0; i<N_LAMPS; i++) { // button titles from array NEW b0311
    assert(!lamps[i]);
    if ([buttonTitles[i] compare:@"-"] == NSOrderedSame) continue;
    NSString *prefix = [NSString stringWithFormat:@"Lamp%d",(1+i)];
    lamps[i] = [[FFlamp alloc] initWithPLC:plc XML:dict prefix:prefix];
    [lamps[i] addObserver:self forKeyPath:KEY_MOVING
                               options:NSKeyValueObservingOptionNew
                               context:(void*)0];
  }
  assert(!q_lamp);
  if ([buttonTitles[N_LAMPS] compare:@"-"] != NSOrderedSame) {
    q_lamp = [[FFvarLamp alloc] initWithPLC:plc XML:dict prefix:@"varLamp"];
    [q_lamp addObserver:self forKeyPath:KEY_MOVING
                             options:NSKeyValueObservingOptionNew
                             context:(void*)0];
  }

  assert(!ff_enable);                  // create FFS-Enable coil
  ff_enable = [[FFcoil alloc] initWithPLC:plc XML:dict prefix:@"FFS_Enable"];
  assert(!ff_screen);                  // create FFS-Screen ?
  ff_screen = [[FFscreen alloc] initWithPLC:plc XML:dict prefix:@"FFS"];
  [ff_screen addObserver:self forKeyPath:KEY_MOVING
             options:NSKeyValueObservingOptionNew context:(void*)0];
  [ff_screen.brake addObserver:self forKeyPath:KEY_MOVING
                   options:NSKeyValueObservingOptionNew context:(void*)0];
  [ff_screen.brake.pswitch addObserver:self forKeyPath:KEY_STATUS
                           options:NSKeyValueObservingOptionNew context:(void*)0];
  ff_disabled = (YprefGetInt(KEY_FFS_DSBLD,0,use_shared)) ? YES : NO;
  if (ff_disabled) [main_logger append:"FFS disabled"];
  
  assert(!mc_enable);
  mc_enable = [[FFcoil alloc] initWithPLC:plc XML:dict prefix:@"MCAL_Enable"];
  assert(!mc_screen);                  // Mcal at 'Clay' only
  if ([telname compare:@"Clay"] == NSOrderedSame) {
    mc_screen = [[FFscreen alloc] initWithPLC:plc XML:dict prefix:@"MCAL"];  
    [mc_screen addObserver:self forKeyPath:KEY_MOVING
               options:NSKeyValueObservingOptionNew context:(void*)0];
    [mc_screen.brake addObserver:self forKeyPath:KEY_MOVING
                     options:NSKeyValueObservingOptionNew context:(void*)0];
    [mc_screen.brake.pswitch addObserver:self forKeyPath:KEY_STATUS
                             options:NSKeyValueObservingOptionNew context:(void*)0];
  } else {
    assert(![NSThread isMainThread]);
    dispatch_sync(dispatch_get_main_queue(),^(void) {  // b0101
      [check_mcins setState:NSControlStateValueOff];
      [check_mcret setState:NSControlStateValueOff];
      [popup_mcbrake selectItemAtIndex:0];
    });
  }
  mc_disabled = (YprefGetInt(KEY_MCAL_DSBLD,0,use_shared)) ? YES : NO;
  if (mc_disabled) [main_logger append:"Mcal disabled"];

  assert(!air_sensor);
  air_sensor = [[FFsensor alloc] initWithPLC:plc XML:dict prefix:@"AIR"];
  
  /* --- setup TCSIS connection --- */
  
  port = 5801;
  status = YprefGetInt(KEY_TELSTAT,1,use_shared);
  tcs = [[TCPIP_Connection alloc] initWithHost:telhost port:port online:status];
  sprintf(buf,"TCS[%s:%d,%d]= %p",telhost,port,status,tcs);
  [main_logger append:buf];

  err = [self getFullStatus];
  if (err) {                           // terminate application
    Ysleep(2000);                      // b0308
    [self terminate:"reading status from PLC failed"]; return;
  }
  
  configured = TRUE;
  [self updateState:YES stage:YES];    // send full state to TCS

  [pool drain];
}

/* ---------------------------------------------------------------- */

- (void)fireTimer:(NSTimer*)theTimer   
{
#if (DEBUG > 2)
  fprintf(stderr,"%s(%p): counter=%lu\n",PREFUN,theTimer,counter);
#endif

 if (!configured) return;
 
#ifndef SIM_ONLY
  if (!plc.online)  // Simulator
#endif
  { NSString *string;
    if (counter % 2) string = [NSString stringWithFormat:@"%@/%@",s_title,telname];
    else             string = [NSString stringWithFormat:@"%@ (Simulator)",s_title];
    [window setTitle:string];
  }

  if ([self running]) return;
      
  if (plc.up_and_running && interval_dog) {
    if (time(NULL) >= next_dog) {      // check cookie
      [self performSelectorInBackground:@selector(run_dog) withObject:nil];
      next_dog = time(NULL) + interval_dog;
    }
  }

  counter++;
}

/* ---------------------------------------------------------------- */

- (int)getFullStatus
{
  int err;
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,self);
#endif
  
  err = [plc open]; 
  if (!err) {
    for (int i=0; i<N_LAMPS; i++) {
      err = [lamps[i] get]; if (err) break; // Note: [nil] is ok
    }
    if (!err) err = [q_lamp get];           // Note: [nil] is ok
    if (!err) err = [ff_screen get];
    if (!err) err = [mc_screen get];
    if (!err) [self checkAir:FALSE];
    [plc close];
  } 
  
  if (err) { NSString *str;                 // NSString b0308
    str = [NSString stringWithFormat:@"getting status from PLC failed, err=%d\r%@",
             err,[app errorString:err]];
    [main_logger message:[str UTF8String] level:2 time:0 file:YES]; // logfile NEW b0307
  }
  
  return err;
}

/* ---------------------------------------------------------------- */

- (void)shutdown 
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,self);
#endif

  if (plc.up_and_running) {
    [main_logger append:"set cookie=0"];
    [plc set32bit:cookie_addr value:0];
    [main_logger append:"PLC shutdown"];
    [plc shutdown];
  }
  if (tcs.up_and_running) {
    [main_logger append:"TCS shutdown"];
    [tcs shutdown];
  }
}

/* ---------------------------------------------------------------- */

- (BOOL)running
{
  if (prog_screen.counter) return TRUE;
  if (prog_mcal.counter) return TRUE;
  return FALSE;
}

/* --- */

- (BOOL)deployed
{
  if (ff_screen.status == SCREEN_INS) return TRUE;
  if (mc_screen.status == SCREEN_INS) return TRUE;
  return FALSE;
}

/* ---------------------------------------------------------------- */

- (NSString*)ffsState                  // NEW b0310
{
  NSString *s=@"unknown";

  if      (ff_screen.status == SCREEN_INS) s = @"deployed";
  else if (ff_screen.status == SCREEN_RET) s = @"retracted";

  return s;
}

/* --- */

- (NSString*)ffsLamps                  // NEW b0310
{
  char buf[256]="";
  
  for (int i=0; i<N_LAMPS; i++) {      // fixed level lamps
    if (!lamps[i]) continue;
    if (lamps[i].status) {             // lamp ON
      if (strlen(buf) > 0) strcat(buf," ");
      strcat(buf,[buttonTitles[i] UTF8String]);
    }
  }
  if (strlen(buf) < 2) strcpy(buf,"Off");
  
  return @(buf);
}

/* ---------------------------------------------------------------- */
#pragma mark "Actions"
/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

- (IBAction)action_lamp:(id)sender
{
  int  req;
  char buf[256];
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,sender);
#endif
  assert([NSThread isMainThread]);

  [sender setEnabled:NO];

  if (sender == popup_quartz) {        // variable quartz lamp
    assert(q_lamp);                    // TODO no var. lamp available yet
    req = 1000*(int)[sender indexOfSelectedItem];
    sprintf(buf,"%s: var= %d",PREFUN,req); [main_logger append:buf];
    NSNumber *pos = [NSNumber numberWithInt:req];
    [q_lamp performSelectorInBackground:@selector(run:) withObject:pos];
  } else { int i;                      // fixed level lamps
    i = (int)[but_lamps indexOfObject:sender];
    assert(i < N_LAMPS);
    assert(lamps[i]);
    req = (lamps[i].status != LAMP_OFF) ? LAMP_OFF : LAMP_ON;
    sprintf(buf,"%s: i= %d req= %d",PREFUN,i,req); [main_logger append:buf];
    NSNumber *pos = [NSNumber numberWithInt:req];
    [lamps[i] performSelectorInBackground:@selector(run:) withObject:pos];
  }
}

/* ---------------------------------------------------------------- */

- (IBAction)action_ffs:(id)sender
{
  int  req=0,err=0;
  char buf[256];
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,sender);
#endif
  assert([NSThread isMainThread]);
  assert(ff_screen);
  assert(app);

  if (!ff_screen || ff_disabled) {     // might be called by script b0207
    err = error_state = E_DCU_error;   // global 'error_state'
    NSString *str = [NSString stringWithFormat:@"FFS-DCU not available (err=%d)\r%@",
      err,[app errorString:err]];
    [Notice openNotice:[str UTF8String] title:P_TITLE time:0 level:2];
    return;
  }

  BOOL script = ([sender isKindOfClass:[NSScriptCommand class]]) ? TRUE : FALSE;
  if (script) {                       
    if (!eng_mode) {
      [Notice openNotice:"Scripting allowed in 'engineering mode' only" 
              title:P_TITLE time:30 level:1];
      error_state = E_DCU_error;       // stops script b0207
      return;
    }
  }
  if (!configured) err = E_DCU_config;  // too early
  if ([self running]) err = E_DCU_busy; // catch scripting error

  if (!err) err = [plc open];
  if (!err) { int value=DCU_FFS;       // Simulator preset
    err = [self checkPLC:YES];
    if (mc_screen) {
      if (!err) err = [mc_screen get]; // other DCU
      if (!err) {                      // is out of the way?
        sprintf(buf,"mc_screen.ret_limit= %d (%d)",
                mc_screen.ret_limit.status,mc_screen.ins_limit.status);
        [main_logger append:buf];
        if (use_interlocks) {
          if ((mc_screen.ret_limit.status != 1) || (mc_screen.ins_limit.status != 0)) {
            err = E_DCU_ret; 
          } 
        }
      }
    }
    if (!err) { if (![self checkAir:TRUE]) { err = E_DCU_air; } }
    if (!err) err = [mc_enable set:0];
    if (!err) err = [ff_enable set:1];
    [NSThread sleepForTimeInterval:PLC_DELAY];  
    if (!err) err = [plc getOutput:selector_addr value:&value];
    if (!err) { if (value != DCU_FFS) err = E_DCU_selector; }
    [plc close];
  }
  if (err) {
    error_state = err;
    [self updateFFS]; if (err == E_DCU_ret) [self updateMcal];
    NSString *str = [NSString stringWithFormat:@"enabling FFS-DCU failed (err=%d)\r%@",
      err,[app errorString:err]];
    [main_logger message:[str UTF8String] level:2 file:YES]; // logfile NEW b0307
    return;
  }

  if (!script) [sender setEnabled:FALSE];

  if (sender == popup_ffbrake) {       // FFS brake
    req = (int)[sender indexOfSelectedItem];
    sprintf(buf,"%s: brake= %d",PREFUN,req); [main_logger append:buf];
    NSNumber *pos = [NSNumber numberWithInt:req];
    [ff_screen.brake performSelectorInBackground:@selector(run:) withObject:pos];
  } else {                             // FFS screen
    switch ([drop_ffs indexOfSelectedItem]) {
      case 1: req = SCREEN_RET; break;
      case 2: req = SCREEN_INS; break;
      default: assert(0); break;
    }
    if (!eng_mode) {
      if (req == ff_screen.status) {
        [self updateFFS];
        sprintf(buf,"FF-Screen already %s",(req==SCREEN_RET) ? "retracted" : "deployed");
        [Notice openNotice:buf title:P_TITLE time:30 level:1];
        return;
      }
    }
    sprintf(buf,"%s: screen= %d",PREFUN,req); [main_logger append:buf];
    NSNumber *pos = [NSNumber numberWithInt:req];
    [ff_screen performSelectorInBackground:@selector(run:) withObject:pos];
    if (req == SCREEN_RET) { int i;    // turn off all lamps
      for (i=0; i<N_LAMPS; i++) {
        if (!lamps[i]) continue;
        if (lamps[i].status != LAMP_OFF) {
          [self action_lamp:[but_lamps objectAtIndex:i]];
        }
      }
      if (q_lamp) {
        if (q_lamp.status != LAMP_OFF) {
          [popup_quartz selectItemAtIndex:0];
          [self action_lamp:popup_quartz];
        }
      }
    }
  }
}

/* ---------------------------------------------------------------- */

- (IBAction)action_mcal:(id)sender     // takes about 35 seconds
{
  int  req=0,err=0;
  char buf[256];
  NSString *str;
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,sender);
#endif
  assert([NSThread isMainThread]);
  assert(app);

  if (!mc_screen || mc_disabled) {     // might be called by script b0207
    err = error_state = E_DCU_error;   // global 'error_state'
    str = [NSString stringWithFormat:@"MCAL-DCU not available (err=%d)\r%@",
            err,[app errorString:err]];
    [Notice openNotice:[str UTF8String] title:P_TITLE time:0 level:2];
    return;
  }
  
  BOOL script = ([sender isKindOfClass:[NSScriptCommand class]]) ? TRUE : FALSE;
  if (script) {
    if (!eng_mode) {
      [Notice openNotice:"Scripting allowed in 'engineering mode' only" 
              title:P_TITLE time:30 level:1];
      error_state = E_DCU_error;       // stops script b0207
      return;
    }
  }
  
  if (!configured) err = E_DCU_config;   // too early
  if ([self running]) err = E_DCU_busy;  // catch scripting error
  
  if (!err) err = [plc open];
  if (!err) { int value=DCU_MCAL;      // Simulator preset
    err = [self checkPLC:YES];
    if (ff_screen) {
      if (!err) err = [ff_screen get]; // other DCU
      if (!err) {                      // is out of the way?
        sprintf(buf,"ff_screen.ret_limit= %d (%d)",
                ff_screen.ret_limit.status,ff_screen.ins_limit.status);
        [main_logger append:buf];
        if (use_interlocks) {
          if ((ff_screen.ret_limit.status != 1) || (ff_screen.ins_limit.status != 0)) {
            err = E_DCU_ret; 
          }
        }
      }
    }
    if (!err) { if (![self checkAir:TRUE]) { err = E_DCU_air; } }
    if (!err) err = [ff_enable set:0];
    if (!err) err = [mc_enable set:1];
    [NSThread sleepForTimeInterval:PLC_DELAY];
    if (!err) err = [plc getOutput:selector_addr value:&value];
    if (!err) { if (value != DCU_MCAL) err = E_DCU_selector; }
    [plc close];
  } 
  if (err) {
    error_state = err;
    [self updateMcal]; if (err == E_DCU_ret) [self updateFFS];
    str = [NSString stringWithFormat:@"enabling MCAL-DCU failed (err=%d)\r%@",
            err,[app errorString:err]];
    [main_logger message:[str UTF8String] level:2 file:YES]; // logfile NEW b0307
    return;
  }

  if (!script) [sender setEnabled:FALSE];

  if (sender == popup_mcbrake) {       // move MCAL brake
    req = (int)[sender indexOfSelectedItem];
    sprintf(buf,"%s: brake= %d",PREFUN,req); [main_logger append:buf];
    NSNumber *pos = [NSNumber numberWithInt:req];
    [mc_screen.brake performSelectorInBackground:@selector(run:) withObject:pos];
  } else {                             // move MCAL screen
    switch ([drop_mcal indexOfSelectedItem]) {
      case 1: req = SCREEN_RET; break;
      case 2: req = SCREEN_INS; break;
      default: assert(0); break;
    }
    if (!eng_mode) {
      if (req == mc_screen.status) {
        [self updateMcal];
        sprintf(buf,"Mcal already %s",(req==SCREEN_RET) ? "retracted" : "deployed");
        [Notice openNotice:buf title:P_TITLE time:30 level:1];
        return;
      }
    }
    sprintf(buf,"%s: screen= %d",PREFUN,req); [main_logger append:buf];
    NSNumber *pos = [NSNumber numberWithInt:req];
    [mc_screen performSelectorInBackground:@selector(run:) withObject:pos];
  }
}

/* ---------------------------------------------------------------- */
#pragma mark "Update"
/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

- (void)updateState:(BOOL)snd_lamps stage:(BOOL)snd_stage // b0101
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%d,%d)\n",PREFUN,snd_lamps,snd_stage);
#endif

  for (int i=0; i<N_LAMPS; i++) [self updateLamp:i];
  
  if (snd_lamps || snd_stage) [self updateTCS:snd_lamps stage:snd_stage];
  
  [self updateVar];
  [self updateFFS];
  [self updateMcal];

  dispatch_async(MAIN_QUEUE,^(void) {    // NSView on MT NEW b0305
    if ((mc_screen.status == SCREEN_INS) || (ff_screen.status == SCREEN_INS)) {
      [view_in setImage:icon_in];        // b0200
    } else {
      [view_in setImage:nil];
    }
  });
}

/* ---------------------------------------------------------------- */

- (void)updateLamp:(int)lamp
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%d)\n",PREFUN,lamp);
#endif

  dispatch_async(dispatch_get_main_queue(),^(void) {   // b0101
    NSButton *button = [but_lamps objectAtIndex:lamp]; // b0200
    BOOL enable=configured;
    if (!lamps[lamp]) { [button setEnabled:NO]; return; }
    if (lamps[lamp].moving) {
      [button setImage:((lamps[lamp].request) ? icon_on : icon_off)];
    } else {
      if (lamps[lamp].request == lamps[lamp].status) {
        [button setImage:((lamps[lamp].status) ? icon_on : icon_off)];
      } else {
        [button setImage:icon_err];
      }
      if (ff_screen.status == SCREEN_INS) {
        [button setEnabled:enable];
      } else {
        if (!eng_mode) {
          [button setEnabled:((lamps[lamp].status != LAMP_OFF) ? enable : NO)];
        } else {
          [button setEnabled:enable];
        }
      }
    } // endif(lamps[i].moving)
  });  
}

/* ---------------------------------------------------------------- */

- (void)updateVar
{
#if (DEBUG > 1)
  fprintf(stderr,"%s()\n",PREFUN);
#endif

  dispatch_async(dispatch_get_main_queue(),^(void) { // b0101
    BOOL enable=configured;
    if (!q_lamp) { [popup_quartz setEnabled:NO]; return; } 
    if (!q_lamp.moving) {
      if (ff_screen.status == SCREEN_INS) {
        [popup_quartz setEnabled:enable];
      } else {
        if (!eng_mode) {
          [popup_quartz setEnabled:((q_lamp.status != 0) ? enable : NO)];
        } else {
          [popup_quartz setEnabled:enable];
        }
      }
    }
    NSString *str = [NSString stringWithFormat:@"%.3f",level2volt(q_lamp.level)];
    [edit_quartz setStringValue:str];
  });
}

/* ---------------------------------------------------------------- */

- (void)updateTCS:(BOOL)snd_lamps stage:(BOOL)snd_stage
{
  char cmd[256],buf[256];
#if (DEBUG > 1)
  fprintf(stderr,"%s(%d,%d): up_and_running=%d\n",PREFUN,
          snd_lamps,snd_stage,tcs.up_and_running);
#endif

  if (!tcs.up_and_running) return;

  int err = [tcs open];                // b0100
  if (err) {
    sprintf(buf,"opening TCP/IP connection to TCSIS failed (err=%d)",err);
    [main_logger message:buf level:1 time:30 file:YES]; // logfile NEW b0307
    return;
  }

  if (snd_lamps) { char state[32];     // send "fflamps"
    strcpy(cmd,"fflamps");
    for (int i=0; i<N_LAMPS; i++) {    // fixed level lamps
      if (!lamps[i]) continue;
      if (lamps[i].status) {           // lamp ON
        strcat(cmd," "); strcat(cmd,[buttonTitles[i] UTF8String]); 
      } 
    }
    if (q_lamp) {                      // variable Quartz lamp
      if (q_lamp.level) {              // TODO no var. lamp available yet
        sprintf(state," vQ%.2f",level2volt(q_lamp.level));
        strcat(cmd,state);
      }
    }
    if (strlen(cmd) < 8) strcat(cmd," Off");
    sprintf(buf,"TCSIS: %s",cmd); [main_logger append:buf];
    strcat(cmd,"\n");                  // append <LF>
    err = [tcs request:cmd response:buf max:sizeof(buf)];
    if (err) {
      sprintf(buf,"sending lamp state to TCSIS failed (err=%d)",err);
      [main_logger message:buf level:1 time:30 file:YES]; // logfile NEW b0307
    }
  } // endif(snd_lamps)
  
  if (snd_stage) {                     // send "dcustat" b0101
    strcpy(cmd,"dcustat ");
    if ([self running]) {              // any one moving
      strcat(cmd,"moving");
    } else
    if ([self deployed]) {             // any one delpoyed
      if      (ff_screen.status == SCREEN_INS) strcat(cmd,"in/ffs");
      else if (mc_screen.status == SCREEN_INS) strcat(cmd,"in/mcal");
      else                                     strcat(cmd,"unknown");
    } else { BOOL ok=YES;              // other
      if (ff_screen) {
        if (ff_screen.status != SCREEN_RET) ok = NO;
      }
      if (mc_screen) {
        if (mc_screen.status != SCREEN_RET) ok = NO;
      }
      strcat(cmd,(ok) ? "out" : "unknown");
    }
    sprintf(buf,"TCSIS: %s",cmd); [main_logger append:buf];
    strcat(cmd,"\n");                  // append <LF>
    err = [tcs request:cmd response:buf max:sizeof(buf)];
    if (err) { NSString *str;
      str = [NSString stringWithFormat:@"sending DCU state to TCSIS failed (err=%d)\r%@",
                      err,[app errorString:err]];
      [main_logger message:[str UTF8String] level:1 time:30 file:YES]; // logfile NEW b0307
    }
  }
  
  [tcs close];
}

/* ---------------------------------------------------------------- */

- (void)updateScreen:(FFscreen*)screen Popup:(NSPopUpButton*)drop
{
  NSButton *but;

  if (!screen) return;                 // 'Baade' has no MCal unit
#if (DEBUG >  1)
  fprintf(stderr,"%s(%p,%p)\n",PREFUN,screen,drop);
#endif
  assert([NSThread isMainThread]);

  if (screen.moving) {                 // screen is moving
    [drop setTitle:@"moving"];
    [drop setEnabled:NO];
  } else { NSString *status;           // not moving
    switch (screen.status) { 
    case SCREEN_RET:
      status = @"retracted";
      break;
    case SCREEN_INS:
      status = @"deployed";
      break;
    case 0:
      status = @"floating";
      break;    
    default:
      status = @"Error";
      break;
    }
    if (screen.status == screen.request) { // consistent state
      [drop setTitle:status];
    } else {
      NSString *string = [NSString stringWithFormat:@"?%@",status];
      [drop setTitle:string];
    }
  }

  but = (screen == ff_screen) ? check_ffins : check_mcins;
  [but setState:((screen.ins_limit.status) ? NSControlStateValueOn : NSControlStateValueOff)];
  but = (screen == ff_screen) ? check_ffret : check_mcret;
  [but setState:((screen.ret_limit.status) ? NSControlStateValueOn : NSControlStateValueOff)];
}

/* ---------------------------------------------------------------- */

- (void)updateFFS
{
  BOOL enable=(configured && !ff_disabled);

#if (DEBUG > 1)
  fprintf(stderr,"%s()\n",PREFUN);
#endif

  if (![NSThread isMainThread]) {      // v0.810
    [self performSelectorOnMainThread:_cmd withObject:nil waitUntilDone:YES];
    return;
  } 
  assert([NSThread isMainThread]);

  [self updateScreen:ff_screen Popup:drop_ffs];
  
  if (!ff_screen.brake.moving) [popup_ffbrake selectItemAtIndex:ff_screen.brake.status];
  [check_ffbrake setState:((ff_screen.brake.pswitch.status) ? NSControlStateValueOn : NSControlStateValueOff)];

  if (!ff_screen.moving && !ff_screen.brake.moving) { BOOL allow=!ff_disabled;
    if (use_interlocks && mc_screen && allow) {
      allow = (mc_screen.status == SCREEN_RET) && (!mc_screen.moving);
      if (mc_screen.brake.status != BRAKE_ON) allow = NO;
    }
    [drop_ffs setEnabled:(allow && enable)];
    allow = (eng_mode && !ff_disabled) ? YES : NO;
    [popup_ffbrake setEnabled:(enable && allow)];
  } else {                             // disable brake and screen
    [popup_ffbrake setEnabled:NO];
    [drop_ffs setEnabled:NO];
  }
}

/* ---------------------------------------------------------------- */

- (void)updateMcal
{
  BOOL enable=(configured && !mc_disabled);

#if (DEBUG > 1)
  fprintf(stderr,"%s()\n",PREFUN);
#endif

  if (![NSThread isMainThread]) {      // v0.810
    [self performSelectorOnMainThread:_cmd withObject:nil waitUntilDone:YES];
    return;
  } 
  assert([NSThread isMainThread]);

  if (mc_screen) {
    [self updateScreen:mc_screen Popup:drop_mcal];
  } else {
    enable = FALSE;
  }

  if (!mc_screen.brake.moving) [popup_mcbrake selectItemAtIndex:mc_screen.brake.status];
  [check_mcbrake setState:((mc_screen.brake.pswitch.status) ? NSControlStateValueOn : NSControlStateValueOff)];

  if (!mc_screen.moving && !mc_screen.brake.moving) { BOOL allow=!mc_disabled;
    if (use_interlocks && allow) {
      allow = (ff_screen.status == SCREEN_RET) && (!ff_screen.moving);
      if (ff_screen.brake.status != BRAKE_ON) allow = NO; // v0.807
    }
    [drop_mcal setEnabled:(allow && enable)];
    allow = (eng_mode && !mc_disabled) ? YES : NO;
    [popup_mcbrake setEnabled:(enable && allow)];
  } else {                             // disable brake and screen
    [popup_mcbrake setEnabled:NO];
    [drop_mcal setEnabled:NO];
  }
}

/* ---------------------------------------------------------------- */

- (void)run_dog
{
  u_int uval=cookie;                   // Simulator preset
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p)\n",PREFUN,self);
#endif

  NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];

  int err = [plc open];
  if (!err) { 
    err = [plc get32bit:cookie_addr value:&uval]; 
    if (!err) {
      if (uval != cookie) { char buf[128];
        sprintf(buf,"Warning: DCU-PLC cookie has changed (%x)",uval);
        strcat(buf,"\rAnother GUI might be accessing the DCU-PLC");
        [main_logger message:buf level:1 time:interval_dog file:YES]; // logfile NEW b0307
        err = [plc set32bit:cookie_addr value:cookie];
        if (!err) {
          [self getFullStatus];
          [self updateState:YES stage:YES];
        }
      }
    }
    if (!err) [self checkAir:FALSE];   // as part of 'watchdog' 
    [plc close];
  }
  
  if (err) { NSString *str;
    str = [NSString stringWithFormat:@"getting/setting cookie at DCU-PLC failed (err=%d)\r%@",
                    err,[app errorString:err]];
    [main_logger message:[str UTF8String] level:1 time:interval_dog file:YES]; // logfile NEW b0307
  }

  [pool drain];
}

/* ---------------------------------------------------------------- */

- (void)updateAir
{
  assert([NSThread isMainThread]);

  if (air_pressure < 0) { 
    [edit_air setStringValue:@"?"];
  } else {
    if (air_pressure < min_pressure+1) [edit_air setTextColor:[NSColor redColor]];
    else                               [edit_air setTextColor:[NSColor blackColor]];
    [edit_air setStringValue:[NSString stringWithFormat:@"%.1f",air_pressure]];
  }
}

/* --- */

- (BOOL)checkAir:(BOOL)interactive
{
  int  err,value;
  BOOL r=FALSE; 
  char buf[128];

  assert(air_sensor);
  value = (int)(10.0*air_pressure);    // [0.1 PSI]
  if (air_pressure <= min_pressure) value += [MTwister intValue:5]-1;
  else                              value += [MTwister intValue:7]-3;
  if (value < 0) value = 0;
  err = [air_sensor get:value];
  if (!err) {
    air_pressure = (float)air_sensor.status/10.0f;
    if (air_pressure < min_pressure) {
      if (interactive) { char buf[128];
        sprintf(buf,"air-pressure too low: %.1f [psi]",air_pressure);
        [Notice openNotice:buf title:P_TITLE time:30 level:1];
      }  
    } else r = TRUE;
    sprintf(buf,"air_pressure= %.1f",air_pressure); [main_logger append:buf];
  } else { 
    if (interactive) {
      sprintf(buf,"reading air-pressure failed (err=%d)",err);
      [main_logger message:buf level:2 time:30 file:YES];
    }
    air_pressure = -1.0f;
  }

  [self performSelectorOnMainThread:@selector(updateAir) 
        withObject:nil waitUntilDone:NO];

  return r;
}

/* ---------------------------------------------------------------- */

static const char *reasons[] = {       // PLC flags
  "ok",
  "Mcal not retracted while using FFS",
  "FFS not retracted while using Mcal",
  ""                                   // terminator
}; 

- (int)checkPLC:(BOOL)interactive     
{
  int err,error=0;
  
#ifdef SIM_ONLY
  if (!plc.online) error = (error_state) ? [MTwister intValue:(0x01 << 2)] : 0;
#endif
  err = [plc getValue:error_addr value:&error]; 
  if (!err) { if (error) err = E_DCU_disabled; }

  if (error) { char text[512],*p;
    sprintf(text,"PLC disabled (code=0x%02x)",error);
    for (int i=0; *reasons[i]; i++) {
      if (error & (0x01 << i)) {
        strcat(text,"\r"); strcat(text,reasons[i]); 
      }
    }
    if (interactive) {
      [Notice openNotice:text title:P_TITLE time:0 level:2];
    }                                  // replace <CR>
    p = text; while (*p) { if (*p =='\r') *p = ' '; p++; }
    [main_logger append:text];
  }
  
  return err;
}

/* ---------------------------------------------------------------- */
#pragma mark "KVO"
/* ---------------------------------------------------------------- */

- (void)observeKey:(NSString*)key ofObject:(id)object // b0101
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p,%p)\n",PREFUN,key,object);
#endif
  assert([NSThread isMainThread]);

  if (object == ff_screen) {           // FF screen
    if ([key compare:KEY_MOVING] == NSOrderedSame) {
      if (ff_screen.moving) {
        error_state = 0;               // set global error
        [prog_screen turnOn];
      } else {
        error_state = (ff_screen.request == ff_screen.status) ? 0 : E_DCU_failed;
        [prog_screen turnOff];
        if (error_state) [self checkPLC:YES];
        else {                         // b0300
          if (ff_screen.status == SCREEN_RET) {
            [[NSDistributedNotificationCenter defaultCenter] postNotificationName:@"DCU/FFS"
                   object:@"retracted" userInfo:nil deliverImmediately:YES];
          } else
          if (ff_screen.status == SCREEN_INS) {
            [[NSDistributedNotificationCenter defaultCenter] postNotificationName:@"DCU/FFS"
                   object:@"deployed" userInfo:nil deliverImmediately:YES];
          }
        }
      }
      [self updateState:NO stage:YES];
    } else {
      assert(0);
    }
  } else 
  if (object == ff_screen.brake) {     // FFS brake
    if ([key compare:KEY_MOVING] == NSOrderedSame) {
      if (ff_screen.brake.moving) [prog_screen turnOn];
      else                        [prog_screen turnOff];
      [self updateFFS]; if (use_interlocks) [self updateMcal];
    }
  } else
  if (object == ff_screen.brake.pswitch) { // FFS brake pressure-switch
    if ([key compare:KEY_STATUS] == NSOrderedSame) {
      [self updateFFS];
    } else {
      assert(0);
    }    
  } else
  if (object == mc_screen) {           // Mcal screen
    if ([key compare:KEY_MOVING] == NSOrderedSame) {
      if (mc_screen.moving) {
        error_state = 0;               // set global error
        [prog_mcal turnOn];
      } else {
        error_state = (mc_screen.request == mc_screen.status) ? 0 : E_DCU_failed;
        [prog_mcal turnOff];
        if (error_state) [self checkPLC:YES];
        else {                         // b0211
          if (mc_screen.status == SCREEN_RET) {
            [[NSDistributedNotificationCenter defaultCenter] postNotificationName:@"DCU/Mcal"
                   object:@"retracted" userInfo:nil deliverImmediately:YES];
          } else
          if (mc_screen.status == SCREEN_INS) {
            [[NSDistributedNotificationCenter defaultCenter] postNotificationName:@"DCU/Mcal"
                   object:@"deployed" userInfo:nil deliverImmediately:YES];
          }
        }
      }
      [self updateState:NO stage:YES];
    } else {
      assert(0);
    }
  } else
  if (object == mc_screen.brake) {     // Mcal DCU brake
    if ([key compare:KEY_MOVING] == NSOrderedSame) {
      if (mc_screen.brake.moving) [prog_mcal turnOn];
      else                        [prog_mcal turnOff]; 
      [self updateMcal]; if (use_interlocks) [self updateFFS];
    }
  } else 
  if (object == mc_screen.brake.pswitch) { // Mcal DCU brake sensor
    if ([key compare:KEY_STATUS] == NSOrderedSame) {
      [self updateMcal]; 
    } else {
      assert(0);
    }    
  } else
  if (object == q_lamp) {              // variable quartz lamp
    if ([key compare:KEY_MOVING] == NSOrderedSame) {
      if (q_lamp.moving) {
        [prog_screen turnOn];
      } else {
        [prog_screen turnOff];
        [self updateTCS:YES stage:NO];
      }
      [self updateVar];      
    } else {
      assert(0);
    }
  } else { int i;
    for (i=0; i<N_LAMPS; i++) {        // fixed level lamps
      if (object == lamps[i]) {
        if ([key compare:KEY_MOVING] == NSOrderedSame) {
          if (lamps[i].moving) {       // move started
            [prog_screen turnOn];
          } else {                     // move ended
            [prog_screen turnOff];
            [self updateTCS:YES stage:NO];
          }
          [self updateLamp:i];
        } else {
          assert(0);
        }
        break;
      }
    }
#ifndef NDEBUG
    if (i == N_LAMPS) { assert(0); }
#endif
  }
}

/* --- */

- (void)observeValueForKeyPath:(NSString*)key
                      ofObject:(id)object
                        change:(NSDictionary*)dict
                       context:(void*)param
{
#if (DEBUG > 2)
  fprintf(stderr,"%s(%p,%p,%p,%p): mainThread=%d\n",PREFUN,
          key,object,dict,param,[NSThread isMainThread]);
#endif

  if ([NSThread isMainThread]) {
    [self observeKey:key ofObject:object];
  } else {
    dispatch_sync(MAIN_QUEUE,^(void) {  // b0101
      [self observeKey:key ofObject:object];
    });
  }
}

/* ---------------------------------------------------------------- */
#pragma mark "NSWindow delegates"
/* ---------------------------------------------------------------- */

#if 0 // unused
// - (void)windowDidExpose:(NSNotification*)note
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p): window=%p\n",PREFUN,note,[note object]);
#endif
}
#endif

/* --- */

#if 0 // unused
// - (void)windowDidBecomeKey:(NSNotification*)note
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p): window=%p\n",PREFUN,note,[note object]);
#endif
}
#endif

/* --- */

#if 0 // unused
// - (void)windowDidBecomeMain:(NSNotification*)note
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p): window=%p\n",PREFUN,note,[note object]);
#endif
}
#endif

/* --- */

- (BOOL)windowShouldClose:(id)sender
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,sender);
#endif

#ifdef NDEBUG
  if ([self deployed]) { // Note: [close] disabled in Release version
    if (YalertQuestion(P_TITLE,"DCU deployed - Quit anyway?",1,"No","Yes")) return NO;
  }
#endif

  return YES;
}

/* ---------------------------------------------------------------- */

@end

/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
