/* ---------------------------------------------------------------- */
//
// Logger.m
//
// Created by Christoph Birk on 2013-04-10
// Copyright 2013 Carnegie Observatories. All rights reserved.
//
// 2014-03-07  /Users/Shared/Library/Logs
// 2014-04-02  set permissions when using Shared
// 2014-04-23  use Ystr2tmp()
// 2014-09-24  [append: toArray:]
// 2014-12-04  remove noticeLock
// 2014-12-10  YsharedLogPath()
// 2015-03-16  add CFBundleShortVersionString
// 2015-04-06  add subFolder, autoChange
// 2015-06-22  showUT
// 2015-10-28  timer=nil in setAutoChange
//
/* ---------------------------------------------------------------- */

#ifndef DEBUG
#define DEBUG      0                   // debug level
#endif

#define M_FILE     512                 // max. file name length

/* ---------------------------------------------------------------- */

#import "Logger.h"
#import "Notice.h"
#import "Yutils.h"

/* ---------------------------------------------------------------- */

static NSDateFormatter *timeFormatter=nil;

/* ---------------------------------------------------------------- */

@implementation Logger

@synthesize logfile,showUT,offset;

/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

+ (void)initialize
{
#if (DEBUG > 0)
  fprintf(stderr,"%s:%p\n",PREFUN,self);
#endif

  if (self == [Logger class]) {        // avoid call by sub-class
    timeFormatter = [[NSDateFormatter alloc] init];
    [timeFormatter setDateFormat:@"HH:mm:ss"];
  }
}

/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */

- (Logger*)init
{
#if (DEBUG > 0)
  fprintf(stderr,"%s:%p\n",PREFUN,self);
#endif

  return [self initWithBase:nil];
}

/* ---------------------------------------------------------------- */

- (Logger*)initWithBase:(const char*)base
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%s)\n",PREFUN,base);
#endif

  return [self initWithBase:base useShared:NO];
}

/* ---------------------------------------------------------------- */

- (Logger*)initWithBase:(const char*)base useShared:(BOOL)shared
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%s,%d)\n",PREFUN,base,shared);
#endif

  return [self initWithBase:base useShared:shared period:LOGGER_WEEKLY];
}

/* ---------------------------------------------------------------- */

- (Logger*)initWithBase:(const char*)base useShared:(BOOL)shared period:(int)in_per
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%s,%d,%d)\n",PREFUN,base,shared,in_per);
#endif

  return [self initWithBase:base useShared:shared period:in_per folder:NULL];
}

/* ---------------------------------------------------------------- */

- (Logger*)initWithBase:(const char*)base useShared:(BOOL)shared period:(int)in_per folder:(const char*)sub
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p,%d,%d,%p)\n",PREFUN,base,shared,in_per,sub);
#endif

  if ((self = [super init]) == nil) return nil;  // Clang complains without
  
  NSBundle *bundle = [NSBundle mainBundle];
  p_exec    = [[bundle objectForInfoDictionaryKey:@"CFBundleExecutable"] retain];
  p_build   = [[bundle objectForInfoDictionaryKey:@"CFBundleVersion"] retain];
  p_version = [[bundle objectForInfoDictionaryKey:@"CFBundleShortVersionString"] retain];
  if (base) baseName = Ychr2str(base);
  else      baseName = [p_exec lowercaseString];
  [baseName retain];

  messageLock = [[NSLock alloc] init];
  fileLock = [[NSLock alloc] init];
  n_msg = 128;                      // table size
  mesgArray = [[NSMutableArray alloc] initWithCapacity:n_msg];
  timeArray = [[NSMutableArray alloc] initWithCapacity:n_msg];
  assert(mesg_head == 0);
  sharedPath = shared;
  period = in_per;
  offset = 9;                       // 9 am local name change
  showUT = NO;
  if (sub) folder = [[NSString stringWithFormat:@"/%s",sub] retain];
  [self setLogfile:[self logfileName]];
  
  return self;
}

/* ---------------------------------------------------------------- */

- (void)dealloc
{
#if (DEBUG > 0)
  fprintf(stderr,"%s:%p\n",PREFUN,self);
#endif

  [p_version release];
  [p_build release];
  [p_exec release];
  [baseName release];
  [messageLock release];
  [fileLock release];
  [mesgArray release];
  [timeArray release];
  [folder release];
  
  [super dealloc];
}

/* ---------------------------------------------------------------- */

- (NSString*)logfileName
{
  time_t    ut; 
  struct tm *now,res;
  char      tstr[32];
  NSString  *name;
  NSMutableString *path=nil;
  
  switch (period) {
#ifndef NDEBUG
  case -1:
    ut = time(NULL);
    now = localtime_r(&ut,&res);
    strftime(tstr,sizeof(tstr),"%Y_%m_%d_%H_%M",now);
    break;
#endif
  case LOGGER_HOURLY:
    ut = time(NULL);                     // UT
    now = localtime_r(&ut,&res);
    strftime(tstr,sizeof(tstr),"%Y_%m_%d_%H",now);
    break;
  case LOGGER_DAILY:
    ut = time(NULL) - offset*3600;
    now = localtime_r(&ut,&res);
    strftime(tstr,sizeof(tstr),"%Y_%m_%d",now);
    break;
  case LOGGER_WEEKLY:
  default:
    ut = time(NULL) - offset*3600;
    now = localtime_r(&ut,&res);
    strftime(tstr,sizeof(tstr),"%Y_%W",now);
    break;
  }
  
  if (sharedPath) {
    path = [NSMutableString stringWithFormat:@"%@/%@",YsharedLogPath(),p_exec];
  } else { NSArray *ary;
    ary = NSSearchPathForDirectoriesInDomains(NSLibraryDirectory,NSUserDomainMask,YES);
    if ([ary count] > 0) {
      path = [NSMutableString stringWithFormat:@"%@/Logs/%@",[ary objectAtIndex:0],p_exec];
    }
  }
  if (!path) path = [NSMutableString stringWithFormat:@"%@",NSHomeDirectory()];
  if (folder) [path appendString:folder];
  
  name = [NSString stringWithFormat:@"%@/%@%s.log",path,baseName,tstr];
  
  return name;
}

/* ---------------------------------------------------------------- */

- (void)setLogfile:(NSString*)file
{
  char buf[M_FILE];
   
  NSString *path = [file stringByDeletingLastPathComponent];
  NSFileManager *fm = [[NSFileManager alloc] init]; // MT-safe
  [fm createDirectoryAtPath:path  withIntermediateDirectories:YES
      attributes:nil error:NULL];      // create path if necessary
  [fm release];
  if (sharedPath) YfilePermissions(Ystr2tmp(file),6,6,6); // rw-,rw-,rw-

  [fileLock lock];
  if (logfile) [logfile release];
  logfile = [file retain];
  [fileLock unlock];
  
  [self message:Ystr2tmp(file) File:YES];
  sprintf(buf,"build= %s (%s)",Ystr2tmp(p_build),Ystr2tmp(p_version));
  [self append:buf];
}

/* ---------------------------------------------------------------- */

- (void)setField:(NSTextField*)field
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,field);
#endif
  [messageLock lock];
  messageField = field;
  [messageLock unlock];
}

/* ---------------------------------------------------------------- */

- (void)setTable:(NSTableView*)table
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,table);
#endif
  
  [messageLock lock];
  messageTable = table;
  if (table) [table setDataSource:self];
  [messageLock unlock];    
}

/* ---------------------------------------------------------------- */

- (void)setAutoChange:(BOOL)autoChange
{
  dispatch_async(dispatch_get_main_queue(),^(void) {
    if (timer) {
      [timer invalidate];
      [timer release]; timer = nil;
    }
    if (autoChange) {
      timer = [NSTimer scheduledTimerWithTimeInterval:60.0
             target:self selector:@selector(fireTimer:)
             userInfo:nil repeats:YES];
      [timer retain];
    }
  });
}

/* ---------------------------------------------------------------- */

- (void)fireTimer:(NSTimer*)theTimer
{
#if (DEBUG > 0)
  BOOL r = [self changeLogfile];
  if (r) fprintf(stderr,"%s: changed logfile '%s'\n",PREFUN,Ystr2tmp(logfile));
#else
  [self changeLogfile];
#endif
}

/* ---------------------------------------------------------------- */

- (BOOL)changeLogfile
{
  BOOL r=FALSE;
#if (DEBUG > 1)
  fprintf(stderr,"%s:%p\n",PREFUN,self);
#endif

  NSString *string = [self logfileName];
  if ([logfile compare:string] != NSOrderedSame) { // name changed
    [self setLogfile:string];                      // set new name
    r = TRUE;
  }

  return r;
}

/* ---------------------------------------------------------------- */

static void replaceElement(NSMutableArray* array,NSString* string,int head)
{
  if (head < [array count]) {  // NSMutableArray protected by 'messageLock'
    [array replaceObjectAtIndex:head withObject:string];
  } else {
    [array addObject:string];
  }
}

/* ---------------------------------------------------------------- */

- (void)addMessage:(const char*)text
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(): '%s'\n",PREFUN,text);
#endif
  assert(messageLock);
  assert(text);

  if (messageField) {                  // overwrite text field
    [messageLock lock];
    [messageField performSelectorOnMainThread:@selector(setStringValue:)
                  withObject:Ychr2str(text) waitUntilDone:YES];
    [messageLock unlock];
  }
  if (messageTable) {                  // add to table
    [messageLock lock];
    if (showUT) { char buf[32]; struct tm *now,res;
      time_t ut = time(NULL);
      now = gmtime_r(&ut,&res);
      strftime(buf,sizeof(buf),"%H:%M:%S",now);
      replaceElement(timeArray,Ychr2str(buf),mesg_head);
    } else {
      assert(timeFormatter);
      replaceElement(timeArray,[timeFormatter stringFromDate:[NSDate date]],mesg_head);
    }
    assert([timeArray count] <= n_msg);
    replaceElement(mesgArray,Ychr2str(text),mesg_head);
    assert([mesgArray count] <= n_msg);
    mesg_head = (mesg_head+1) % n_msg;
    [messageLock unlock];
    [messageTable performSelectorOnMainThread:@selector(reloadData)
                  withObject:nil waitUntilDone:NO];
  }
}

/* ---------------------------------------------------------------- */

- (void)message:(const char*)text Level:(int)level Time:(int)dt
           File:(BOOL)file Suppress:(BOOL)suppress
{
  char *ptr=NULL;
#if (DEBUG > 1)
  fprintf(stderr,"%s(%s)\n",PREFUN,text);
#endif

  if (file || (level >= 0)) { char *p,*q=(char*)text; // sanitize
    p = ptr = (char*)malloc(1+strlen(text));
    while (*q) {
      switch (*q) {
      case '\r': case '\n':            // replace <CR>,<LF> with <SPC>
        *p = ' ';
        break;
      default:
        *p = *q;
        break;
      }
      p++; q++;
    }
    *p = *q;                           // terminator
  }

  if (file) [self append:ptr];

  if (level >= 0) {                    // write to message window
    [self addMessage:ptr];
#ifdef INCLUDE_NOTICE_H
    if (level > 0) { BOOL show=YES;    // open notice?
      if (suppress) {                  // same text as last time?
        NSString *string = Ychr2str(text);
        @synchronized(self) {
          if (!last_text || ([last_text compare:string] != NSOrderedSame) || (dt > 0)) {
            if (last_text) [last_text release];
            last_text = [string retain];
            show = YES;
          } else show = NO;
        }
      }
      if (show) {
        [Notice openNotice:text Title:Ystr2tmp(p_exec) Time:dt Level:level];
      }
    }
#endif
  }

  if (ptr) free((void*)ptr);
}

/* --- */

- (void)message:(const char*)text Level:(int)level Time:(int)dt File:(BOOL)file
{
  [self message:text Level:level Time:dt File:file Suppress:NO]; // bug-fix 2014-12-08
}

/* --- */

- (void)message:(const char*)text Level:(int)level Time:(int)dt
{
  [self message:text Level:level Time:dt File:NO];
}

/* --- */

- (void)message:(const char*)text Level:(int)level File:(BOOL)file
{
  [self message:text Level:level Time:0 File:file];
}

/* --- */

- (void)message:(const char*)text File:(BOOL)file
{
  [self message:text Level:0 Time:0 File:file];
}

/* --- */

- (void)message:(const char*)text Level:(int)level
{
  [self message:text Level:level Time:0 File:NO];
}

/* --- */

- (void)message:(const char*)text
{
  [self message:text Level:0 Time:0 File:NO];
}

/* ---------------------------------------------------------------- */

- (void)append:(const char*)text
{
  [self append:text toArray:nil];
}

/* ---------------------------------------------------------------- */

- (void)append:(const char*)text toArray:(NSMutableArray*)array
{
  char file[M_FILE];
  FILE *fp;
  NSString *string=nil;
#if (DEBUG > 1)
  fprintf(stderr,"%s(%s)\n",PREFUN,text);
#endif
  
  [fileLock lock];
  fp = fopen(Ystr2chr(logfile,file,sizeof(file)),"a");
  if (fp == NULL) {                  // append to logfile
    fprintf(stderr,"%s(): cannot open file '%s'\n",PREFUN,file);
  } else { char buf[32]; struct tm *now,res; time_t ut;
    ut = time(NULL);  
    if (showUT) now = gmtime_r(&ut,&res);
    else        now = localtime_r(&ut,&res);
    strftime(buf,sizeof(buf),"%b %d %H:%M:%S",now);
    fprintf(fp,"%s %s\n",buf,text);
    fclose(fp);
    if (array) string = [NSString stringWithFormat:@"%s %s",buf,text];
  }
  [fileLock unlock];
  
  if (string) {                      // append to array
    @synchronized(array) {
      [array addObject:string];
    }
  }
}

/* ---------------------------------------------------------------- */
/* NSTableView dataSource */
/* ---------------------------------------------------------------- */

- (NSInteger)numberOfRowsInTableView:(NSTableView*)table
{
#if (DEBUG > 2)
  fprintf(stderr,"%s(%p): msg_n=%d\n",PREFUN,table,msg_n);
#endif
  return [mesgArray count];
}

/* ---------------------------------------------------------------- */

- (id)tableView:(NSTableView*)table
      objectValueForTableColumn:(NSTableColumn*)column
      row:(NSInteger)row
{
  int      i;
  NSString *string=nil;

#if (DEBUG > 1)
  fprintf(stderr,"%s(%p,%p,%d)\n",PREFUN,table,column,row);
#endif
  assert(row < [mesgArray count]);
  if (row >= [mesgArray count]) return nil;
  
#if (DEBUG > 0) && !defined(NDEBUG) && defined(INCLUDE_NOTICE_H)
  if (![NSThread isMainThread]) { char buf[128];
    sprintf(buf,"isMainThread=%d, row=%ld\n",[NSThread isMainThread],(long)row);
    [Notice openNotice:buf Title:PREFUN Time:600 Level:1];
  }
#endif

  [messageLock lock];
  i = mesg_head-(int)row-1; if (i < 0) i += n_msg;
  switch ([[column identifier] intValue]) {
  case 0:                              // time
    string = [timeArray objectAtIndex:i];
    break;
  case 1:                              // text
    string = [mesgArray objectAtIndex:i];
    break;
  default:
    assert(0);
  }
  [messageLock unlock];

  return string;
}

/* ---------------------------------------------------------------- */

@end

/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
